package de.lmu.ifi.pms.cep.dura;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Arrays;
import java.util.Scanner;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.compilation.Compiler;
import de.lmu.ifi.pms.cep.dura.compiler.dura.DuraCompiler;
import de.lmu.ifi.pms.cep.tsa.TSA_Program;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;

import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.file.FileNameFilter;
import de.lmu.ifi.pms.java.util.filter.Filter;
import de.lmu.ifi.pms.java.util.string.RegExFilter;


public class Dura_TSA_Compiler extends Compiler<String[], TSA_Program>{
	
	public Dura_TSA_Compiler() {
		super(String[].class, TSA_Program.class);
	}
	
	Filter<String> inclusionFilter = new RegExFilter(
			".+[.]dura"
			);
	
	Filter<String> exclusionFilter = new RegExFilter(
			".*[.]svn",
			".*[.]git"
			);

	@Override
	public TSA_Program compile(final String[] deal_program){
		String namespace = deal_program[0];
		String simple_name = deal_program[1];
		String module = TSA_SyntaxPatterns.programFullName(namespace, simple_name);
		String[] arguments = Arrays.copyOfRange(deal_program, 2, deal_program.length);
		
		Filter<File> inclusionFilter = new FileNameFilter(this.inclusionFilter);
		Filter<File> exclusionFilter = new FileNameFilter(this.exclusionFilter);
		
		ArrayList<File> source = new ArrayList<File>();
		for(String path: arguments){
			this.collectSourceFiles(new File(path), inclusionFilter, exclusionFilter, source);
		}
		
		if(logger.isDebugEnabled()){
			for(File file: source){
				String text;
				try {
					text = new Scanner( file, "UTF-8" ).useDelimiter("\\A").next();
					this.debug("Source file: \n\n" + text + "\n");
				} catch (FileNotFoundException e) {
					this.debug("Source file not found: " + file.getPath(), e);
					throw new RuntimeException("Source file not found: " + file.getPath(), e);
				}
			}
		}
		
		String[] fileNames = new String[source.size()];
		for(int sourceFile = 0; sourceFile < fileNames.length; ++sourceFile){
			fileNames[sourceFile] = source.get(sourceFile).getPath();
		}		
		
		try {
			return DuraCompiler.compile(module, new FixedArrayNavigableSet<String>(false, fileNames).javaCollection());
		} catch (IOException e) {
			RuntimeException r = new RuntimeException(e.getMessage(), e);
			throw r;
		}
	}
	
	public void collectSourceFiles(final File sourcePath, final Filter<File> inclusionFilter, final Filter<File> exclusionFilter, final ArrayList<File> source){
		if(!exclusionFilter.accepts(sourcePath)){
			if(sourcePath.exists()){
				if(sourcePath.isFile()){
					if(inclusionFilter.accepts(sourcePath)){
						source.add(sourcePath);
					}
				}
				else{assert sourcePath.isDirectory() : "Source path exists, but is neither a file nor a directory: " + sourcePath.getAbsolutePath();
					java.io.FileFilter javaExclusionFilter = new java.io.FileFilter() {				
						@Override
						public boolean accept(final File pathname) {
							return !exclusionFilter.accepts(pathname);
						}
					};
					for(File file: sourcePath.listFiles(javaExclusionFilter)){
						this.collectSourceFiles(file, inclusionFilter, exclusionFilter, source);
					}
				}
			}
			else{
				this.debug("Source path does not exist: {}", sourcePath.getAbsolutePath());
			}
		}
	}
	
	protected static final Logger logger = LoggerFactory.getLogger(Dura_TSA_Compiler.class);
	
	@Override
	protected Logger logger(){
		return logger;
	}
	
}
