tree grammar DuraCAstValidation;

options {
	tokenVocab=Dura;
	ASTLabelType=CommonTree;
}

@members {
final private boolean OUTPUT_HTML = false;
final private StringBuilder sb = new StringBuilder();

private int depth = 0;

private void ind() {
  for (int i = 0; i < depth; i++) {
//    System.out.print("  ");
		sb.append("  ");
  }
}

private void em(String str) {
//  System.out.print(str);
		sb.append(str);
}
private void em(String cla, String str) {
  emTagOpen(cla);
  sb.append(str);
//  System.out.print(str);
  emTagClose();
}

private void emLn(String str) {
//  System.out.println(str);
  sb.append(str+"\n");
}

private void emLn(String cla, String str) {
  emTagOpen(cla);
//  System.out.println(str);
	sb.append(str+"\n");  
  emTagClose();
}

private void indEm(String str) {
  ind();
//  System.out.print(str);
  sb.append(str);  
}
private void indEm(String cla, String str) {
  emTagOpen(cla);
  ind();
//  System.out.print(str);
	sb.append(str);  
  emTagClose();
}

private void indEmLn(String str) {
  ind();
//  System.out.println(str);
	sb.append(str+"\n");  
}
private void indEmLn(String cla, String str) {
  emTagOpen(cla);
  ind();
//  System.out.println(str);
	sb.append(str+"\n");  
  emTagClose();
}

private void emHtmlPre() {
  if (!OUTPUT_HTML) {
    return;
  }

	sb.append(  
//	System.out.println(
    "<html>" +
    "<head>" +
    "<style type=\"text/css\">" +
    "body {background-color: #101010; font-weight: bold;}" +
    ".typeDefinition {color: #bf5930;}" +
    ".constDefinition {color: #ffad40;}" +
    ".eventDefinition {color: #3f92d2;}" +
    ".stateDefinition {color: #36d792;}" +
    ".actionDefinition {color: #bf8230;}" +
    ".eventSpecification {color: #ff7640;}" +
    ".reactiveRule {color: #ff9200;}" +
    ".eventQuery {color: #ff9b73;}" +
    ".flatEventQuery {color: #ff8b00;}" +
    ".atomicEventQuery {color: #0b6ca2;}" +
    ".action {color: #bbbe30;}" +
    ".atomicAction {color: #00af64;}" +
    ".querySupplement {color: #ff7640;}" +
    ".actionSupplement {color: #bfb230;}" +
    ".let {color: #ff3100;}" +
    ".unification {color: #bfae30;}" +
    ".actionSupplement {color: #7945d6;}" +
    ".grouping {color: #ffe740;}" +
    ".binding {color: #ffce40;}" +
    ".aggregation {color: #f13c76;}" +
    ".where {color: #4e51d8;}" +
    ".atomicConditionFormula {color: #36d986;}" +
    ".atomicMathFormula {color: #b939d3;}" +
    ".atomicIntervalFormula {color: #edfc72;}" +
    ".timeInterval {color: #4f4fd9;}" +
    ".timePoint {color: #f16d95;}" +
    ".duration {color: #8ef13c;}" +
    ".time {color: #ffcf40;}" +
    ".term {color: #ff1e00;}" +
    ".termLeaf {color: #7247d7;}" +
    ".constTerm {color: #ff5640;}" +
    ".schemaDefinition {color: #8d6dd7;}" +
    ".expr {color: #248f40;}" +
    ".mathExpr {color: #a5ef00;}" +
    ".label {color: #bfaa30;}" +
    ".type {color: #fe0a00;}" +
    ".basicType {color: #ffff00;}" +
    ".expr {color: #1049a9;}" +
    ".compositeType {color: #00ffff;}" +
    ".typedDataVariable {color: #ff00ff;}" +
    ".dataVariable {color: #a43dd4;}" +
    ".timeRelation {color: #e165b9;}" +
    ".relativeTimerOp {color: #7c71d8;}" +
    ".intervalOp {color: #ffce40;}" +
    ".aggregationOp {color: #bf4e30;}" +
    ".timeUnit {color: #7872d8;}" +
    ".arithmeticRelation {color: #ffde73;}" +
    ".path {color: #ee0000;}" +
    ".typeSupplement {color: #00ee00;}" +
    ".conditionFormula {color: #0000ee;}" +
    ".constant {color: #bf7e30;}" +
    ".identifier {color: #bf7330;}" +
    ".pckg {color: #0e00ee;}" +
    ".description {color: #0000ee;}" +
    ".schemaSupplement {color: #ff00fe;}" +
    ".tupleConstraints {color: #f300fe;}" +
    ".relationConstraints {color: #0f002e;}" +
    ".relationConstraintFormula {color: #2200fe;}" +
    ".relationConstraint {color: #21fffe;}" +
    ".modifiers {color: #00f20fe;}" +
    ".modifier {color: #0ff2fe;}" +
    ".clockFunction {color: #20ef0e;}" +
    "</style>" +
    "</head>" +
    "<body>" +
    "<pre>" +
    "\n"
  );
}

private void emHtmlPost() {
  if (!OUTPUT_HTML) {
    return;
  }

	sb.append(  
//  System.out.println(
    "</pre>" +
    "</body>" +
    "\n"
  );
}

private void emTagOpen(String cla) {
  if (!OUTPUT_HTML) {
    return;
  }
  
	sb.append("<code class=\"" + cla + "\" title=\"" + cla + "\">");  
//  System.out.print("<code class=\"" + cla + "\" title=\"" + cla + "\">");
}

private void emTagClose() {
  if (!OUTPUT_HTML) {
    return;
  }
  
  sb.append("</code>");
//  System.out.print("</code>");
}
}

@header {
package de.lmu.ifi.pms.cep.dura.compiler.antlr;
}

/* DURA Subset*/

program returns [String value]
  : {emHtmlPre();} ^(PROGRAM ^(PACKAGE preamble ^(DEFINITIONS (eventDefinition | stateDefinition | actionDefinition | reactiveRule)+))) {emHtmlPost(); $value=sb.toString();}
  ;

preamble
  : ^(PREAMBLE pckg description constDefinition* typeDefinition*)
  ;

pckg
  : {emTagOpen("pckg");}
    ^(PACKAGE_NAME ({indEm("PACKAGE ");} path {indEmLn(""); emLn("");})?)
    {emTagClose();}
  ;

description
  : {emTagOpen("description");}
    ^(DESCRIPTION (STRING {indEmLn("DESCRIPTION " + $STRING.text); emLn("");})?)
    {emTagClose();}
  ;

typeDefinition 
  : {emTagOpen("typeDefinition");}
    ^(TYPEDEF {indEmLn("TYPE"); depth++; indEmLn($TYPEDEF.text); depth--; indEmLn("IS");}
      {depth++; ind();} basicType {emLn("");} typeSupplement? {depth--; indEmLn("END"); emLn("");})
    {emTagClose();}
  | {emTagOpen("typeDefinition");}
    ^(TYPEDEF {indEmLn("TYPE"); depth++; indEmLn($TYPEDEF.text); depth--; indEmLn("IS");}
      {depth++; indEmLn("{"); depth++; ind();} schemaDefinition ({emLn(","); ind();} schemaDefinition)*
      {emLn(""); depth--; indEmLn("}");} typeSupplement? {depth--; indEmLn("END"); emLn("");})
    {emTagClose();}
  ;
  
constDefinition
  : {emTagOpen("constDefinition");}
    ^(CONSTDEF {indEmLn("CONST"); depth++;} {indEmLn($CONSTDEF.text); depth--; indEmLn("IS");}
      {depth++; ind();} constTerm {depth--; indEmLn("END"); emLn("");})
    {emTagClose();}
  ;
  
eventDefinition
  : {emTagOpen("eventDefinition");}
    ^(EVENT modifiers {indEmLn("EVENT");}
      ^(SCHEMA {depth++; ind();} schemaDefinition {emLn("");} schemaSupplement {depth--;})
      ({indEmLn("WITH"); depth++;} eventSpecification ({emLn("");} eventSpecification)* {depth--;})? {indEmLn("END");}
      (^(ANNOTATION {emLn("@annotation:"); depth++;} (STRING {indEmLn("* " + $STRING.text);})+ ) {depth--;})* {emLn("");}
    )
    {emTagClose();}
  ;

stateDefinition
  : {emTagOpen("stateDefinition");}
    ^(STATEFUL_OBJECT modifiers {indEmLn("STATEFUL OBJECT");}
      ^(SCHEMA {depth++; ind();} schemaDefinition {emLn(""); depth--;} schemaSupplement {indEmLn("END");})
      (^(ANNOTATION {emLn("@annotation:"); depth++;} (STRING {indEmLn("* " + $STRING.text);})+ ) {depth--;})* {emLn("");}
    )
    {emTagClose();}
  ;

actionDefinition
  : {emTagOpen("actionDefinition");}
    ^(ACTION modifiers {indEmLn("ACTION");}
      ^(SCHEMA {depth++; ind();} schemaDefinition {emLn(""); depth--;} schemaSupplement {indEmLn("END");})
      (^(ANNOTATION {emLn("@annotation:"); depth++;} (STRING {indEmLn("* " + $STRING.text);})+ ) {depth--;})* {emLn("");}
    )
    {emTagClose();}
  ;

eventSpecification
  : {emTagOpen("eventSpecification");}
    ^(DETECT {indEmLn("DETECT");}
      ^(HEAD {depth++; ind();} term {emLn(""); depth--; indEmLn("ON");})
      ^(BODY {depth++;} eventQuery {depth--; indEmLn("END");})
    )
    {emTagClose();}
  ;

reactiveRule
  : {emTagOpen("reactiveRule");}
    ^(DO {indEmLn("ON");}
      ^(BODY {depth++;} eventQuery {depth--; indEmLn("DO");})
      ^(HEAD {depth++;} action {depth--; indEmLn("END");})
      (^(ANNOTATION {emLn("@annotation:"); depth++;} (STRING {indEmLn("* " + $STRING.text);})+ ) {depth--;})* {emLn("");}
    )
    {emTagClose();}
  ;

eventQuery
  : {emTagOpen("eventQuery");} (
    ^(CONJUNCTION {indEmLn("and {");} {depth++;} flatEventQuery ({emLn(",");} flatEventQuery)* {emLn(""); depth--; indEm("}");} querySupplement {emLn("");})
  | ^(DISJUNCTION {indEmLn("or {");} {depth++;} flatEventQuery ({emLn(",");} flatEventQuery)* {emLn(""); depth--; indEm("}");} querySupplement {emLn("");})
    ) {emTagClose();}
  ;

flatEventQuery
  : {emTagOpen("flatEventQuery");} (
    ^(NEGATION {indEmLn("not { ");} {depth++; ind();} atomicEventQuery {emLn(""); depth--; indEm("}");} querySupplement)
  | {ind();} atomicEventQuery
    ) {emTagClose();}
  ;

atomicEventQuery
  : {emTagOpen("atomicEventQuery");} 
    ^(EVENT_QUERY ^(IDENTIFIER {em($IDENTIFIER.text);} (ID {em(" " + $ID.text);})?) {em(": ");} ^(QUERY term) querySupplement)
    {emTagClose();}
  | {emTagOpen("atomicEventQuery");}
    ^(STATE_QUERY ^(IDENTIFIER {em($IDENTIFIER.text);} (ID  {em(" " + $ID.text);})?) {em(": ");} ^(QUERY term) querySupplement)
    {emTagClose();} 
  ;

action
  : {emTagOpen("action");}
    ^(CONCURRENT {indEmLn("concurrent{");} {depth++;} flatAction ({emLn(",");} flatAction)* {emLn(""); depth--; indEmLn("}");}  EXECUTION_CONSTRAINTS SUCCEEDS_ON FAILS_ON)
    {emTagClose();}
  ;

flatAction
  : {emTagOpen("atomicAction");}
    ^(ACTION_EXECUTION ^(IDENTIFIER {indEm($IDENTIFIER.text);} (ID {em(" " + $ID.text);})?) {em(": ");} term)
    {emTagClose();}
  ;
  
atomicAction
  : {emTagOpen("atomicAction");}
    ^(ACTION_EXECUTION {indEm("action: ");} term)
    {emTagClose();}
  ;

querySupplement
  : {emTagOpen("querySupplement");}
    ^(SUPPLEMENT ({emLn("");} (where | let | grouping))* )
    {emTagClose();}
  ;

actionSupplement
  : {emTagOpen("actionSupplement");}
    ^(SUPPLEMENT (where ({emLn("");} where)*)?)
    {emTagClose();}
  ;
  
typeSupplement
  : {emTagOpen("typeSupplement");}
    ^(SUPPLEMENT tupleConstraints?)
    {emTagClose();}
  ;
  
schemaSupplement
  : {emTagOpen("schemaSupplement");}
    ^(SUPPLEMENT tupleConstraints? relationConstraints?)
    {emTagClose();}
  ;
  
let
  : {emTagOpen("let");}
    ^(LET {indEmLn("let {");} {depth++; indEm("");} unification ({emLn(","); indEm("");} unification)* {emLn(""); depth--; indEm("}");})
    {emTagClose();}
  ;

unification
  : {emTagOpen("unification");}
    ^(UNIFICATION dataVariable {em(" = ");} expr)
    {emTagClose();}
  ;

grouping
  : {emTagOpen("grouping");} (
    ^(AGGREGATION ^(GROUPING {indEmLn("group by {");} {depth++; indEm("");} binding ({emLn(","); indEm("");} binding)*) {emLn(""); depth--; indEm("}");})
  | ^(AGGREGATION ^(GROUPING {indEmLn("group by {");} {depth++; indEm("");} binding ({emLn(","); indEm("");} binding)*) {emLn(""); depth--; indEmLn("}");}
      {indEmLn("aggregate {");} {depth++; indEm("");} aggregation ({emLn(","); indEm("");} aggregation)* {emLn(""); depth--; indEm("}");}
    )
    ) {emTagClose();}
  ;
  
binding
  : {emTagOpen("binding");} (
    dataVariable
  | identifier 
    ) {emTagClose();}
  ;

aggregation
  : {emTagOpen("aggregation");}
    ^(UNIFICATION typedDataVariable {em(" = ");} ^(aggregationOp {em($aggregationOp.text + "(");} dataVariable {em(")");}))
    {emTagClose();}
  ;

tupleConstraints
  : {emTagOpen("tupleConstraints");}
    ^(TUPLE_CONSTRAINTS {indEmLn("where {");} {depth++; indEm("");}
      atomicMathFormula ({emLn(",");  indEm("");} atomicMathFormula)* {emLn(""); depth--; indEmLn("}");}
    )
    {emTagClose();}
  ;

relationConstraints
  : {emTagOpen("relationConstraints");}
    ^(RELATION_CONSTRAINTS {indEmLn("constraints {");} {depth++; indEm("");}
      relationConstraintFormula ({emLn(",");  indEm("");} relationConstraintFormula)* {emLn(""); depth--; indEmLn("}");}
    )
    {emTagClose();}
  ;

relationConstraintFormula
  : {emTagOpen("relationConstraintFormula");} (
    ^(CONJUNCTION {emLn("and {"); depth++; ind();} relationConstraintFormula ({emLn(","); ind();} relationConstraintFormula)* {emLn(""); depth--; indEm("}");})
  | ^(DISJUNCTION {emLn("or {"); depth++; ind();} relationConstraintFormula ({emLn(","); ind();} relationConstraintFormula)* {emLn(""); depth--; indEm("}");})
  | relationConstraint
    ) {emTagClose();}
  ;

relationConstraint
  : {emTagOpen("relationConstraint");}
    ^(CONSTRAINT {indEm($CONSTRAINT.text + " (");} path+ {emLn(")");})
    {emTagClose();}
  ;

where
  : {emTagOpen("where");}
    ^(CONDITION {indEm("where ");} conditionFormula)
    {emTagClose();}
  ;

conditionFormula
  : {emTagOpen("conditionFormula");} (
    ^(CONJUNCTION {emLn("and {"); depth++; ind();} conditionFormula ({emLn(","); ind();} conditionFormula)* {emLn(""); depth--; indEm("}");})
  | ^(DISJUNCTION {emLn("or {"); depth++; ind();} conditionFormula ({emLn(","); ind();} conditionFormula)* {emLn(""); depth--; indEm("}");})
  | atomicConditionFormula
    ) {emTagClose();}
  ;

atomicConditionFormula
  : {emTagOpen("atomicConditionFormula");} (
    atomicIntervalFormula
  | atomicMathFormula
    ) {emTagClose();}
  ;

atomicMathFormula 
  : {emTagOpen("mathFormula");}
    ^(arithmeticRelation expr {em(" " + $arithmeticRelation.text + " ");} expr)
    {emTagClose();}
  ;
  
atomicIntervalFormula
  // TODO: The semantic predicate is a workaround required due to the excessive overloading of the
  //       TIME_RELATION token and overlapping of timeInterval and timePoint. 
  : {emTagOpen("intervalFormula");}
    ^(timeRelation
        ( {em("{");} ^(IDENTIFIERS timeInterval ({em(", ");} timeInterval)*) {em("} " + $timeRelation.text + " ");} duration
        | timeInterval {em(" " + $timeRelation.text + " ");} ({ $timeRelation.text.equals("valid-at") }?=>timePoint | timeInterval)
        )
     )
    {emTagClose();} 
  ;

timeInterval 
  : {emTagOpen("timeInterval");} (
    dataVariable
  | identifier
  | ^(relativeTimerOp {em($relativeTimerOp.text + "(");} timeInterval {em(", ");} duration {em(")");})
    ) {emTagClose();}
  ;

timePoint
  : {emTagOpen("timePoint");} (
    dataVariable
  | ^(intervalOp {em($intervalOp.text + "(");} timeInterval {em(")");})
    ) {emTagClose();} 
  ;

duration
  : {emTagOpen("duration");}
    ^(DURATION time ({em(" ");} time)*)
    {emTagClose();}
  ;

time
  : {emTagOpen("time");}
    ^(timeUnit NUMBER {em($NUMBER.text + " " + $timeUnit.text);})
    {emTagClose();}
  ;

/* XCerpt- */

term
  : {emTagOpen("term");} (
    label {em("{}");}
  | ^(label {em("{ ");} termLeaf {em(" }");})
  | ^(label {emLn("{"); depth++; ind();} term ({emLn(","); ind();} term)* {emLn(""); depth--; indEm("}");})
    ) {emTagClose();}
  ;

termLeaf
  : {emTagOpen("termLeaf");} (
    dataVariable
  | identifier
  | constant
  | duration
  | STRING {em($STRING.text);}
  | NUMBER {em($NUMBER.text);}
  | BOOLEAN {em($BOOLEAN.text);}
    ) {emTagClose();}
  ;

constTerm
  : {emTagOpen("constTerm");} (
    STRING {emLn($STRING.text);}
  | NUMBER {emLn($NUMBER.text);}
  | BOOLEAN {emLn($BOOLEAN.text);}
  | constant {emLn("");}
  | label {emLn("{}");}
  | ^(label {emLn("{"); depth++; ind();} constTerm ({emLn(","); ind();} constTerm)* {emLn(""); depth--; indEm("}");})
    ) {emTagClose();}
  ;

schemaDefinition
  : {emTagOpen("schemaDefinition");} (
    label {em("{}");}
  | ^(label {em("{ ");} basicType {em(" }");})
  | ^(label {em("{ ");} compositeType {em(" }");})
  | ^(label {emLn("{"); depth++; ind();} schemaDefinition ({emLn(","); ind();} schemaDefinition)* {emLn(""); depth--; indEm("}");})
    ) {emTagClose();}
  ;

modifiers
  : {emTagOpen("modifiers");}
    ^(MODIFIERS modifier*)
    {emTagClose();}
  ;

modifier 
	: {emTagOpen("modifier");} (
	  (EXTERNAL_MODIFIER {indEmLn($EXTERNAL_MODIFIER.text);})
	| (DYNAMABILITY_MODIFIER {indEmLn($DYNAMABILITY_MODIFIER.text);}) 
	  ) {emTagClose();}
	;
  
expr  
  : {emTagOpen("expr");} (
    mathExpr
  | identifier
    ) {emTagClose();}
  ;

mathExpr
  : {emTagOpen("mathExpr");}
    ^(ARITHMETIC_OP
      (  (mathExpr mathExpr)=> {em("(");} mathExpr {em(") " + $ARITHMETIC_OP.text + " (");} mathExpr {em(")");}
      |  (mathExpr)=> {em($ARITHMETIC_OP.text +"(");} mathExpr {em(")");}
      )      
    )
    {emTagClose();}
  | {emTagOpen("mathExpr");}
    ^(intervalOp {em($intervalOp.text + "(");} timeInterval {em(")");})
    {emTagClose();}
  | {emTagOpen("mathExpr");}
    clockFunction
    {emTagClose();}
  | {emTagOpen("mathExpr");}
    dataVariable
    {emTagClose();}
  | {emTagOpen("mathExpr");}
    constant
    {emTagClose();}
  | {emTagOpen("mathExpr");}
    path
    {emTagClose();}
  | {emTagOpen("mathExpr");}
    NUMBER {em($NUMBER.text);}
    {emTagClose();}
  | {emTagOpen("mathExpr");}
    ^(DURATION ^(timeUnit NUMBER {em($NUMBER.text + " " + $timeUnit.text);}) ({em(" ");} time)*)
    {emTagClose();}
  | {emTagOpen("mathExpr");}
    STRING {em($STRING.text);}
    {emTagClose();}
  | {emTagOpen("mathExpr");}
    BOOLEAN {em($BOOLEAN.text);}
    {emTagClose();}

  ;

path
  : {emTagOpen("path");}
    ^(PATH label ({em(".");} label)* )
    {emTagClose();}
  ;

 
clockFunction
  : {emTagOpen("clockFunction");}
    CLOCK_FUNCTION {em("system." + $CLOCK_FUNCTION.text + "()");}
    {emTagClose();}
  ;

label 
  : {emTagOpen("label");}
    TERM {em($TERM.text);}
    {emTagClose();}
  ;

type
  : {emTagOpen("type");} ( 
    basicType
  | compositeType
    ) {emTagClose();}
  ;

basicType
  : {emTagOpen("basicType");}
    BASIC_TYPE {em($BASIC_TYPE.text);}
    {emTagClose();}
  ;

compositeType
  : {emTagOpen("compositeType");}
    COMPOSITE_TYPE {em($COMPOSITE_TYPE.text);}
    {emTagClose();}
  ;

identifier
  : {emTagOpen("identifier");}
    ^(IDENTIFIER ID {em($IDENTIFIER.text + " " + $ID.text);})
    {emTagClose();}
  ;

dataVariable
  : {emTagOpen("dataVariable");}
    ^(VARIABLE ID {em("var " + $ID.text);})
    {emTagClose();}
  ;
  
typedDataVariable
  : {emTagOpen("typedDataVariable");}
    ^(TYPED_VARIABLE {em("var ");} type {em(" ");} ID {em($ID.text);})
    {emTagClose();}
  | dataVariable		//Hack
  ;

constant
  : {emTagOpen("constant");}
    ^(CONSTANT ID {em("const " + $ID.text);})
    {emTagClose();}
  ;

timeRelation returns [String text]
  : TIME_RELATION {$text = (OUTPUT_HTML ? "<code class=\"timeRelation\" title=\"timeRelation\">" + $TIME_RELATION.text + "</code>" : $TIME_RELATION.text);}
  ;

relativeTimerOp returns [String text]
  : RELATIVE_TIMER_OP {$text = (OUTPUT_HTML ? "<code class=\"relativeTimerOp\" title=\"relativeTimerOp\">" + $RELATIVE_TIMER_OP.text + "</code>" : $RELATIVE_TIMER_OP.text);}
  ;

intervalOp returns [String text]
  : INTERVAL_OP {$text = (OUTPUT_HTML ? "<code class=\"intervalOp\" title=\"intervalOp\">" + $INTERVAL_OP.text + "</code>" : $INTERVAL_OP.text);}
  ;

aggregationOp returns [String text]
  : AGGREGATION_OP {$text = (OUTPUT_HTML ? "<code class=\"aggregationOp\" title=\"aggregationOp\">" + $AGGREGATION_OP.text + "</code>" : $AGGREGATION_OP.text);}
  ;

timeUnit returns [String text]
  : TIME_UNIT {$text = (OUTPUT_HTML ? "<code class=\"timeUnit\" title=\"timeUnit\">" + $TIME_UNIT.text + "</code>" : $TIME_UNIT.text);}
 ;

arithmeticRelation returns [String text]
  : ARITHMETIC_RELATION {$text = (OUTPUT_HTML ? "<code class=\"arithmeticRelation\" title=\"arithmeticRelation\">" + $ARITHMETIC_RELATION.text + "</code>" : $ARITHMETIC_RELATION.text);}
  ;