package de.lmu.ifi.pms.cep.dura.compiler.antlr.utils;

import java.io.File;

import org.antlr.runtime.RecognitionException;
import org.antlr.runtime.tree.CommonTree;
import org.antlr.runtime.tree.CommonTreeNodeStream;
import org.antlr.runtime.tree.Tree;

import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCAstValidation;
import de.lmu.ifi.pms.cep.dura.compiler.dura.desugaring.Desugarer;

public class DuraTree extends CommonTree {
	private final String filename;
	
	DuraTree(final CommonTree tree) {
		super(tree);
		
		this.filename = null;
	}
	
	public DuraTree(final Object tree, final String filename) {
		this((CommonTree) tree, null, filename);
	}
	
	public DuraTree(final CommonTree tree, final String filename) {
		this(tree, null, filename);
	}
	
	public DuraTree(final CommonTree tree, final DuraTree parent, final String filename) {
		super(tree);
		
		this.setChildIndex(tree.childIndex);
		this.setParent(parent);
		
		int n = tree.getChildCount();
		for (int i = 0; i < n; i++) {
			CommonTree child = (CommonTree) tree.getChild(i);
			DuraTree newSubTree = new DuraTree(child, this, filename);
			this.addChild(newSubTree);
		}

		this.filename = new File(filename).getAbsolutePath();//FIXME modified to obtain absolute path 
	}
	
	@Override
	public DuraTree getChild(final int i) {
		if (this.children==null || i>=this.children.size()) {
			return null;
		}
		
		return (DuraTree) this.children.get(i);
	}
	
	@Override
	public DuraTree getParent() {
		return (DuraTree) this.parent;
	}
	
	public String getFilename() {
		return this.filename;
	}
	
	public String getPositionDescription() {
		return getPositionDescription(this);
	}
	
	public String getPositionDescription(DuraTree tree) {
		if (tree.comprisesMultipleLines()) {
			return String.format("\tat %s:%d-%d", tree.getFilename(), tree.getLine(), tree.getLastLine());
		} else {
			return String.format("\tat %s:%d", tree.getFilename(), tree.getLine());
		}
	}
	
	public int getLastLine() {
		return getLastLine(this);
	}
	
	public int getLastLine(DuraTree tree) {
		int n = tree.getChildCount();
		
		if (n == 0) {
			return tree.getLine();
		} else {
			return tree.getChild(n-1).getLastLine();
		}
	}
	
	/*
	public int getLine() {
		if ( token==null || token.getLine()==0 ) {
			if ( getChildCount()>0 ) {
				return getChild(0).getLine();
			} else if (parent.token != null){
				return parent.token.getLine();
			}
			return 0;
		}
		return token.getLine();
	}*/
	
	public boolean comprisesMultipleLines() {
		return this.getLine() < this.getLastLine();
	}
	
	public String toString() {
		if (comprisesMultipleLines()) {
			return String.format("%s[%d-%d]", super.toString(), getLine(), getLastLine());			
		} else {
			return String.format("%s[%d]", super.toString(), getLine());
		}
	}
	
	public String toStringTree() {
		return toStringTree("");
	}
	
    private String toStringTree(String indent) {
		if ( children==null || children.size()==0 ) {
			return this.toString();
		}
		
		StringBuffer buf = new StringBuffer();
		
		if ( !isNil() ) {
			buf.append("(");
			buf.append(this.toString());
		}
		
		for (int i = 0; children!=null && i < children.size(); i++) {
			DuraTree t = (DuraTree)children.get(i);
			buf.append("\n");
			buf.append(indent+"  ");
			
			buf.append(t.toStringTree(indent+"  "));
			
			if (i !=  children.size()-1) {
				buf.append("\n");
			}
		}
		
		if ( !isNil() ) {
			buf.append(")");
		}
		
		return buf.toString();
	}
}
