package de.lmu.ifi.pms.cep.dura.compiler.dura.desugaring;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;

import org.antlr.runtime.CommonToken;
import org.antlr.runtime.tree.CommonTree;

import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraParser;


public class ComplexSubqueries extends Transformation {

	@Override
	public void apply(CommonTree tree) {
		for (CommonTree eventSpecification : collectEventSpecificationNodes(tree)) {
			detachComplexQueries(eventSpecification);
		}
		
		resetNextNameId();
	}
	
	private CommonTree detachComplexQueries(CommonTree eventSpecification) {
		CommonTree after = (CommonTree) eventSpecification.getParent();
		
		if (eventSpecification.getChildCount() > 0) {
			CommonTree annotation = null;
			if (EMIT_ANNOTATIONS) {
				annotation = new CommonTree(new CommonToken(DuraParser.ANNOTATION, "ANNOTATION"));
				annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Result of transformation " + this.getClass().getSimpleName() + ".\"")));
			}
			
			
			CommonTree head = (CommonTree) eventSpecification.getChild(0);
			
			assert(head.getType() == DuraParser.HEAD);
			
			CommonTree composition = (CommonTree) eventSpecification.getChild(1).getChild(0);
			CommonTree supplement = supplement(composition);
			
			List<CommonTree> queries = queries(composition);
			for (CommonTree q : queries) {
				boolean isNegative = false;
				CommonTree subqueryCompositionOrTerm;
				if (q.getType() == DuraParser.NEGATION || q.getType() == DuraParser.EXISTS) {
					subqueryCompositionOrTerm = (CommonTree) q.getChild(0).getChild(1).getChild(0);
					isNegative = true;
				} else {
					subqueryCompositionOrTerm = (CommonTree) q.getChild(1).getChild(0);
				}
				
				boolean isComplex = subqueryCompositionOrTerm.getType() != DuraParser.TERM;
				if (isComplex) {					
					CommonTree supp;
					if (isNegative) {
						supp = (CommonTree) q.getChild(q.getChildCount() - 1);
					} else {
						supp = supplement;
					}
					HashSet<CommonTreeSetWrapper> relBindings = relevantBindings(q, head, getEnvironment(q), supp);
				
					CommonTree term = createTerm(relBindings);
					
					
					
					CommonTree annot = null;
					if (EMIT_ANNOTATIONS) {
						annot = new CommonTree(new CommonToken(DuraParser.ANNOTATION, "ANNOTATION"));
						annot.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Result of transformation " + this.getClass().getSimpleName() + ".\"")));
						annot.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Translation of " + subqueryCompositionOrTerm.getParent().toStringTree() + " in deductive rule " + eventSpecification.toStringTree() + ".\"")));
					}
					
					
					
					
					int index = subqueryCompositionOrTerm.getChildIndex();
					subqueryCompositionOrTerm.getParent().replaceChildren(index, index, term);
					
					CommonTree eventDef = addEventDefinition(copyTree(term), subqueryCompositionOrTerm, after);
					
					
			
					if (EMIT_ANNOTATIONS) {
						eventDef.addChild(annot);
					}
					
					
					
					if (EMIT_ANNOTATIONS) {
						annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Lifting / extraction of " + subqueryCompositionOrTerm.getParent().toStringTree() + " to event " + term.getText() + ".\"")));
					}
					
					
					after = detachComplexQueries((CommonTree)  eventDef.getChild(2));
				}
			}
			
			if (EMIT_ANNOTATIONS) {
				if (annotation.getChildCount() > 1) {
					eventSpecification.getParent().addChild(annotation);
				}
			}
		}
		
		return after;
	}
	
	@SuppressWarnings(value = "unchecked")
	private CommonTree addEventDefinition(CommonTree headTerm, CommonTree composition, CommonTree after) {
		CommonTree eventDef = new CommonTree(new CommonToken(DuraParser.EVENT, "EVENT"));
		eventDef.addChild(new CommonTree(new CommonToken(DuraParser.MODIFIERS, "MODIFIERS")));
		
		CommonTree schema = new CommonTree(new CommonToken(DuraParser.SCHEMA, "SCHEMA"));
		
		// The pseudo schema term contains bindings to be replaced by types.
		CommonTree pseudoSchemaTerm = copyTree(headTerm);
		
		for (int i = 0; i < pseudoSchemaTerm.getChildCount(); i++) {
			CommonTree subTerm = (CommonTree) pseudoSchemaTerm.getChild(i);
			CommonTree bindingNode = (CommonTree) subTerm.getChild(0);
			
			subTerm.deleteChild(0);
			
			if (bindingNode.getType() == DuraParser.IDENTIFIER) {
				CommonTree typeNode = new CommonTree(new CommonToken(DuraParser.BASIC_TYPE, "identifier"));
				subTerm.addChild(typeNode);
			} else {
				List<CommonTree> typeNodes = determineAndGetTypeNodes(composition, bindingNode);
				
				for (CommonTree t : typeNodes) {
					subTerm.addChild(copyTree(t));
				}
			}
		}
		
		explicitEventSchemataMap.put(pseudoSchemaTerm.getText(), pseudoSchemaTerm);
		
		CommonTree pseudoSchemaTermCopy = copyTree(pseudoSchemaTerm);
		joinTermCopiesWithTermChildren(implicitEventSchemaTerms, pseudoSchemaTermCopy);
		
		eventSchemataMap.put(pseudoSchemaTerm.getText(), pseudoSchemaTermCopy);
		// Keep in mind that this does not contain the implicit additions like id, reception-time, etc.
		// We don't need that because it's not exposed to a programmer anyway.
		// Still, as with all schemata related stuff, this id TODO!
		
		
		
		// TODO: Instead of doing this here, one could build a binding-to-type-mapping once
		//       (during semantic analysis!!!) for the event specification and take use of that here.
		//
		//       Or at least some form of memoization.
		
		schema.addChild(pseudoSchemaTerm);
		schema.addChild(new CommonTree(new CommonToken(DuraParser.SUPPLEMENT, "SUPPLEMENT")));
		
		eventDef.addChild(schema);
		
		CommonTree eventSpecification = new CommonTree(new CommonToken(DuraParser.DETECT, "DETECT"));
		
		CommonTree head = new CommonTree(new CommonToken(DuraParser.HEAD, "HEAD"));
		head.addChild(headTerm);
		eventSpecification.addChild(head);
		
		CommonTree body = new CommonTree(new CommonToken(DuraParser.BODY, "BODY"));
		body.addChild(composition);
		eventSpecification.addChild(body);
		
		eventDef.addChild(eventSpecification);
		
		CommonTree parent = (CommonTree) after.getParent();
		parent.getChildren().add(after.getChildIndex() + 1, eventDef);
		parent.freshenParentAndChildIndexes();
		
		return eventDef;
	}
	
	private CommonTree createTerm(HashSet<CommonTreeSetWrapper> bindings) {
		String termName = createName();
		CommonTree term = new CommonTree(new CommonToken(DuraParser.TERM, termName));
		
		int localNameId = 0; 
		for (CommonTreeSetWrapper b : bindings) {
			CommonTree subTerm = new CommonTree(new CommonToken(DuraParser.TERM, termName + "_" + localNameId));
			localNameId++;
			
			subTerm.addChild(copyTree(b.getTree()));
			
			term.addChild(subTerm);
		}
		
		return term;
	}
	
	private HashSet<CommonTreeSetWrapper> relevantBindings(CommonTree query, CommonTree head, List<CommonTree> environment, CommonTree supplement) {
		assert(query.getType() == DuraParser.NEGATION || query.getType() == DuraParser.EXISTS || query.getType() == DuraParser.EVENT_QUERY || query.getType() == DuraParser.STATE_QUERY);
		assert(head.getType() == DuraParser.HEAD);
		assert(supplement.getType() == DuraParser.SUPPLEMENT);

		HashSet<CommonTreeSetWrapper> bindingsSet = bindings(query);
		HashSet<CommonTreeSetWrapper> set = termBindings(term(head));
		boolean queryIsPositive = query.getType() != DuraParser.NEGATION && query.getType() != DuraParser.EXISTS;
		if (queryIsPositive) {
			set.addAll(suppRefs(supplement(head)));
		}
		set.addAll(suppRefs(supplement));
		for (CommonTree q : environment) {
			if (queryIsPositive || (q.getType() != DuraParser.NEGATION && q.getType() != DuraParser.EXISTS)) {
				set.addAll(bindings(q));
			}
		}
		set.retainAll(bindingsSet);
		
		return set;
	}
	
	private HashSet<CommonTreeSetWrapper> suppRefs(CommonTree supplement) {
		assert(supplement.getType() == DuraParser.SUPPLEMENT);
		
		HashSet<CommonTreeSetWrapper> set = new HashSet<CommonTreeSetWrapper>();
		
		Queue<CommonTree> nodeQueue = new LinkedList<CommonTree>();
		nodeQueue.add(supplement);
		
		while(!nodeQueue.isEmpty()) {
			CommonTree node = (CommonTree) nodeQueue.remove();			
			
			if (node.getType() == DuraParser.VARIABLE || node.getType() == DuraParser.IDENTIFIER) {
				set.add(new CommonTreeSetWrapper(node));
			}
			
			if (node.getChildCount() > 0) {
				for (Object c : node.getChildren()) {
					CommonTree child = (CommonTree) c;
					nodeQueue.add(child);
				}
			}
		}
			
		return set;
	}
	
	private IdentityHashMap<CommonTree, HashSet<CommonTreeSetWrapper>> bindingsMemoizationMap = new IdentityHashMap<CommonTree, HashSet<CommonTreeSetWrapper>>(); 
	private HashSet<CommonTreeSetWrapper> bindings(CommonTree query) {
		assert(query.getType() == DuraParser.NEGATION || query.getType() == DuraParser.EXISTS || query.getType() == DuraParser.EVENT_QUERY || query.getType() == DuraParser.STATE_QUERY);
		
		if (query.getType() == DuraParser.NEGATION || query.getType() == DuraParser.EXISTS) {
			query = (CommonTree) query.getChild(0);
		}
		
		HashSet<CommonTreeSetWrapper> set = bindingsMemoizationMap.get(query);
		if (set != null) {
			return set;
		}
		
		set = new HashSet<CommonTreeSetWrapper>();
		set.add(new CommonTreeSetWrapper(identifier(query)));
		boolean isAtomicQuery = query.getChild(1).getChild(0).getType() == DuraParser.TERM;
		if (isAtomicQuery) {
			set.addAll(termBindings(term(query)));
			set.addAll(suppDefs(supplement(query)));
		} else {
			CommonTree composition = composition(query);
			boolean isConjunction = composition.getType() == DuraParser.CONJUNCTION;
			if (isConjunction) {
				for (CommonTree q : queries(composition)) {
					if (q.getType() != DuraParser.NEGATION && q.getType() != DuraParser.EXISTS) {
						set.addAll(bindings(q));
					}
				}
			} else {
				HashSet<CommonTreeSetWrapper> disjunctionSet = new HashSet<CommonTreeSetWrapper>();
				for (CommonTree q : queries(composition)) {
					if (q.getType() != DuraParser.NEGATION && q.getType() != DuraParser.EXISTS) {
						if (disjunctionSet.isEmpty()) {
							disjunctionSet.addAll(bindings(q));
						} else {
							disjunctionSet.retainAll(bindings(q));
						}
					}
				}
				set.addAll(disjunctionSet);
			}
			set.addAll(suppDefs(supplement(composition)));
		}
		
		bindingsMemoizationMap.put(query, set);
		return set;
	}
	
	private CommonTree supplement(CommonTree queryOrCompositionOrHead) {
		assert( queryOrCompositionOrHead.getType() == DuraParser.NEGATION || 
				queryOrCompositionOrHead.getType() == DuraParser.EXISTS ||
				queryOrCompositionOrHead.getType() == DuraParser.CONJUNCTION ||
				queryOrCompositionOrHead.getType() == DuraParser.DISJUNCTION ||
				queryOrCompositionOrHead.getType() == DuraParser.HEAD ||
				(
					(queryOrCompositionOrHead.getType() == DuraParser.EVENT_QUERY || queryOrCompositionOrHead.getType() == DuraParser.STATE_QUERY) &&
					queryOrCompositionOrHead.getChild(1).getChild(0).getType() != DuraParser.TERM
				)
				);
		assert(queryOrCompositionOrHead.getChildCount() > 1);
		
		return (CommonTree) queryOrCompositionOrHead.getChild(queryOrCompositionOrHead.getChildCount() - 1);
	}
	
	private List<CommonTree> queries(CommonTree composition) {
		assert(composition.getType() == DuraParser.CONJUNCTION || composition.getType() == DuraParser.DISJUNCTION);
		
		ArrayList<CommonTree> nodes = new ArrayList<CommonTree>();
		
		int endIndex = composition.getChildCount() - 1;
		if (composition.getChild(endIndex).getType() == DuraParser.SUPPLEMENT) {
			endIndex--;
		}
		
		for (int i = 0; i <= endIndex; i++) {
			nodes.add((CommonTree) composition.getChild(i));
		}
		
		return nodes;
	}
	
	private CommonTree composition(CommonTree query) {
		assert(query.getType() == DuraParser.NEGATION || query.getType() == DuraParser.EXISTS || query.getType() == DuraParser.EVENT_QUERY || query.getType() == DuraParser.STATE_QUERY);
		
		if (query.getType() == DuraParser.NEGATION || query.getType() == DuraParser.EXISTS) {
			query = (CommonTree) query.getChild(0);
		}
		
		assert(query.getChild(1).getChild(0).getType() != DuraParser.TERM);
		
		return (CommonTree) query.getChild(1).getChild(0);
	}
	
	private CommonTree term(CommonTree queryOrHead) {
		assert(queryOrHead.getType() == DuraParser.NEGATION || queryOrHead.getType() == DuraParser.EXISTS ||
			   queryOrHead.getType() == DuraParser.EVENT_QUERY || queryOrHead.getType() == DuraParser.STATE_QUERY ||
			   queryOrHead.getType() == DuraParser.HEAD);
		
		if (queryOrHead.getType() == DuraParser.HEAD) {
			return (CommonTree) queryOrHead.getChild(0);
		}
		
		if (queryOrHead.getType() == DuraParser.NEGATION || queryOrHead.getType() == DuraParser.EXISTS) {
			queryOrHead = (CommonTree) queryOrHead.getChild(0);
		}
		
		assert(queryOrHead.getChild(1).getChild(0).getType() == DuraParser.TERM);
			
		return (CommonTree) queryOrHead.getChild(1).getChild(0);
	}
}
