package de.lmu.ifi.pms.cep.dura.compiler.durac.collections;

import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraLexer.DESCRIPTION;
import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraLexer.IMPORT;
import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraLexer.MODULE;
import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraLexer.PREAMBLE;

import java.util.HashMap;
import java.util.Map;

import org.antlr.runtime.tree.Tree;

import de.lmu.ifi.pms.cep.dura.compiler.antlr.utils.DuraTree;
import de.lmu.ifi.pms.cep.dura.compiler.antlr.utils.TreeIterator;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;


public class Preamble {
	public final String package_name;
	public final String description;

	public final Map<String,String> imports = new HashMap<String, String>();

	public Preamble(DuraTree preamble) {
		assert(preamble.getType() == PREAMBLE);
		
		Tree descriptionNode = new TreeIterator(preamble, DESCRIPTION).next().getChild(0);
		Tree packageNode = new TreeIterator(preamble, MODULE).next().getChild(0);

		this.description = descriptionNode==null ? "" : descriptionNode.getText();
		this.package_name = packageNode==null ? "" : packageNode.getText()+".";
		
		
		//FIXME: check whether the imported definition is available/visible 
		for (Tree imprt : new TreeIterator(preamble, IMPORT)) {
			String fullName = imprt.getChild(0).getText();
			String name = TSA_SyntaxPatterns.bufferFullName(fullName)[1];
			
			assert(!imports.containsKey(name));
			
			imports.put(name, fullName);
		}
	}
}