package de.lmu.ifi.pms.cep.dura.compiler.durac.formula_transformations;

import de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraAggregationFunctionMapping;
import de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraMathFunctionMapping;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraCExceptionWrapper;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraNoTypeException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraUnexpectedTypeException;
import de.lmu.ifi.pms.cep.tsa.conditions.AggregationFunction;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.conditions.TupleFunction;
import de.lmu.ifi.pms.cep.tsa.schema.Type;
import de.lmu.ifi.pms.java.util.adaption.Adapter;

public class MapTypeFunctions implements Adapter<Term<?>, Term<?>> {

	public static MapTypeFunctions instance = new MapTypeFunctions();
	
	@Override
	public Term<? extends Type> adapt(Term<?> source) {
		if (source instanceof AggregationFunction) {
			AggregationFunction<?> function = (AggregationFunction<?>) source;
			
			Term<? extends Type> argument = this.adapt(function.arg(0));
			
			try {
				return AggregationFunction.apply(function.ID, DuraAggregationFunctionMapping.getType(function, argument), argument);
			} catch (DuraUnexpectedTypeException e) {
				throw new DuraCExceptionWrapper(e);
			}

		} else if (source instanceof TupleFunction) {
			TupleFunction<?> function = (TupleFunction<?>) source;
			
			Term<?>[] typedArguments = new Term[function.arity()];
			
			for (int i=0; i<function.arity(); i++) {
				typedArguments[i] = this.adapt(function.arg(i));
			}
			
			try {
				return TupleFunction.apply(function.ID, DuraMathFunctionMapping.getType(function, typedArguments), typedArguments);
			} catch (DuraUnexpectedTypeException e) {
				throw new DuraCExceptionWrapper(e);
			}
		} else {
			if (! source.isTyped) {
				throw new DuraCExceptionWrapper(new DuraNoTypeException(source));
			}
			
			return (Term<Type>) source;
		}
	}

}
