package de.lmu.ifi.pms.cep.dura.compiler.durac.formula_transformations;

import java.util.HashMap;
import java.util.Map;

import de.lmu.ifi.pms.cep.dura.compiler.durac.collections.ConstDefinitions;
import de.lmu.ifi.pms.cep.dura.compiler.durac.collections.Preamble;
import de.lmu.ifi.pms.cep.dura.compiler.durac.collections.UnambiguousQueryVariables;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraCExceptionWrapper;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraUnspecificCompliationException;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;

/**
 * Transforms terms and formulas with variables, constants, time functions and
 * relations into a more basic representation that is suitable for TSA.
 * 
 * @author hausmann
 * 
 */
public class VariableSubstitution {
	final private MapRelpaceConstants replaceConstants;
	final private MapReplaceVariables replaceVariables;
	final private MapSubstituteAttributePrefix substitutePrefix;
	
	public VariableSubstitution(VariableSubstitution substitution) {
		this.replaceConstants = substitution.replaceConstants;
		this.replaceVariables = substitution.replaceVariables;
		this.substitutePrefix = substitution.substitutePrefix;
	}
	
	public VariableSubstitution(final UnambiguousQueryVariables variables, final ConstDefinitions constants, final Preamble preamble) {
		this(variables, new HashMap<String, String>(), constants, preamble);
	}

	public VariableSubstitution(final UnambiguousQueryVariables variables, Map<String, String> substitutePrefix, final ConstDefinitions constants, final Preamble preamble) {
		this.replaceConstants = new MapRelpaceConstants(constants, preamble);
		this.replaceVariables = new MapReplaceVariables(variables);
		this.substitutePrefix = new MapSubstituteAttributePrefix(substitutePrefix);
	}

	/**
	 * Replaces constants and variables by values and typed attributes.
	 * @throws DuraUnspecificCompliationException 
	 */
	public Term<?> apply(final Term<?> term) throws DuraUnspecificCompliationException {
		try {
			Term<?> result = term;
			
			result = result.replaceAtomicTerm(this.substitutePrefix);
			result = result.replaceAtomicTerm(this.replaceConstants);
			result = result.replaceAtomicTerm(this.replaceVariables);
			
			return result;
		} catch (DuraCExceptionWrapper e) {
			throw e.cause;
		}
	}

	/**
	 * Replaces constants and variables by values and typed attributes.
	 * 
	 * @throws DuraUnspecificCompliationException 
	 */
	public ConditionFormula apply(final ConditionFormula formula) throws DuraUnspecificCompliationException {
		try {
			ConditionFormula result = formula;
	
			result = result.replaceAtomicTerm(this.substitutePrefix);
			result = result.replaceAtomicTerm(this.replaceConstants);
			result = result.replaceAtomicTerm(this.replaceVariables);
			
			return result;
		} catch (DuraCExceptionWrapper e) {
			throw e.cause;
		}
	}
}
