package de.lmu.ifi.pms.cep.event_mill.configuration;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.NoSuchElementException;

import de.lmu.ifi.pms.cep.Version;

import de.lmu.ifi.pms.java.lang.iteration.BufferingIteratorWrapper;

public class ExtendibleURLClassLoader extends URLClassLoader{
	
	static{
		if(Version.Java.greaterOrEqual1dot7){
//			registerAsParallelCapable();
		}
	}
	
	protected final ClassLoader parent;
	
	public final DriverReader driverReader;
	
	public static URL[] extendURLs(final URL[] urls, final ClassLoader parent){
		URL url = resourceURL(DriverReaderImpl.class, parent);
		if(url != null){
			URL[] extendedUrls = Arrays.copyOf(urls, urls.length + 1);
			extendedUrls[urls.length] = url;
			return extendedUrls;
		}
		else{
			return urls;
		}		
	}
	
	public static URL resourceURL(final Class<?> clazz, final ClassLoader classLoader){
		return resourceURL(clazz.getName(), classLoader);
	}
	
	public static URL resourceURL(final String className, final ClassLoader classLoader){
		URL url = classLoader.getResource(className.replace('.', '/')+".class");
		if(url != null){
			String protocol = url.getProtocol();
			if(protocol.equals("file")){
				File directory;
				try {
					directory = new File(url.toURI());
				} catch (URISyntaxException e) {
					throw new Error(e);
				}
				for(int nameLength = className.split("\\.").length; nameLength > 0; --nameLength){
					directory = directory.getParentFile();
				}		
				try {
					return new File(directory.getAbsolutePath() + "/").toURI().toURL();
				} catch (MalformedURLException e) {}
			}
			else if(protocol.equals("jar")){				
				String file = url.getFile();
				if(file.startsWith("file:/") & file.contains(".jar!")){
					try {
						URL jarURL = new URL(file.substring(0, file.indexOf('!')));
						return jarURL;
					} catch (MalformedURLException e1) {
					}
				}
			}
		}
		return null;
	}

	public ExtendibleURLClassLoader(final URL[] urls, final ClassLoader parent) {
		super(extendURLs(urls, parent), null);
		this.parent = parent;
		try {
			this.driverReader = (DriverReader) this.loadClass(DriverReaderImpl.class.getName()).newInstance();
			assert this.driverReader.getClass().getClassLoader() == this;
		} catch (Exception e) {
			throw new Error(e);
		}
	}
	
    @Override
	protected Class<?> findClass(final String name) throws ClassNotFoundException
	{
		if(name.equals(DriverReaderImpl.class.getName())) {
			return super.findClass(name);
		}
		else{
			if(this.parent != null){
				try{
					return this.parent.loadClass(name);
				}
				catch (ClassNotFoundException e){}
			}
			return super.findClass(name);
		}
	}

	public ExtendibleURLClassLoader(final URL[] urls) {
		this(urls, null);
	}

	@Override
	public void addURL(final URL url) {
		super.addURL(url);
	}
	
	@Override
	public URL getResource(final String name) {
		URL resource = null;
		if(this.parent != null){
			resource = this.parent.getResource(name);
		}
		if(resource == null){
			resource = super.getResource(name);
		}
		return resource;
	}
	
	@Override
	public InputStream getResourceAsStream(final String name) {
		InputStream resource = null;
		if(this.parent != null){
			resource = this.parent.getResourceAsStream(name);
		}
		if(resource == null){
			resource = super.getResourceAsStream(name);
		}
		return resource;
	}
	
	@SuppressWarnings("unchecked")
	@Override
	public Enumeration<URL> getResources(final String name) throws IOException {
		if(this.parent != null){
			return new CompoundEnumeration<URL>(this.parent.getResources(name), super.getResources(name));
		}
		else{
			return super.getResources(name);
		}
	}
	
	public static class CompoundEnumeration<E> implements Enumeration<E>{
		
		BufferingIteratorWrapper<Enumeration<E>> enumerations;
		
		public CompoundEnumeration(final Enumeration<E>... enumerations) {
			this.enumerations = new BufferingIteratorWrapper<Enumeration<E>>(Arrays.asList(enumerations).iterator());
			if(this.enumerations.hasNext()){
				this.enumerations.next();
			}
		}

		@Override
		public boolean hasMoreElements() {
			if(this.enumerations.hasCurrent()){
				while(this.enumerations.current().hasMoreElements() | this.enumerations.hasNext()){
					if(this.enumerations.current().hasMoreElements()){
						return true;
					}
					else{
						assert this.enumerations.hasNext();
						this.enumerations.next();
					}
				}
			}
			return false;
		}

		@Override
		public E nextElement() {
			if(this.hasMoreElements()){
				return this.enumerations.current().nextElement();
			}
			else{
				throw new NoSuchElementException();
			}
		}
		
	}
	
}
