package de.lmu.ifi.pms.cep.event_mill.console;

import java.io.BufferedReader;
import java.io.PrintStream;
import java.util.Iterator;

import org.slf4j.Logger;
import org.slf4j.Marker;

import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.map.Map;
import de.lmu.ifi.pms.java.util.tupel.Single;


public abstract class AbstractMarkedConsoleThread extends AbstractConsoleThread {
	
	protected final Marker marker;
	
	public AbstractMarkedConsoleThread(final String name, final Map<String, String, ?> variables, final Iterator<String> immediateCommands, final BufferedReader inputStream, final PrintStream outputStream, final Marker marker) {
		this(name, variables, new Single<Iterator<String>, Iterator<String>>(immediateCommands), inputStream, outputStream, marker);
	}
	
	public AbstractMarkedConsoleThread(final String name, final Map<String, String, ?> variables, final Collection<Iterator<String>> immediateCommands, final BufferedReader inputStream, final PrintStream outputStream, final Marker marker) {
		super(name, variables, immediateCommands, inputStream, outputStream);
		this.marker = marker;
	}

	@Override
	public void run() {

	}
	
	//logging	
	
	@Override
	protected abstract Logger logger();

	@Override
	protected void trace(final String msg) {
		this.logger().trace(this.marker, msg);
	}

	@Override
	protected void trace(final String format, final Object arg) {
		this.logger().trace(this.marker, format, arg);
	}

	@Override
	protected void trace(final String format, final Object arg1, final Object arg2) {
		this.logger().trace(this.marker, format, arg1, arg2);
	}

	@Override
	protected void trace(final String format, final Object... argArray) {
		this.logger().trace(this.marker, format, argArray);
	}

	@Override
	protected void trace(final String msg, final Throwable t) {
		this.logger().trace(this.marker, msg, t);
	}

	@Override
	protected void debug(final String msg) {
		this.logger().debug(this.marker, msg);
	}

	@Override
	protected void debug(final String format, final Object arg) {
		this.logger().debug(this.marker, format, arg);
	}

	@Override
	protected void debug(final String format, final Object arg1, final Object arg2) {
		this.logger().debug(this.marker, format, arg1, arg2);
	}

	@Override
	protected void debug(final String format, final Object... argArray) {
		this.logger().debug(this.marker, format, argArray);
	}

	@Override
	protected void debug(final String msg, final Throwable t) {
		this.logger().debug(this.marker, msg, t);
	}

	@Override
	protected void info(final String msg) {
		this.logger().info(this.marker, msg);
	}

	@Override
	protected void info(final String format, final Object arg) {
		this.logger().info(this.marker, format, arg);
	}

	@Override
	protected void info(final String format, final Object arg1, final Object arg2) {
		this.logger().info(this.marker, format, arg1, arg2);
	}

	@Override
	protected void info(final String format, final Object... argArray) {
		this.logger().info(this.marker, format, argArray);
	}

	@Override
	protected void info(final String msg, final Throwable t) {
		this.logger().info(this.marker, msg, t);
	}

	@Override
	protected void warn(final String msg) {
		this.logger().warn(this.marker, msg);
	}

	@Override
	protected void warn(final String format, final Object arg) {
		this.logger().warn(this.marker, format, arg);
	}

	@Override
	protected void warn(final String format, final Object arg1, final Object arg2) {
		this.logger().warn(this.marker, format, arg1, arg2);
	}

	@Override
	protected void warn(final String format, final Object... argArray) {
		this.logger().warn(this.marker, format, argArray);
	}

	@Override
	protected void warn(final String msg, final Throwable t) {
		this.logger().warn(this.marker, msg, t);
	}

	@Override
	protected void error(final String msg) {
		this.logger().error(this.marker, msg);
	}

	@Override
	protected void error(final String format, final Object arg) {
		this.logger().error(this.marker, format, arg);
	}

	@Override
	protected void error(final String format, final Object arg1, final Object arg2) {
		this.logger().error(this.marker, format, arg1, arg2);
	}

	@Override
	protected void error(final String format, final Object... argArray) {
		this.logger().error(this.marker, format, argArray);
	}

	@Override
	protected void error(final String msg, final Throwable t) {
		this.logger().error(this.marker, msg, t);
	}

}
