package de.lmu.ifi.pms.cep.event_mill.execution.dummy;

import java.util.Arrays;
import java.util.NoSuchElementException;
import java.util.concurrent.locks.ReentrantLock;

import de.lmu.ifi.pms.cep.event_mill.execution.program.ExecutableQuery;
import de.lmu.ifi.pms.cep.event_mill.sql.SQL_Translation;
import de.lmu.ifi.pms.cep.tsa.DTSA_BufferType;
import de.lmu.ifi.pms.cep.tsa.DTSA_QueryType;

import de.lmu.ifi.pms.java.math.Math;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;

public class DummyExecutableQuery extends ExecutableQuery<DummyExecutableProgram, DummyStreamBuffer>
{
	
	public final int id;
	
	long lastRound;
	
	//protected final int numberOfReferencedClocks;
	//protected final FixedArrayNavigableMap<String, Clock> referencedClocks;

	
	protected final ReentrantLock executionLock = new ReentrantLock();

	//protected final Object minWriteTimesLock = new Object();
	
	protected final long[] lastTupleCounts;
	protected final FixedArrayNavigableMap<String, DummyStreamBuffer> inputs;
	protected final long[] distances;
	
	
	protected DummyStreamBuffer output;
	
	protected final boolean isMonotonic;
	
	protected long executionTime;
	
	protected long progress;
	
	public DummyExecutableQuery(final DummyExecutableProgram program, final int id, final DTSA_QueryType type, final String prog_package, final boolean isMonotonic, final DTSA_BufferType input_type, final String[] inputs, final long[] distances, final DTSA_BufferType outputBufferType, final String output, final long executionTime){		
		this(program, id, type, prog_package, isMonotonic, input_type, new FixedArrayNavigableSet<String>(inputs), distances, outputBufferType, output, executionTime);
	}
	
	public DummyExecutableQuery(final DummyExecutableProgram program, final int id, final DTSA_QueryType type, final String prog_package, final boolean isMonotonic, final DTSA_BufferType input_type, final FixedArrayNavigableSet<String> inputs, final long[] distances, final DTSA_BufferType outputBufferType, final String output, final long executionTime){
		this(program, id, type, prog_package, isMonotonic, input_type, inputs, distances, outputBufferType, output, 0, executionTime);
	}

	public DummyExecutableQuery(final DummyExecutableProgram program, final int id, final DTSA_QueryType type, final String prog_package, final boolean isMonotonic, final DTSA_BufferType input_type, final FixedArrayNavigableSet<String> inputs, final long[] distances, final DTSA_BufferType outputBufferType, final String output, final long lastRound, final long executionTime)
	{
		super(type, prog_package, program);
		this.id = id;
		
		this.lastRound = lastRound;
		
		int outputStreamBufferIndex = program.StreamBuffers().get(outputBufferType).keys().indexOf(output);
		if(outputStreamBufferIndex >= 0){
			this.output = program.StreamBuffers().get(outputBufferType).values().get(outputStreamBufferIndex);
		}
		else{
			throw new RuntimeException("Query " + this.id + "requires buffer that is not available in the program.");
		}
		
		this.output.RegisterInputQuery(this);

		if(!SQL_Translation.isPossibleInputBuffersType(type, input_type)){
			throw new RuntimeException("The type of query " + this.id + " and its specified input type are incompatible.");
		}
		
		DummyStreamBuffer[] inputBuffers = new DummyStreamBuffer[inputs.size()];
		for(int index = 0; index < inputBuffers.length; ++index){
			String input = inputs.get(index);
			int streamBufferIndex = program.StreamBuffers().get(input_type).keys().indexOf(input);
			if(streamBufferIndex >= 0){
				inputBuffers[index] = program.StreamBuffers().get(input_type).values().get(streamBufferIndex);
			}
			else{
				throw new RuntimeException("Query " + this.id + "requires buffer \"" + input + "\", but \"" + input + "\" is not available in the program.");
			}
		}		
		this.inputs = new FixedArrayNavigableMap<String, DummyStreamBuffer>(inputs, new FixedArray<DummyStreamBuffer>(false, inputBuffers));
		for(DummyStreamBuffer input: this.inputs.values()){
			input.RegisterOutputQuery(this);
		}
		
		this.distances = Arrays.copyOf(distances, distances.length);
		
		this.lastTupleCounts = new long[inputs.size()];
		Arrays.fill(this.lastTupleCounts, Long.MIN_VALUE);//Will lead to at most one unnecessary query execution

		this.isMonotonic = isMonotonic;
		
		this.executionTime = executionTime;
		
		this.program.register(this);
		
		this.init();
	}
	
	public long distance(final int input){
		return this.distances[input];
	}
	
	public long distance(final String input){
		int index = this.inputs.keys().indexOf(input);
		if(index < 0){
			throw new NoSuchElementException();
		}
		else{
			return this.distance(index);
		}
	}
	
	public long getExecutionTime(){
		return this.executionTime;
	}
	
	public synchronized long progress(){
		return this.progress;
	}
	
	public long achievableProgress(){
		long progress = Math.MaxLongValue;
		long localBound;
		for(int input = 0; input < this.distances.length; ++input){
			localBound = this.inputs.values().get(input).progress() - this.distances[input];
			if(localBound < progress){
				progress = localBound;
			}
		}
		return progress;
	}
	
	protected void init(){
		this.executionLock.lock();

		this.executionLock.unlock();	
	}

	@Override
	public FixedArrayNavigableMap<String, DummyStreamBuffer> inputs()
	{
		return this.inputs;
	}

	@Override
	public FixedArrayNavigableMap<String, DummyStreamBuffer> outputs()
	{
		return new FixedArrayNavigableMap<String, DummyStreamBuffer>(this.output.name(), this.output);
	}
	

	@Override
	public void execute()
	{
		this.executionLock.lock();
		try{
			
			long currentRound = this.lastRound + 1;
			
			this.info("QUERY {} (execution {}) - Start", this.id, currentRound);
			
				long achievableProgress = achievableProgress();
	
				boolean mightComputeNewTuples = false;
				for(int inputIndex = 0; inputIndex < this.inputs.size(); ++inputIndex){
					DummyStreamBuffer input = this.inputs.values().get(inputIndex);
					synchronized (input) {
						long inputTupleCount = input.tupleCount(); 
						if(inputTupleCount > this.lastTupleCounts[inputIndex]){//test if input contains any new tuple compared with the last execution of the query
							this.lastTupleCounts[inputIndex] = inputTupleCount;
							mightComputeNewTuples = true;
						}
					}
				}
				
				if(!this.isMonotonic){
					mightComputeNewTuples = true;
					//FIXME
				}
				if(this.type == DTSA_QueryType.INPUT){
					mightComputeNewTuples = true;
					//FIXME					
				}

				int tupleNumber;
				if(mightComputeNewTuples){

					this.debug("QUERY {} (execution {}) - Executing eval statment.", this.id, currentRound);
					
					long currentTime = System.currentTimeMillis();
					long endTime = currentTime + this.executionTime;
					while(currentTime < endTime){
						try {
							Thread.sleep(endTime - currentTime);
						} catch (InterruptedException e) {
	
						}
						currentTime = System.currentTimeMillis();
					}
					
					tupleNumber = 1;//TODO compute based on selectivity + random
					
					this.debug("QUERY {} (execution {}) - Executed eval statment.", this.id, currentRound);
					
				}
				else{
					this.debug("QUERY {} (execution {}) - Execution of eval statement is omited because no new tuples could have been produced.", this.id, currentRound);
					tupleNumber = 0;
				}
				
				synchronized (this) {
					this.progress = achievableProgress;
				}
				
				//access to new outputMinTimes is still protected by execution lock 
				assert this.executionLock.isHeldByCurrentThread();
						
				this.output.notifyWrite(tupleNumber);			
				for(DummyStreamBuffer input: this.inputs.values()){
					input.notifyRead();
				}
				
				//TODO better log message
				this.info("QUERY {} (execution {}) - Finished", this.id, currentRound);			
				this.progLogger.info("QUERY {} (execution {}) - Finished", this.id, currentRound);

		}
		catch(RuntimeException e){
			this.error("QUERY " + this.id, e);
			throw e;
		}
		catch(Error e){
			this.error("QUERY " + this.id, e);
			throw e;
		}
		finally{//FIXME executes finally also when catch was sucessful?
			this.executionLock.unlock();
		}
	}
	


}
