package de.lmu.ifi.pms.cep.event_mill.io.bus;

import java.util.Arrays;

import org.simpleframework.xml.Attribute;
import org.simpleframework.xml.ElementList;
import org.simpleframework.xml.ElementListUnion;


import de.lmu.ifi.pms.java.util.collection.Collection;


public class CompositeMessageBusProvider extends MessageBusProvider {
	
	protected final MessageBusProvider[] messageBusProviders;
	
	@ElementListUnion({
		@ElementList(entry="empty-bus", inline=true, type=EmptyMessageBusProvider.class, required=false, empty=false),
		@ElementList(entry="composite-bus", inline=true, type=CompositeMessageBusProvider.class, required=false, empty=false),
		@ElementList(entry="pipe-bus", inline=true, type=PipeMessageBusProvider.class, required=false, empty=false),
		@ElementList(entry="heap-broadcast-bus", inline=true, type=HeapBroadcastMessageBusProvider.class, required=false, empty=false),
		@ElementList(entry="bus", type=MessageBusProvider.class, inline=true, required=false, empty=false)
	})
	protected java.util.ArrayList<MessageBusProvider> getBuses(){
		if(this.messageBusProviders != null){
			return new java.util.ArrayList<MessageBusProvider>(java.util.Arrays.asList(this.messageBusProviders));
		}
		else{
			return null;
		}
	}
	
	public CompositeMessageBusProvider(final String identifier) {
		super(identifier);
		this.messageBusProviders = new MessageBusProvider[0];
	}
	
	public CompositeMessageBusProvider(final String identifier, final MessageBusProvider[] messageBusProviders) {
		super(identifier);
		this.messageBusProviders = Arrays.copyOf(messageBusProviders, messageBusProviders.length);
	}
	
	public CompositeMessageBusProvider(final String identifier, final Collection<MessageBusProvider> messageBusProviders) {
		this(identifier, messageBusProviders.javaCollection());
	}
	
	public CompositeMessageBusProvider(@Attribute(name="iri") final String identifier, 
			@ElementList(entry="bus", inline=true, type=MessageBusProvider.class, required=false, empty=false)
			final java.util.ArrayList<MessageBusProvider> messageBusProviders) {
		this(identifier, (java.util.Collection<MessageBusProvider>) messageBusProviders);
	}
	
	public CompositeMessageBusProvider(final String identifier, final java.util.Collection<MessageBusProvider> messageBusProviders) {
		super(identifier);
		if(messageBusProviders == null){
			this.messageBusProviders = new MessageBusProvider[0];
		}
		else{
			this.messageBusProviders = messageBusProviders.toArray(new MessageBusProvider[messageBusProviders.size()]);
		}
	}

	@Override
	public boolean accepts(final String identity) {
		for(MessageBusProvider provider: this.messageBusProviders){
			if(provider.accepts(identity)){
				return true;
			}
		}
		return false;
	}

	@Override
	public MessageBus<?> connect(final String identity) {
		for(MessageBusProvider provider: this.messageBusProviders){
			if(provider.accepts(identity)){
				return provider.connect(identity);
			}
		}
		throw new IllegalArgumentException();
	}

}
