package de.lmu.ifi.pms.cep.event_mill.io.bus;

import org.simpleframework.xml.ElementList;
import org.simpleframework.xml.ElementListUnion;

import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;

public class ExplicitMessageBusProviderCollection extends MessageBusProviderCollection{
	
	public final FixedArrayNavigableMap<String, MessageBusProvider> buses;
	
	@ElementListUnion({
		@ElementList(entry="empty-bus", inline=true, type=EmptyMessageBusProvider.class, required=false, empty=false),
		@ElementList(entry="composite-bus", inline=true, type=CompositeMessageBusProvider.class, required=false, empty=false),
		@ElementList(entry="pipe-bus", inline=true, type=PipeMessageBusProvider.class, required=false, empty=false),
		@ElementList(entry="heap-broadcast-bus", inline=true, type=HeapBroadcastMessageBusProvider.class, required=false, empty=false),
		@ElementList(entry="bus", inline=true, type=MessageBusProvider.class, required=false, empty=false)
	})
	protected java.util.ArrayList<MessageBusProvider> getBuses(){
		if(this.buses.size() > 0){
			return new java.util.ArrayList<MessageBusProvider>(this.buses.values().javaCollection());
		}
		else{
			return null;
		}
	}
	
	public ExplicitMessageBusProviderCollection() {
		this(new MessageBusProvider[0]);
	}
	
	public ExplicitMessageBusProviderCollection(final MessageBusProvider[] busProviders) {
		super();		
		String[] busIdentifiers = new String[busProviders.length];
		for(int stream = 0; stream < busProviders.length; ++stream){
			busIdentifiers[stream] = busProviders[stream].identifier; 
		}
		
		this.buses = new FixedArrayNavigableMap<String, MessageBusProvider>(busIdentifiers, busProviders);
	}
	
	public ExplicitMessageBusProviderCollection(final Collection<MessageBusProvider> messageBusProviders) {
		this(messageBusProviders.javaCollection());
	}
	
	public ExplicitMessageBusProviderCollection(
			@ElementList(entry="bus", inline=true, required=false, empty=false)
			final java.util.ArrayList<MessageBusProvider> messageBusProviders) {
		this((java.util.Collection<MessageBusProvider>) messageBusProviders);
	}
	
	public ExplicitMessageBusProviderCollection(final java.util.Collection<MessageBusProvider> messageBusProviders) {
		this((messageBusProviders == null ? new MessageBusProvider[0] : messageBusProviders.toArray(new MessageBusProvider[messageBusProviders.size()])));
	}
	
	@Override
	public boolean contains(final String identifier) {
		return this.buses.has(identifier);
	}
	
	@Override
	public MessageBusProvider create(final String identifier) {
		return this.buses.get(identifier);
	}

}
