package de.lmu.ifi.pms.cep.event_mill.io.bus;

import java.util.Arrays;
import java.util.concurrent.ConcurrentLinkedQueue;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.event_mill.io.MessageSink;
import de.lmu.ifi.pms.cep.event_mill.io.MessageSource;
import de.lmu.ifi.pms.cep.event_mill.io.format.MessageFormatProvider;

public class PipeMessageBus<MessageType> extends MessageBus<MessageType> {
	
	volatile ConcurrentLinkedQueue<MessageType> inputPipe;
	int sourceCount;
	
	volatile ConcurrentLinkedQueue<MessageType> outputPipe;
	
	final MessageFormatProvider<MessageType> formater;
	
	public final PipeMessageBus<MessageType> complement;
	
	public PipeMessageBus(final MessageFormatProvider<MessageType> formater){
		this(null,formater);
	}
	
	public PipeMessageBus(final PipeMessageBusProvider<MessageType> provider) {
		this(provider, provider.formater);
	}
	
	protected PipeMessageBus(final PipeMessageBusProvider<MessageType> provider, final MessageFormatProvider<MessageType> formater) {
		super(formater.type, provider);
		this.formater = formater;
		this.complement = new PipeMessageBus<MessageType>(this);
		
		this.sourceCount = 0;
	}
	
	protected PipeMessageBus(final PipeMessageBus<MessageType> pipe) {
		super(pipe.type, pipe.providerReference);
		this.formater = pipe.formater;
		this.inputPipe = pipe.outputPipe;
		this.outputPipe = pipe.inputPipe;
		this.complement = pipe;
		
		this.sourceCount = 0;
	}

	@Override
	public MessageFormatProvider<MessageType> formatter() {
		return this.formater;
	}
	
	@Override
	public MessageSink<MessageType> sink() {
		return new MessageSink<MessageType>(this.type) {
			@Override
			public void writeBlock(final MessageType[] messages) {
				ConcurrentLinkedQueue<MessageType> outputPipe = PipeMessageBus.this.outputPipe;
				if(outputPipe != null){
					outputPipe.addAll(Arrays.asList(messages));
				}
			}
			
			@Override
			public void writeBlock(final java.util.Collection<MessageType> messages) {
				ConcurrentLinkedQueue<MessageType> outputPipe = PipeMessageBus.this.outputPipe;
				if(outputPipe != null){
					outputPipe.addAll(messages);
				}
			}

			@Override
			public void write(final MessageType message) {
				ConcurrentLinkedQueue<MessageType> outputPipe = PipeMessageBus.this.outputPipe;
				if(outputPipe != null){
					outputPipe.offer(message);
				}
			}
			
			@Override
			public void close() {

			}
		};
	}
	
	@Override
	public synchronized MessageSource<MessageType> source() {
		if(this.sourceCount == 0){
			this.inputPipe = new ConcurrentLinkedQueue<MessageType>();
			this.complement.outputPipe = this.inputPipe;
		}
		++this.sourceCount;
		assert this.inputPipe != null;
		
		return new MessageSource<MessageType>(this.type) {
			
			@Override
			public MessageType read() {
				return PipeMessageBus.this.inputPipe.poll();
			}
			
			@Override
			public MessageType[] readBlock() {
				MessageType message = PipeMessageBus.this.inputPipe.poll();
				if(message == null){
					return this.emptyBlock;
				}
				else{
					java.util.ArrayList<MessageType> block = new java.util.ArrayList<MessageType>();			
					while(message != null){
						block.add(message);
						message = PipeMessageBus.this.inputPipe.poll();
					}
					return block.toArray(this.newBlock(block.size()));
				}
			}
			
			@Override
			public void close() {
				synchronized(PipeMessageBus.this){
					--PipeMessageBus.this.sourceCount;
					if(PipeMessageBus.this.sourceCount == 0){
						PipeMessageBus.this.inputPipe = null;
						PipeMessageBus.this.complement.outputPipe = null;
					}
				}
			}
			
		};
	}
	
	protected static final Logger logger = LoggerFactory.getLogger(PipeMessageBus.class);
	
	@Override
	protected Logger logger() {
		return logger;
	}

}
