package de.lmu.ifi.pms.cep.event_mill.io.bus;

import org.simpleframework.xml.Attribute;
import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementUnion;

import de.lmu.ifi.pms.cep.event_mill.io.catalog.DataCatalogProvider;
import de.lmu.ifi.pms.cep.event_mill.io.format.MessageFormatProvider;

public class WrapperMessageBusProvider<MessageType> extends MessageBusProvider{
	
	@Element(name="catalog")
	protected final DataCatalogProvider dataCatalog;
	
	@Element(name="format")
	protected final MessageFormatProvider<MessageType> formatter;//needs bijective formatter
	
	@ElementUnion({
		@Element(name="empty-bus", type=EmptyMessageBusProvider.class),
		@Element(name="composite-bus", type=CompositeMessageBusProvider.class),
		@Element(name="pipe-bus", type=PipeMessageBusProvider.class),
		@Element(name="heap-broadcast-bus", type=HeapBroadcastMessageBusProvider.class),
		@Element(name="bus", type=MessageBusProvider.class)
	})
	protected final MessageBusProvider bus;
	
	public WrapperMessageBusProvider(@Attribute(name="iri") final String identifier
			, @Element(name="catalog") final DataCatalogProvider dataCatalog
			, @Element(name="format") final MessageFormatProvider<MessageType> formatter
			, @Element(name="bus") final MessageBusProvider bus
		) {
		super(identifier);
		this.dataCatalog = dataCatalog;
		this.formatter = formatter;
		this.bus = bus;
	}

	@Override
	public boolean accepts(final String identifier) {
		return this.bus.accepts(identifier);
	}
	
	@SuppressWarnings({ "rawtypes", "unchecked" })
	@Override
	public MessageBus<?> connect(final String identifier) {
		MessageBus<?> bus = this.bus.connect(identifier);
		return new WrapperMessageBus(this, bus);
	}

}
