package de.lmu.ifi.pms.cep.event_mill.io.connect;

import java.util.Arrays;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.event_mill.io.MessageAdapter;
import de.lmu.ifi.pms.cep.event_mill.io.MessageSink;
import de.lmu.ifi.pms.cep.event_mill.io.MessageSource;
import de.lmu.ifi.pms.cep.event_mill.io.bus.MessageBus;
import de.lmu.ifi.pms.cep.event_mill.io.catalog.DataCatalog;

public class BusConnector<MessageType1, MessageType2> extends Connector {
	
	protected final MessageSource<MessageType1> source;
	protected final MessageSink<MessageType2> sink;
	protected final MessageAdapter<MessageType1, MessageType2> adapter;
	
	public BusConnector(final DataCatalog dataCatalog, final MessageBus<MessageType1> bus1, final MessageBus<MessageType2> bus2){
		super();
		this.source = bus1.source();
		this.sink = bus2.sink();
		if(this.sink.type.isAssignableFrom(this.source.type) & bus1.formatter().identifier.equals(bus2.formatter().identifier)){
			this.adapter = null;
		}
		else{
			this.adapter = new MessageAdapter<MessageType1, MessageType2>(dataCatalog, bus1.formatter(), bus2.formatter());
		}
	}

	@Override
	public void nextBlock() {
		MessageType1[] block1 = this.source.readBlock();
		if(block1.length > 0){
			if(this.adapter != null){
				MessageType2[] block2 = this.sink.newBlock(block1.length);
				MessageType2 message2;
				int index2 = 0;
				for(int index1 = 0; index1 < block1.length; ++index1){
					message2 = this.adapter.adapt(block1[index1]);
					if(message2 != null){
						block2[index2] = message2;
						++index2;
					}
				}
				if(index2 < block2.length){
					this.sink.writeBlock(Arrays.copyOf(block2, index2));
				}
				else{
					this.sink.writeBlock(block2);
				}
			}
			else{
				@SuppressWarnings("unchecked")
				MessageType2[] block2 = (MessageType2[])block1;
				this.sink.writeBlock(block2);
			}
		}
	}

	protected static final Logger logger = LoggerFactory.getLogger(BusConnector.class);
	
	@Override
	protected Logger logger(){
		return logger;
	}
	
}
