package de.lmu.ifi.pms.cep.event_mill.io.format;

import org.simpleframework.xml.ElementList;
import org.simpleframework.xml.ElementListUnion;

import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;

public class ExplicitMessageFormatProviderCollection extends MessageFormatProviderCollection{
	
	public final FixedArrayNavigableMap<String, MessageFormatProvider<?>> formats;
	
	@ElementListUnion({
		@ElementList(entry="alias", inline=true, type=MessageFormatAliasProvider.class, required=false, empty=false),
		@ElementList(entry="format", inline=true, type=MessageFormatProvider.class, required=false, empty=false)
	})
	protected java.util.ArrayList<MessageFormatProvider<?>> getFormats(){
		if(this.formats.size() > 0){
			java.util.ArrayList<MessageFormatProvider<?>> basicFormats = new java.util.ArrayList<MessageFormatProvider<?>>();
			java.util.ArrayList<MessageFormatProvider<?>> aliases = new java.util.ArrayList<MessageFormatProvider<?>>();
			for(MessageFormatProvider<?> format: this.formats.values().javaCollection()){
				if(format instanceof MessageFormatAliasProvider<?>){
					aliases.add(format);
				}
				else{
					basicFormats.add(format);
				}
			}
			basicFormats.addAll(aliases);
			return new java.util.ArrayList<MessageFormatProvider<?>>(basicFormats);
		}
		else{
			return null;
		}
	}
	
	public ExplicitMessageFormatProviderCollection() {
		this(new MessageFormatProvider<?>[0]);
	}
	
	public ExplicitMessageFormatProviderCollection(final MessageFormatProvider<?>[] formatProviders) {
		super();		
		String[] formatIdentifiers = new String[formatProviders.length];
		for(int stream = 0; stream < formatProviders.length; ++stream){
			formatIdentifiers[stream] = formatProviders[stream].identifier; 
		}
		
		this.formats = new FixedArrayNavigableMap<String, MessageFormatProvider<?>>(formatIdentifiers, formatProviders);
	}
	
	public ExplicitMessageFormatProviderCollection(final Collection<MessageFormatProvider<?>> messageBusProviders) {
		this(messageBusProviders.javaCollection());
	}
	
	public ExplicitMessageFormatProviderCollection(
			@ElementList(entry="format", inline=true, required=false, empty=false)
			final java.util.ArrayList<MessageFormatProvider<?>> messageBusProviders) {
		this((java.util.Collection<MessageFormatProvider<?>>) messageBusProviders);
	}
	
	public ExplicitMessageFormatProviderCollection(final java.util.Collection<MessageFormatProvider<?>> messageFormatProviders) {
		this((messageFormatProviders == null ? new MessageFormatProvider<?>[0] : messageFormatProviders.toArray(new MessageFormatProvider<?>[messageFormatProviders.size()])));
	}
	
	@Override
	public boolean contains(final String identifier) {
		return this.formats.contains(identifier);
	}
	
	@Override
	public MessageFormatProvider<?> get(final String identifier) {
		return this.formats.get(identifier);
	}

}
