package de.lmu.ifi.pms.cep.event_mill.io.format;

import java.util.Arrays;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;


public class JavaRegexMessageReader extends MessageReader<String>{
	
	public final String messageRegEx;
	Matcher matcher;
	final int[]capturingGroups;
	
	public JavaRegexMessageReader(final String messageRegEx, final String[] parameters, final BasicTypeID[] types, final int[]capturingGroups) {
		this(messageRegEx, new FixedArray<String>(parameters), new FixedArray<BasicTypeID>(types), capturingGroups);
	}

	public JavaRegexMessageReader(final String messageRegEx, final FixedArray<String> parameters, final FixedArray<BasicTypeID> types, final int[]capturingGroups) {
		super(String.class, parameters, types);
		this.messageRegEx = messageRegEx;
		this.matcher = Pattern.compile(messageRegEx).matcher("");
		int groupCount = this.matcher.groupCount();
		if(capturingGroups.length != parameters.size()){
			throw new IllegalArgumentException();
		}
		for(int parameter = 0; parameter < capturingGroups.length; ++parameter){
			if(capturingGroups[parameter] > groupCount){
				throw new IllegalArgumentException();
			}
		}
		
		this.capturingGroups = Arrays.copyOf(capturingGroups, capturingGroups.length);
		
	}

	@Override
	public void decompose(final String message) {
		this.matcher.reset(message);
		if(this.matcher.matches()){
			
		}
		else{
			throw new IllegalArgumentException("Message does not match the specified format.");
		}
	}

	@Override
	public void clearMessage() {
		this.matcher.reset("");
	}

	@Override
	public boolean isNull(final int parameterIndex) {
		return this.matcher.group(this.capturingGroups[parameterIndex - 1]) == null;
	}

	@Override
	public boolean getBoolean(final int parameterIndex) {
		return Boolean.valueOf(this.matcher.group(this.capturingGroups[parameterIndex - 1]));
	}

	@Override
	public byte getByte(final int parameterIndex) {
		return Byte.valueOf(this.matcher.group(this.capturingGroups[parameterIndex - 1]));
	}

	@Override
	public short getShort(final int parameterIndex) {
		return Short.valueOf(this.matcher.group(this.capturingGroups[parameterIndex - 1]));
	}

	@Override
	public int getInt(final int parameterIndex) {
		return Integer.valueOf(this.matcher.group(this.capturingGroups[parameterIndex - 1]));
	}

	@Override
	public long getLong(final int parameterIndex) {
		return Long.valueOf(this.matcher.group(this.capturingGroups[parameterIndex - 1]));
	}

	@Override
	public long getTimestamp(final int parameterIndex) {
		return Long.valueOf(this.matcher.group(this.capturingGroups[parameterIndex - 1]));
	}

	@Override
	public long getDuration(final int parameterIndex) {
		return Long.valueOf(this.matcher.group(this.capturingGroups[parameterIndex - 1]));
	}

	@Override
	public long getIdentifier(final int parameterIndex) {
		return Long.valueOf(this.matcher.group(this.capturingGroups[parameterIndex - 1]));
	}

	@Override
	public float getFloat(final int parameterIndex) {
		return Float.valueOf(this.matcher.group(this.capturingGroups[parameterIndex - 1]));
	}

	@Override
	public double getDouble(final int parameterIndex) {
		return Double.valueOf(this.matcher.group(this.capturingGroups[parameterIndex - 1]));
	}

	@Override
	public String getString(final int parameterIndex) {
		return this.matcher.group(this.capturingGroups[parameterIndex - 1]);
	}
	
	protected static final Logger logger = LoggerFactory.getLogger(JavaRegexMessageReader.class);
	
	@Override
	protected Logger logger(){
		return logger;
	}

}
