package de.lmu.ifi.pms.cep.event_mill.io.format;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.simpleframework.xml.Element;

import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;

public class ParameterFormatterImpl extends ParameterFormatter{
	
	public static final String defaultBooleanRegEx = "true|false";//TODO case insensitive
	public static final String defaultByteRegEx = "[+-]?+\\d++";
	public static final String defaultShortRegEx = "[+-]?+\\d++";
	public static final String defaultIntRegEx = "[+-]?+\\d++";
	public static final String defaultLongRegEx = "[+-]?+\\d++";
	public static final String defaultTimestampRegEx = "[+-]?+\\d++";
	public static final String defaultDurationRegEx = "[+-]?+\\d++";
	public static final String defaultIdentifierRegEx = "[+-]?+\\d++";
	public static final String defaultFloatRegEx = "[+-]?+\\d++(?:\\.\\d++)?+";//FIXME
	public static final String defaultDoubleRegEx = "[+-]?+\\d++(?:\\.\\d++)?+";//FIXME
	public static final String defaultStringRegEx = "\"(?:[^\"\\\\]|\\\\\\\\|\\\\\")*+\"";
	
	@Element(name="boolean-pattern", required=false)
	public final Pattern booleanPattern;
	@Element(name="byte-pattern", required=false)
	public final Pattern bytePattern;
	@Element(name="short-pattern", required=false)
	public final Pattern shortPattern;
	@Element(name="int-pattern", required=false)
	public final Pattern intPattern;
	@Element(name="long-pattern", required=false)
	public final Pattern longPattern;
	@Element(name="timestamp-pattern", required=false)
	public final Pattern timestampPattern;
	@Element(name="duration-pattern", required=false)
	public final Pattern durationPattern;
	@Element(name="identifier-pattern", required=false)
	public final Pattern identifierPattern;
	@Element(name="float-pattern", required=false)
	public final Pattern floatPattern;
	@Element(name="double-pattern", required=false)
	public final Pattern doublePattern;
	@Element(name="string-pattern", required=false)
	public final Pattern stringPattern;
	
	public ParameterFormatterImpl() {
		this(defaultBooleanRegEx, defaultByteRegEx, defaultShortRegEx, defaultIntRegEx, defaultLongRegEx, defaultTimestampRegEx, defaultDurationRegEx, defaultIdentifierRegEx, defaultFloatRegEx, defaultDoubleRegEx, defaultStringRegEx);
	}	
	
	public ParameterFormatterImpl(@Element(name="boolean-pattern", required=false) final String booleanRegEx, 
			@Element(name="byte-pattern", required=false) final String byteRegEx,
			@Element(name="short-pattern", required=false) final String shortRegEx, 
			@Element(name="int-pattern", required=false) final String intRegEx, 
			@Element(name="long-pattern", required=false) final String longRegEx,
			@Element(name="timestamp-pattern", required=false) final String timestampRegEx, 
			@Element(name="duration-pattern", required=false) final String durationRegEx,
			@Element(name="identifier-pattern", required=false) final String identifierRegEx, 
			@Element(name="float-pattern", required=false) final String floatRegEx,
			@Element(name="double-pattern", required=false) final String doubleRegEx, 
			@Element(name="string-pattern", required=false) final String stringRegEx) {
		super();
		this.booleanPattern = (booleanRegEx == null ? Pattern.compile(defaultBooleanRegEx) : Pattern.compile(booleanRegEx));
		this.bytePattern = (byteRegEx == null ? Pattern.compile(defaultByteRegEx) : Pattern.compile(byteRegEx));
		this.shortPattern = (shortRegEx == null ? Pattern.compile(defaultShortRegEx) : Pattern.compile(shortRegEx));
		this.intPattern = (intRegEx == null ? Pattern.compile(defaultIntRegEx) : Pattern.compile(intRegEx));
		this.longPattern = (longRegEx == null ? Pattern.compile(defaultLongRegEx) : Pattern.compile(longRegEx));
		this.timestampPattern = (timestampRegEx == null ? Pattern.compile(defaultTimestampRegEx) : Pattern.compile(timestampRegEx));
		this.durationPattern = (durationRegEx == null ? Pattern.compile(defaultDurationRegEx) : Pattern.compile(durationRegEx));
		this.identifierPattern = (identifierRegEx == null ? Pattern.compile(defaultIdentifierRegEx) : Pattern.compile(identifierRegEx));
		this.floatPattern = (floatRegEx == null ? Pattern.compile(defaultFloatRegEx) : Pattern.compile(floatRegEx));
		this.doublePattern = (doubleRegEx == null ? Pattern.compile(defaultDoubleRegEx) : Pattern.compile(doubleRegEx));
		this.stringPattern = (stringRegEx == null ? Pattern.compile(defaultStringRegEx) : Pattern.compile(stringRegEx));
	}



	@Override
	public String matchParameter(final String message, final int start, int maxEnd, final BasicTypeID type){
		if(maxEnd > message.length()){
			maxEnd = message.length();
		}
		Matcher matcher;
		switch (type) {
		case BOOLEAN:{
			matcher = this.booleanPattern.matcher(message);
			break;
		}
		case DOUBLE:{
			matcher = this.doublePattern.matcher(message);
			break;
		}
		case DURATION:{
			matcher = this.durationPattern.matcher(message);
			break;
		}
		case FLOAT:{
			matcher = this.floatPattern.matcher(message);
			break;
		}
		case IDENTIFIER:{
			matcher = this.identifierPattern.matcher(message);
			break;
		}
		case INT:{
			matcher = this.intPattern.matcher(message);
			break;
		}
		case LONG:{
			matcher = this.longPattern.matcher(message);
			break;
		}
		case REFERENCE:{
			throw new IllegalArgumentException("Type \"reference\" not supported.");
		}
		case STRING:{
			matcher = this.stringPattern.matcher(message);
			break;
		}
		case TIMESTAMP:{
			matcher = this.timestampPattern.matcher(message);
			break;
		}
		default:
			throw new Error("Unimplemented case: " + type.name());
		}
		matcher.region(start, maxEnd);
		if(matcher.lookingAt()){
			return matcher.group();
		}
		else{
			throw new IllegalArgumentException("Expecting parameter of type " + type + " at position " + start);
		}
	}

	@Override
	public boolean parseBoolean(final String value){
		return Boolean.parseBoolean(value);
	}

	@Override
	public byte parseByte(final String value) {
		return Byte.parseByte(value);
	}

	@Override
	public short parseShort(final String value) {
		return Short.parseShort(value);
	}

	@Override
	public int parseInt(final String value) {
		return Integer.parseInt(value);
	}

	@Override
	public long parseLong(final String value) {
		return Long.parseLong(value);
	}

	@Override
	public long parseTimestamp(final String value) {
		return Long.parseLong(value);
	}

	@Override
	public long parseDuration(final String value) {
		return Long.parseLong(value);
	}

	@Override
	public long parseIdentifier(final String value) {
		return Long.parseLong(value);
	}

	@Override
	public float parseFloat(final String value) {
		return Float.parseFloat(value);
	}

	@Override
	public double parseDouble(final String value) {
		return Double.parseDouble(value);
	}

	@Override
	public String parseString(final String value) {
		assert value.length() >= 2;
		return value.substring(1, value.length() - 1);
	}

	@Override
	public String formatBoolean(final boolean value) {
		return String.valueOf(value);
	}

	@Override
	public String formatByte(final byte value) {
		return String.valueOf(value);
	}

	@Override
	public String formatShort(final short value) {
		return String.valueOf(value);
	}

	@Override
	public String formatInt(final int value) {
		return String.valueOf(value);
	}


	@Override
	public String formatLong(final long value) {
		return String.valueOf(value);
	}

	@Override
	public String formatTimestamp(final long value) {
		return String.valueOf(value);
	}

	@Override
	public String formatDuration(final long value) {
		return String.valueOf(value);
	}

	@Override
	public String formatIdentifier(final long value) {
		return String.valueOf(value);
	}

	@Override
	public String formatFloat(final float value) {
		return String.valueOf(value);
	}

	@Override
	public String formatDouble(final double value) {
		return String.valueOf(value);
	}

	@Override
	public String formatString(final String value) {
		return "\"" + value + "\"";
	}
	
}
