package de.lmu.ifi.pms.cep.event_mill.io.format;

import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;

import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;

public abstract class StringMessageFormatProvider extends MessageFormatProvider<String> {
	
	protected final ParameterFormatter formatter;
	
	public StringMessageFormatProvider(final String identifier,	final ParameterFormatter formatter) {
		super(String.class, identifier);
		this.formatter = formatter;
		assert this.formatter != null;
	}

	@Override
	public MessageGenerator<String> messageGenerator(final String messageType, final FixedArrayNavigableMap<String, BasicTypeID> types) {
		return new StringPatternGenerator(this.formatter, this.getGeneratorPattern(messageType, types), types);
	}

	@Override
	public MessageReader<String> messageReader(final String messageType, final FixedArrayNavigableMap<String, BasicTypeID> types) {	
		return new StringPatternReader(this.formatter, this.getGeneratorPattern(messageType, types), types);
	}
	
	public static FixedArrayNavigableMap<String, BasicType> basicTypes(final FixedArrayNavigableMap<String, BasicTypeID> types){
		int numberOfAttributes = types.size();
		FixedArray<BasicTypeID> basicTypeIDs = types.values();
		BasicType[] basicTypes = new BasicType[numberOfAttributes];
		for(int attribute = 0; attribute < numberOfAttributes; ++attribute){
			basicTypes[attribute] = basicTypeIDs.get(attribute).type;
		}
		return new FixedArrayNavigableMap<String, BasicType>(types.keys(), new FixedArray<BasicType>(false, basicTypes));
	}
	
	protected abstract String getGeneratorPattern(final String messageType, final FixedArrayNavigableMap<String, BasicTypeID> types);
	
	@Override
	public abstract String messageType(final String message);
	
}
