package de.lmu.ifi.pms.cep.event_mill.io.format.impl.xml;

import static de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID.*;

import org.simpleframework.xml.Attribute;
import org.simpleframework.xml.Element;

import de.lmu.ifi.pms.cep.event_mill.io.format.MessageFormatProvider;
import de.lmu.ifi.pms.cep.event_mill.io.format.MessageGenerator;
import de.lmu.ifi.pms.cep.event_mill.io.format.MessageReader;
import de.lmu.ifi.pms.cep.event_mill.io.format.ParameterFormatter;
import de.lmu.ifi.pms.cep.event_mill.io.format.StructuredStringMessageFormatProvider;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;

import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;

public class XmlMessageFormatProvider extends StructuredStringMessageFormatProvider {
	
	public static final String defaultIdentifier = "de.lmu.ifi.pms.cep.io.format.xml";
	public static final ParameterFormatter defaultFormatter = new ParameterFormatter(){
		
		@Override
		public String matchParameter(final String message, final int start, final int maxEnd, final BasicTypeID type) {
			return message.substring(start, message.indexOf('<', start)).trim();
		}
		
		@Override
		public String formatString(final String value) {
			//FIXME xml escape and resolve strings
			return value.trim();
		}
		
		@Override
		protected ParameterFormatter resolve() {
			return defaultFormatter;
		};
		
	};
	
	@Element(name="parameter-formatter", required=false)
	ParameterFormatter getFormatter(){
		if(this.formatter == defaultFormatter){
			return null;
		}
		else{
			return this.formatter;
		}
	}
	
	public XmlMessageFormatProvider() {
		this(defaultIdentifier);
	}
	
	public XmlMessageFormatProvider(final String identifier) {
		this(identifier, defaultFormatter);
	}
	
	public XmlMessageFormatProvider(final ParameterFormatter formatter) {
		this(defaultIdentifier, formatter);
	}
	
	public XmlMessageFormatProvider(@Attribute(name="iri", required=false) final String identifier, @Element(name="parameter-formatter", required=false) final ParameterFormatter formatter) {
		super(identifier, (formatter == null ? defaultFormatter : formatter));
	}

	@Override
	public String messageType(final String message) {
		final int firstTagStart = message.indexOf('<');
		final int firstTagEnd = message.indexOf('>', firstTagStart);
		return message.substring(firstTagStart + 1, firstTagEnd); 
	}
	
	@Override
	protected void getMessageHead(final String messageType, final StringBuilder sb, final boolean hasAttributes) {
		if(hasAttributes){
			sb.append("<");
			sb.append(messageType);
			sb.append(">\n");
		}
		else{
			sb.append("<");
			sb.append(messageType);
		}
	}
	
	@Override
	protected void getMessageFoot(final String messageType, final StringBuilder sb, final boolean hasAttributes) {
		if(hasAttributes){
			sb.append("\n</");
			sb.append(messageType);
			sb.append(">");
		}
		else{
			sb.append("/>");
		}
	}
	
	@Override
	protected void getAttributePrefix(final String attribute_local_name, final StringBuilder sb, final boolean hasSubAttributes, final boolean hasValue, final int depth) {
		indentTab(sb, depth);
		if(hasSubAttributes | hasValue){
			sb.append("<");
			sb.append(attribute_local_name);
			sb.append(">");
			if(hasSubAttributes){
				sb.append("\n");;
			}
		}
		else{
			sb.append("<");
			sb.append(attribute_local_name);
		}
	}
	
	@Override
	protected String getAttributeDelimiter(final int depth) {
		return "\n";
	}
	
	@Override
	protected void getAttributePostfix(final String attribute_local_name, final StringBuilder sb, final boolean hasSubAttributes, final boolean hasValue, final int depth) {
		if(hasSubAttributes | hasValue){
			if(hasSubAttributes){
				sb.append("\n");
				indentTab(sb, depth);
			}
			sb.append("</");
			sb.append(attribute_local_name);
			sb.append(">");
		}
		else{
			sb.append("/>");
		}
	}

	public static void main(final String[] args) {
		String messageType = "identity";
		FixedArrayNavigableMap<String, BasicTypeID> types = new FixedArrayNavigableMap<String, BasicTypeID>(
				new String[]{		"person.name", "person.address.street", "person.address.number"}, 
				new BasicTypeID[]{	STRING, STRING, INT}
				);
		
		MessageFormatProvider<String> formatProvider = new XmlMessageFormatProvider();
		
		MessageGenerator<String> messageGenerator = formatProvider.messageGenerator(messageType, types);
		
		messageGenerator.setString("person.name", "Hans Meier");
		messageGenerator.setString("person.address.street", "Münchner Straße");
		messageGenerator.setInt("person.address.number", 15);
		
		String message = messageGenerator.compose();
		System.out.println(message);
		messageGenerator.clearParameters();
		
		MessageReader<String> messageReader = formatProvider.messageReader(messageType, types);
		messageReader.decompose(message);
		
		messageGenerator.setString("person.name", messageReader.getString("person.name"));
		messageGenerator.setString("person.address.street", messageReader.getString("person.address.street"));
		messageGenerator.setInt("person.address.number", messageReader.getInt("person.address.number"));
		
		message = messageGenerator.compose();
		System.out.println(message);
	}
	
}
