package de.lmu.ifi.pms.cep.event_mill.io.stream;

import org.slf4j.Logger;

import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;

import de.lmu.ifi.pms.java.LoggingObject;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;


public abstract class Stream extends LoggingObject{
	
	protected boolean isOpen;
	public final String name;
	public final FixedArray<String> attributes;
	public final FixedArrayNavigableMap<String, BasicTypeID> types;
	final int[] positions;
	
	public Stream(final String name, final FixedArrayNavigableMap<String, BasicTypeID> types){
		this.name = name;
		this.attributes = types.keys();
		this.types = types;
		this.positions = new int[this.types.size()];
		for(int attribute = 0; attribute < this.positions.length; ++attribute){
			this.positions[attribute] = attribute + 1;
		}
		
		this.isOpen = true;
	}
	
	public Stream(final String name, final FixedArray<String> attributes, final FixedArray<BasicTypeID> types) {
		super();
		if(attributes.size() != types.size()){
			throw new IllegalArgumentException("Number of parameters and number of specified types do not match.");
		}
		this.name = name;
		this.attributes = attributes;
		String[] sortedParameters = attributes.toArray(new String[attributes.size()]);
		BasicTypeID[] sortedParameterTypes = new BasicTypeID[sortedParameters.length];
		this.positions = new int[attributes.size()];
		{
			int[] parameterNewPositions = new int[attributes.size()];
			int sortedSize = de.lmu.ifi.pms.java.util.collection.array.Arrays.sort(sortedParameters, parameterNewPositions, false);
			if(sortedSize != attributes.size()){
				throw new IllegalArgumentException("Dublicate parameter!");
			}
			for(int parameter = 0; parameter < this.positions.length; ++parameter){
				sortedParameterTypes[parameterNewPositions[parameter]] = types.get(parameter);
				this.positions[parameterNewPositions[parameter]] = parameter + 1;
			}
		}
		
		this.types = new FixedArrayNavigableMap<String, BasicTypeID>(new FixedArrayNavigableSet<String>(false, false, sortedParameters), new FixedArray<BasicTypeID>(false, sortedParameterTypes));
		
		this.isOpen = true;		
	}
	
	public void close() {
			this.isOpen = false;
	}
	
	
	public boolean isClosed() {
		return !this.isOpen;			
	}
	
	public int position(final int index){
		return this.positions[index];
	}
	
	public int findColumn(final String columnLabel){
		int index = this.types.keys().indexOf(columnLabel);
		if(index < 0){
			return -1;
		}
		else{
			return this.positions[index];
		}
	}

	@Override
	protected void finalize() throws Throwable {
		this.close();
	}
	
//	public boolean isFirst() {
//
//	}
//
//	
//	public boolean isLast() {
//
//	}

	
//	public void setFetchSize(final int rows) {
//
//	}
//
//	
//	public int getFetchSize() {
//
//	}
	
	@Override
	protected abstract Logger logger();

}