package de.lmu.ifi.pms.cep.event_mill.sql;

import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementList;
import org.simpleframework.xml.ElementMap;
import org.simpleframework.xml.Root;
import org.simpleframework.xml.Serializer;
import org.simpleframework.xml.core.Persister;
import org.simpleframework.xml.stream.Format;

import de.lmu.ifi.pms.cep.tsa.DTSA_BufferType;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;
import de.lmu.ifi.pms.java.util.string.StringBuilder;


@Root(name="buffer")
public class SQL_Buffer_Spec {
	
	@Element(name="prog_package", required=false)
	public final String prog_package;
	@Element(name="simple_name")
	public final String simple_name;
	public final String fullName;
	
	@Element(name="description", required=false)
	public final String description;
	
	@Element(name="type")
	public final DTSA_BufferType type;
	
	public final FixedArrayNavigableMap<String, BasicType> attributes;
	
	@ElementMap(name="attributes", entry="attribute", key="name", value="type", attribute=true)
	java.util.HashMap<String, BasicType> getAttributes(){
		return new java.util.HashMap<String, BasicType>(this.attributes.javaMap());
	}
	
	/*@ElementList(name="attributes", entry="attribute", type=TypedAttribute.class)
	java.util.ArrayList<TypedAttribute<BasicType>> getAttributes(){
		return new java.util.ArrayList<TypedAttribute<BasicType>>(new CollectionAdapter<Entry<String, BasicType>, BasicTypeAttribute>(this.attributes) {
			
			@Override
			public BasicTypeAttribute adapt(final Entry<String, BasicType> source) {
				return BasicTypeAttribute.create(source.key(), source.value());
			}
			
		}.javaCollection());
	}*/
	
	public final FixedArrayNavigableSet<String> stream_bounds;	
	
	@ElementList(name="min_times", entry="min_time", type=String.class)
	java.util.ArrayList<String> getMin_times(){
		return new java.util.ArrayList<String>(this.stream_bounds.javaCollection());
	}
	
	public SQL_Buffer_Spec(final String prog_package, final String simple_name, final String description, final DTSA_BufferType type,
							final FixedArrayNavigableMap<String, BasicType> attributes,
							final FixedArrayNavigableSet<String> stream_bounds) {
		super();
		assert simple_name.matches(TSA_SyntaxPatterns.BufferNameRegEx);
		assert prog_package==null || prog_package.matches(TSA_SyntaxPatterns.PackageNameRegEx);
		
		this.prog_package = (prog_package == null ? "" : prog_package);
		this.simple_name = simple_name;
		this.fullName = TSA_SyntaxPatterns.bufferFullName(this.prog_package, this.simple_name);
		this.description = (description == null ? "" : description);
		this.type = type;
		this.attributes = attributes;
		this.stream_bounds = stream_bounds;
	}

	SQL_Buffer_Spec(	@Element(name="prog_package", required=false) final String prog_package, 
			@Element(name="simple_name") final String simple_name, @Element(name="description",required=false) final String description, @Element(name="type") final DTSA_BufferType type, 
			@ElementMap(name="attributes", entry="attribute", key="name", value="type", attribute=true) final java.util.HashMap<String, BasicType> attributes,
			@ElementList(name="min_times", entry="min_time") final java.util.ArrayList<String> min_times) {
		this(prog_package, simple_name, description, type, attributesFrom(attributes), mintimesFrom(min_times));
	}
	
	@Override
	public String toString() {
		Serializer serializer = new Persister();
		StringBuilder result = new StringBuilder();
		try {
			serializer.write(this, result);
			return result.toString();
			//return serializer.read(FixedArray.class, xml_buffer_specs);
		} catch (Exception e) {
			throw new IllegalArgumentException(e);
		}
	}
	
	static FixedArrayNavigableMap<String, BasicType> attributesFrom(final java.util.HashMap<String, BasicType> attributes){
		
		return new FixedArrayNavigableMap<String, BasicType>(attributes);
		//TODO avoid sorting by checking if sorted before
	}
	
	/*
	static FixedArrayNavigableMap<String, BasicType> attributesFrom(final java.util.ArrayList<TypedAttribute<BasicType>> attributes){
		String[] attributeNames = new String[attributes.size()];
		BasicType[] attributeTypes = new BasicType[attributes.size()];
		int index = 0;
		for(TypedAttribute<BasicType> attribute : attributes){
			attributeNames[index] = attribute.name;
			attributeTypes[index] = attribute.type;
			++index;
		}
		
		return new FixedArrayNavigableMap<String, BasicType>(attributeNames, attributeTypes);
		//TODO avoid sorting by checking if sorted before
	}*/
	
	static FixedArrayNavigableSet<String> mintimesFrom(final java.util.Collection<String> min_times){
		return new FixedArrayNavigableSet<String>(min_times);
		//TODO avoid sorting by checking if sorted before
	}

	@SuppressWarnings("unchecked")
	public static FixedArray<SQL_Buffer_Spec> deserialize(final String xml_buffer_specs){
		Serializer serializer = new Persister(new Format("<?xml version=\"1.0\" encoding=\"UTF-8\"?>"));
		try {
			return new FixedArray<SQL_Buffer_Spec>(serializer.read(java.util.ArrayList.class, xml_buffer_specs));
		} catch (Exception e) {
			throw new IllegalArgumentException(e);
		}
	}
	
	public static String serialize(final FixedArray<SQL_Buffer_Spec> buffer_specs){
		Serializer serializer = new Persister(new Format("<?xml version=\"1.0\" encoding=\"UTF-8\"?>"));
		StringBuilder result = new StringBuilder();
		try {
			serializer.write(new java.util.ArrayList<SQL_Buffer_Spec>(buffer_specs.javaCollection()), result);
			return result.toString();
			//return serializer.read(FixedArray.class, xml_buffer_specs);
		} catch (Exception e) {
			throw new IllegalArgumentException(e);
		}
	}

}
