package de.lmu.ifi.pms.cep.event_mill.sql;

import java.sql.SQLException;

import de.lmu.ifi.pms.cep.COMMON_CONSTANTS;
import de.lmu.ifi.pms.cep.event_mill.configuration.DEFAULTS;
import de.lmu.ifi.pms.cep.event_mill.execution.program.StreamBuffer;
import de.lmu.ifi.pms.cep.tsa.DTSA_BufferType;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.cep.tsa.time.Clock;

import de.lmu.ifi.pms.java.util.collection.array.Arrays;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArraySet;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;

public class SQL_StreamBuffer extends StreamBuffer<SQL_StreamBuffer, SQL_ExecutableQuery, SQL_ExecutableProgram>{
	
	public final int id;
	public final String schema;
	public final String table_name;
	
	public final int numberOfStreamBounds;
	public final FixedArrayNavigableSet<String> stream_bounds;
	protected final long[] streamBoundWriteValues;
	protected final long[] streamBoundReadValues;
	
	protected long tupleCounter;
	protected long size;
	
	protected FixedArraySet<SQL_ExecutableQuery> inputQueries;

	protected FixedArraySet<SQL_ExecutableQuery> outputQueries;
	int[] inputIndexInOutputQuery;
	
	protected final Clock block_Counter;
	
//	public SQL_StreamBuffer(CERA_MaterialisationPoint materialisationPoint, SQL_StreamBufferStore bufferStore) {
//		this(materialisationPoint, bufferStore, materialisationPoint.Name);
//	}

	public SQL_StreamBuffer(final SQL_ExecutableProgram program, final int id, final String prog_package, final String simple_name, final FixedArrayNavigableMap<String, BasicTypeID> attributes, final FixedArrayNavigableSet<String> min_times, final DTSA_BufferType type, final String SQL_TableName) {
		super(prog_package, simple_name, type, program);
		
		this.id = id;
		this.schema = program.bufferSchemas.get(this.type);
		this.table_name = SQL_TableName;
		
		this.stream_bounds = min_times;
		this.numberOfStreamBounds = this.stream_bounds.size();
		this.streamBoundWriteValues = new long[this.numberOfStreamBounds];
		Arrays.fill(this.streamBoundWriteValues, SQL_CONSTANTS.BIGINT.MIN_VALUE);
		this.streamBoundReadValues = new long[this.numberOfStreamBounds];
		Arrays.fill(this.streamBoundReadValues, SQL_CONSTANTS.BIGINT.MIN_VALUE);
		
		this.tupleCounter = 0;

		this.size = this.readSize();
		
		this.inputQueries = new FixedArraySet<SQL_ExecutableQuery>();
		this.outputQueries = new FixedArraySet<SQL_ExecutableQuery>();
		this.inputIndexInOutputQuery = new int[0];
		
		this.program.register(this);
		
		//FIXME
		this.block_Counter = program.clocks.get(COMMON_CONSTANTS.BLOCK_COUNTER_CLOCK_NAME(this.type, this.full_name));
		assert this.block_Counter != null;
	}
	
	public synchronized long[] streamBoundWriteValues()
	{
		if(this.type == DTSA_BufferType.INPUT){
			//FIXME
			if(this.numberOfStreamBounds > 0){
				assert this.stream_bounds.get(0).equals(TSA_SyntaxPatterns.streamBoundQualifiedName(this.full_name, DEFAULTS._o_block));
				this.streamBoundWriteValues[0] = this.block_Counter.min();
			}
		}
		else{
			//set all values to Long.MAX_VALUE
			this.program.erase(this.streamBoundWriteValues);
			long[] queryOutputStreamBoundWriteValues;
			//compute component-wise minimum over all queries writing to this buffer
			for(SQL_ExecutableQuery query: this.inputQueries()){
				queryOutputStreamBoundWriteValues = query.outputStreamBoundWriteValues();
				for(int index = 0; index < this.numberOfStreamBounds; ++index){
					if(queryOutputStreamBoundWriteValues[index] < this.streamBoundWriteValues[index]){
						this.streamBoundWriteValues[index] = queryOutputStreamBoundWriteValues[index];
					}
				}
			}
		}
		return this.streamBoundWriteValues;//FIXME copy
	}
	
	protected synchronized long[] atStartInitWriteValues(final long[] clockValues)
	{
		if(this.type == DTSA_BufferType.INPUT){
			//FIXME
			if(this.numberOfStreamBounds > 0){
				assert this.stream_bounds.get(0).equals(TSA_SyntaxPatterns.streamBoundQualifiedName(this.full_name, DEFAULTS._o_block));
				this.streamBoundWriteValues[0] = clockValues[this.program.clocks.keys().indexOf(COMMON_CONSTANTS.BLOCK_COUNTER_CLOCK_NAME(this.type, this.full_name))];
			}
		}
		else{
			this.program.erase(this.streamBoundWriteValues);
			long[] queryOutputMinTimes;
			for(SQL_ExecutableQuery query: this.inputQueries()){
				queryOutputMinTimes = query.lastOutputStreamBoundWriteValues;
				for(int index = 0; index < this.numberOfStreamBounds; ++index){
					if(queryOutputMinTimes[index] < this.streamBoundWriteValues[index]){
						this.streamBoundWriteValues[index] = queryOutputMinTimes[index];
					}
				}
			}
		}
		return this.streamBoundWriteValues;//FIXME copy
	}
	
	/**
	 * unit of relevance values
	 */
	
	public synchronized long[] streamBoundReadValues()
	{//FIXME implementation / synchronization
		if(this.type == DTSA_BufferType.INPUT){
			//FIXME
			if(this.numberOfStreamBounds > 0){
				assert this.stream_bounds.get(0).equals(TSA_SyntaxPatterns.streamBoundQualifiedName(this.full_name, DEFAULTS._o_block));
				assert this.outputQueries.size() <= 2;				
				SQL_ExecutableQuery inputQuery = this.outputQueries().get(0);
				if(inputQuery.type.isGarbageQuery){
					inputQuery = this.outputQueries().get(1);
				}
				synchronized (inputQuery.boundValuesLock) {
					this.streamBoundReadValues[0] = inputQuery.lastInputStreamBoundAndClockMinValues[0];
				}
			}
		}
		else if(this.type == DTSA_BufferType.OUTPUT | this.type == DTSA_BufferType.LOG){
			//FIXME
		}
		else{
			//set all values to Long.MAX_VALUE
			this.program.erase(this.streamBoundReadValues);
			long[] queryInputStreamBoundReadValues;
			//compute component-wise minimum over all queries reading from this buffer
			for(int queryIndex = 0; queryIndex < this.outputQueries.size(); ++queryIndex){
				SQL_ExecutableQuery query = this.outputQueries.get(queryIndex);
				if(!query.type.isGarbageQuery){
					queryInputStreamBoundReadValues = query.inputStreamBoundReadValues(this.inputIndexInOutputQuery[queryIndex]);
					assert queryInputStreamBoundReadValues.length == this.streamBoundReadValues.length;
					for(int index = 0; index < this.numberOfStreamBounds; ++index){						
						assert this.stream_bounds.get(index).equals(query.inputStreamBounds.get(query.inputNumberOfPreceedingStreamBoundIdentifiers[this.inputIndexInOutputQuery[queryIndex]] + index));
						if(queryInputStreamBoundReadValues[index] < this.streamBoundReadValues[index]){
							this.streamBoundReadValues[index] = queryInputStreamBoundReadValues[index];
						}
					}
				}
			}
		}
		return this.streamBoundReadValues;//FIXME copy
	}
	
	/**
	 * end of unit
	 */
	
	@Override
	public synchronized FixedArraySet<SQL_ExecutableQuery> inputQueries()
	{
		return this.inputQueries;
	}
	
	public synchronized void RegisterInputQuery(final SQL_ExecutableQuery query){
		SQL_ExecutableQuery[] inputQueries = new SQL_ExecutableQuery[this.inputQueries.size() + 1];
		for(int input = 0; input < this.inputQueries.size(); ++input){
			if(query == this.inputQueries.get(input)){//TODO better test?
				return;
			}
			inputQueries[input] = this.inputQueries.get(input);
		}
		inputQueries[inputQueries.length - 1] = query;
		this.inputQueries = new FixedArraySet<SQL_ExecutableQuery>(false, false, inputQueries);
	}
	
	@Override
	public synchronized FixedArraySet<SQL_ExecutableQuery> outputQueries()
	{
		return this.outputQueries;
	}
	
	public synchronized void RegisterOutputQuery(final SQL_ExecutableQuery query, final int inputIndexInQuery){
		SQL_ExecutableQuery[] outputQueries = new SQL_ExecutableQuery[this.outputQueries.size() + 1];
		int[] inputIndexInOutputQuery = new int[outputQueries.length];
		
		for(int output = 0; output < this.outputQueries.size(); ++output){
			if(query == this.outputQueries.get(output)){//TODO better test?
				return;
			}
			outputQueries[output] = this.outputQueries.get(output);
		}
		System.arraycopy(this.inputIndexInOutputQuery, 0, inputIndexInOutputQuery, 0, this.inputIndexInOutputQuery.length);
		
		outputQueries[outputQueries.length - 1] = query;
		
		assert query.inputs.size() > inputIndexInQuery && query.inputs.values().get(inputIndexInQuery) == this;//TODO better test?
		
		inputIndexInOutputQuery[outputQueries.length - 1] = inputIndexInQuery;

		this.outputQueries = new FixedArraySet<SQL_ExecutableQuery>(false, false, outputQueries);
		this.inputIndexInOutputQuery = inputIndexInOutputQuery;
	}
	
	public synchronized void notifyRead(){

	}
	
	public synchronized void notifyWrite(final long tupleNumber){
		if(tupleNumber > 0){
			this.tupleCounter += tupleNumber;
			this.size += tupleNumber;
		}
		
		assert(true); //FIXME check whether this.size == SELECT COUNT() FROM
	}
	
	public synchronized void notifyDelete(final long tupleNumber){
		if(tupleNumber > 0){
			this.size -= tupleNumber;			
		}
		
		assert(true); //FIXME check whether this.size == SELECT COUNT() FROM
	}
	
	protected synchronized long tupleCount(){
		return this.tupleCounter;
	}
	
	@Override
	public boolean exists(){
		try
		{
			return this.program.sql_translation.existsTable(this.program.bufferStore.connection().getMetaData(), null, this.schema, this.table_name);
		}
		catch (SQLException e)
		{
			throw new RuntimeException(e); //TODO better exception handling
		}
	}
	
	@Override
	public boolean isEmpty(){
		try
		{
			return this.program.sql_translation.isEmpty(this.program.bufferStore.connection(), this.schema, this.table_name);
		}
		catch (SQLException e)
		{
			throw new RuntimeException(e); //TODO better exception handling
		}
	}
	
	public long readSize(){
		try
		{
			return this.program.sql_translation.tableSize(this.program.bufferStore.connection(), this.schema, this.table_name);
		}
		catch (SQLException e)
		{
			throw new RuntimeException(e); //TODO better exception handling
		}
	}

	@Override
	public void clear()
	{
		//FIXME move to SQL_Translation
		try
		{
			//FIXME
			this.program.bufferStore.statement().execute("DELETE FROM \"" + this.schema + "\".\"" + this.table_name + "\";");
		}
		catch (SQLException e)
		{
			throw new RuntimeException(e);//TODO better exception
		}
	}
	
	public void writeTo(final StringBuilder sb){
		sb.append("\n");
		sb.append(this.type.toString());
		sb.append(" ");
		sb.append(this.full_name);
		sb.append("\n");
		sb.append("\n");
		SQL_ExecutableQuery.writeValues(sb, new FixedArray<String>("OutputStreamBoundsWriteValues"), this.stream_bounds, this.streamBoundWriteValues);
		sb.append("\n");
	}
	
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		this.writeTo(sb);
		return sb.toString();
	}
	
}
