package de.lmu.ifi.pms.cep.event_mill.sql;

import org.simpleframework.xml.Attribute;
import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementList;
import org.simpleframework.xml.Path;

import de.lmu.ifi.pms.cep.compilation.CompilerFactory;
import de.lmu.ifi.pms.cep.event_mill.configuration.DEFAULTS;
import de.lmu.ifi.pms.cep.event_mill.configuration.SQL_ConnectionConfiguration;
import de.lmu.ifi.pms.cep.event_mill.configuration.SQL_DatabaseDriver;
import de.lmu.ifi.pms.cep.event_mill.execution.program.StreamBufferStoreFactory;
import de.lmu.ifi.pms.cep.event_mill.sql.compiler.DTSA_SQL_CompilerFactory;
import de.lmu.ifi.pms.cep.tsa.DTSA_Program;
import de.lmu.ifi.pms.cep.tsa.optimization.DTSA_OptimizerFactory;

import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;


public class SQL_StreamBufferStoreFactory extends StreamBufferStoreFactory<SQL_StreamBufferStore, SQL_Program_Spec> {
	
	@Element(name="driver")
	public final SQL_DatabaseDriver driver;
	
	@Element(name="connection")
	public final SQL_ConnectionConfiguration connection;
	
	@Attribute(name="name")
	@Path("meta_schema")
	public final String meta_schema;
	
	@Element(name="start-command", required=false)
	public final String db_start_command;
	
	@Element(name="stop-command", required=false)
	public final String db_stop_command;
	
	public final FixedArray<DTSA_OptimizerFactory> preWriteOptimizers;
	
	@ElementList(inline=true, entry="optimizer", type=DTSA_OptimizerFactory.class, required=false)
	java.util.ArrayList<DTSA_OptimizerFactory> getOptimizers(){
		return new java.util.ArrayList<DTSA_OptimizerFactory>(this.preWriteOptimizers.javaCollection());
	}

	public SQL_StreamBufferStoreFactory( @Element(name="driver") final SQL_DatabaseDriver driver
			, @Element(name="connection") final SQL_ConnectionConfiguration connection
			, @Attribute(name="name") @Path("meta_schema") final String meta_schema
			, @Element(name="start-command", required=false) final String db_start_command
			, @Element(name="stop-command", required=false) final String db_stop_command
			, @ElementList(inline=true, entry="optimizer", type=DTSA_OptimizerFactory.class, required=false) final java.util.ArrayList<DTSA_OptimizerFactory> optimizers
		) {
		super(SQL_StreamBufferStore.class, SQL_Program_Spec.class);
		this.driver = driver;
		this.connection = connection;
		this.meta_schema = meta_schema;
		this.db_start_command = db_start_command;
		this.db_stop_command = db_stop_command;
		this.preWriteOptimizers = (optimizers == null ? new FixedArray<DTSA_OptimizerFactory>(DEFAULTS.defaultDTSAOptimizer()) : new FixedArray<DTSA_OptimizerFactory>(optimizers));
	}

	@Override
	public SQL_StreamBufferStore make() {
		return new SQL_StreamBufferStore(this.compiler(), this.driver, this.connection, this.meta_schema, this.db_start_command, this.db_stop_command);
	}
	
	@Override
	public CompilerFactory<DTSA_Program, SQL_Program_Spec> compiler() {
		return new DTSA_SQL_CompilerFactory(DTSA_SQL_CompilerFactory.defaultIdentifier, this.preWriteOptimizers, this.driver.sqlWriter());
	}

}
