package de.lmu.ifi.pms.cep.event_mill.sql;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.tsa.time.Clock;


public class TupleIdSequenceClock extends Clock {
	
	protected static final Logger logger = LoggerFactory.getLogger(TupleIdSequenceClock.class);
	
	protected final SQL_StreamBufferStore bufferStore;
	protected final String statement;
	protected PreparedStatement getValue;
	boolean usePreparedStatements;
	
	/*
	public TupleIdSequenceClock(final SQL_StreamBufferStore bufferStore, final String schema, final String sequence){
		this(bufferStore, schema, sequence, false);
	}*/
	
	public TupleIdSequenceClock(final SQL_StreamBufferStore bufferStore, final String program_meta_schema, final String sequence, final boolean usePreparedStatements){
		assert program_meta_schema.matches(SQL_Translation.schema_name_RegEx);
		assert sequence.matches(SQL_Translation.sequence_name_RegEx);
		
		this.bufferStore = bufferStore;
		this.statement = bufferStore.sql_translation.sqlGetTupleIdSequenceValue(program_meta_schema);
		this.usePreparedStatements = usePreparedStatements;
		
		this.init();
	}
	
	void init(){
		int tryCount = 2;
		
		while(true){
			Connection connection = this.bufferStore.connection();
			try
			{
				this.getValue = this.bufferStore.sql_translation.prepareStatement(connection, this.statement, this.usePreparedStatements);
				break;
			} catch (SQLException f)
			{
				if(SQL_EXCEPTIONS.isConnectionFailure(f)){
					logger.debug("Connection failure. Trying again.", f);
					this.bufferStore.reconnect(connection);
				}
				else if(SQL_EXCEPTIONS.isMonetDBBug(f) & tryCount > 0){
					logger.warn("MonetDB bug. Trying to reconnect.", f);
					this.bufferStore.reconnect(connection);
					--tryCount;
				}
				else{
					logger.error("", f);
					throw new RuntimeException(f);
					//TODO better exception
				}
			}
		}
	}

	@Override
	public synchronized long now() {
		try {
			while(true){
				try{
						ResultSet value = this.getValue.executeQuery();
						value.next();
						long now = value.getLong(1);
						logger.debug("Retrieved current value of tuple id sequence and return now value: " + now);
						try{ this.getValue.getMoreResults(); } catch (SQLException e){}
						return now;
				}
				catch(SQLException e){
					if(SQL_EXCEPTIONS.isConnectionFailure(e)){
						logger.debug("Connection failure. Trying again.", e);
						this.bufferStore.reconnect(this.getValue.getConnection());
						this.init();
					}
					else if(SQL_EXCEPTIONS.isRollback(e)){
						logger.debug("Reading tuple_id_seq rolled back. Trying again.", e);
					}
					else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
						logger.debug("Handle of read tuple_id_seq statement is invalid. Repreparing statement and trying again.", e);
						this.init();
					}
					else{
						throw e;
					}
				}				
			}
		} catch (SQLException e) {
			logger.error("", e);
			throw new RuntimeException(e);
			//TODO better exception
		}
	}
	
	@Override
	public synchronized long min() {
		try {
			while(true){
				try{
					ResultSet value = this.getValue.executeQuery();
					value.next();
					long min = value.getLong(1) - 1;
					logger.debug("Retrieved current value of tuple id sequence and return min value (= now-1): " + min);
					try{ this.getValue.getMoreResults(); } catch (SQLException e){}
					return min;
				}
				catch(SQLException e){
					if(SQL_EXCEPTIONS.isConnectionFailure(e)){
						logger.debug("Connection failure. Trying again.", e);
						this.bufferStore.reconnect(this.getValue.getConnection());
						this.init();
					}
					else if(SQL_EXCEPTIONS.isRollback(e)){
						logger.debug("Reading tuple_id_seq rolled back. Trying again.", e);
					}
					else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
						logger.debug("Handle of read tuple_id_seq statement is invalid. Repreparing statement and trying again.", e);
						this.init();
					}
					else{
						throw e;
					}
				}				
			}
		} catch (SQLException e) {
			logger.error("", e);
			throw new RuntimeException(e);
			//TODO better exception
		}
	}

}
