package de.lmu.ifi.pms.cep.tsa;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;

import org.simpleframework.xml.Attribute;
import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementList;
import org.simpleframework.xml.Root;

import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_Exception;

import de.lmu.ifi.pms.java.util.adaption.ArrayAdaptor;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.collection.fixed.FixedCollection;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;
import de.lmu.ifi.pms.java.util.map.fixed.FixedEnumMap;


@Root(name="program")
public class DTSA_Program {
		
	public final String namespace;
	
	public final String simple_name;
	@Attribute(name="name")
	public final String full_name;
	@Element(name="description", required=false)
	public final String description;
	
	@ElementList(entry="buffer", inline=true)
	java.util.ArrayList<DTSA_StreamBuffer> getBuffers(){
		java.util.ArrayList<DTSA_StreamBuffer> buffers = new java.util.ArrayList<DTSA_StreamBuffer>();
		for(FixedArrayNavigableMap<String, DTSA_StreamBuffer> bufferSet: this.streamBuffers.values()){
			buffers.addAll(bufferSet.values().javaCollection());
		}
		return buffers;
	}
	
	@ElementList(entry="query", inline=true)
	java.util.ArrayList<DTSA_Query> getQueries(){
		java.util.ArrayList<DTSA_Query> queries = new java.util.ArrayList<DTSA_Query>();
		for(FixedArray<DTSA_Query> querySet: this.queries.values()){
			queries.addAll(querySet.javaCollection());
		}
		return queries;
	}
	
	public final FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>> streamBuffers;
	
	public final FixedEnumMap<DTSA_QueryType, FixedArray<DTSA_Query>> queries;
	
	public final FixedCollection<String> origin;
	
	@ElementList(entry="origin", inline=true, required=false)
	java.util.ArrayList<String> getOriginList(){
		return new java.util.ArrayList<String>(this.origin.javaCollection());
	}
	

	/*public DTSA_Program(final String namespace, final String simple_name, final String description, final FixedArray<DTSA_StreamBuffer> streamBuffers, final FixedArray<DTSA_Query> queries, final FixedCollection<String> origin) {
		this(namespace, simple_name, description, streamBuffers(streamBuffers), queries(queries), origin);
	}
	
	public DTSA_Program(final String namespace, final String simple_name, final String description, final FixedArray<DTSA_StreamBuffer> streamBuffers, final DTSA_Query[] queries, final FixedCollection<String> origin) {
		this(namespace, simple_name, description, streamBuffers, new FixedArray<DTSA_Query>(queries), origin);
	}*/
	public DTSA_Program(
			  @Attribute(name="name") final String full_name
			, @Element(name="description", required=false) final String description
			, @ElementList(entry="buffer", inline=true) final java.util.ArrayList<DTSA_StreamBuffer> buffers
			, @ElementList(entry="query", inline=true) final java.util.ArrayList<DTSA_Query> queries
			, @ElementList(entry="origin", inline=true) final java.util.ArrayList<String> origin
			) throws TSA_Exception{
			this(TSA_SyntaxPatterns.programFullName(full_name)[0], TSA_SyntaxPatterns.programFullName(full_name)[1], (description == null ? "" : description), streamBuffers(new FixedArray<DTSA_StreamBuffer>(buffers)), queries(new FixedArray<DTSA_Query>(queries)), (origin == null ? new FixedArray<String>() : new FixedArray<String>(origin)));
	}
	
	public DTSA_Program(  final String namespace
						, final String simple_name
						, final String description
						, final FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>> streamBuffers
						, final FixedEnumMap<DTSA_QueryType, FixedArray<DTSA_Query>> queries
						, final FixedCollection<String> origin
			) throws TSA_Exception{
		super();
		if(TSA_SyntaxPatterns.isValidNamespace(namespace)){
			this.namespace = namespace;
		}
		else{
			throw new TSA_Exception(origin, "Invalid program namespace: %s", namespace);
		}
		if(TSA_SyntaxPatterns.isValidNamespace(namespace)){
			this.simple_name = simple_name;
		}
		else{
			throw new TSA_Exception(origin, "Invalid simple program name: %s", simple_name);
		}
		this.full_name = TSA_SyntaxPatterns.programFullName(this.namespace, this.simple_name);
		this.description = description;
		
		this.streamBuffers = streamBuffers;
		
		this.queries = queries;
		
		this.origin = origin;
		
		//FIXME check
	}
	
	@SuppressWarnings({ "rawtypes", "unchecked" })
	protected static final FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>> streamBuffers(final FixedArray<DTSA_StreamBuffer> streamBuffers){
		
		FixedEnumMap<DTSA_BufferType, ArrayList<DTSA_StreamBuffer>> groupedStreamBuffers;
		{
			ArrayList<DTSA_StreamBuffer>[] groupedStreamBuffersArray = (ArrayList<DTSA_StreamBuffer>[]) new ArrayList<?>[DTSA_BufferType.values().length];
			for(int type = 0; type < groupedStreamBuffersArray.length; ++type){
				groupedStreamBuffersArray[type] = new ArrayList<DTSA_StreamBuffer>();
			}
			groupedStreamBuffers = new FixedEnumMap<DTSA_BufferType, ArrayList<DTSA_StreamBuffer>>(DTSA_BufferType.class, groupedStreamBuffersArray);
		}
		
		for(DTSA_StreamBuffer streamBuffer: streamBuffers){
			groupedStreamBuffers.get(streamBuffer.type).add(streamBuffer);
		} 
		
		return new FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>>(DTSA_BufferType.class, new ArrayAdaptor<ArrayList<DTSA_StreamBuffer>, FixedArrayNavigableMap<String, DTSA_StreamBuffer>>((Class)FixedArrayNavigableMap.class){
			@Override
			public FixedArrayNavigableMap<String, DTSA_StreamBuffer> map(final ArrayList<DTSA_StreamBuffer> source) {
				DTSA_StreamBuffer[] streamBuffers = source.toArray(new DTSA_StreamBuffer[source.size()]);
				String[] streamBufferNames = new String[streamBuffers.length];
				for (int streamBuffer = 0; streamBuffer < streamBuffers.length; ++streamBuffer) {
					streamBufferNames[streamBuffer] = streamBuffers[streamBuffer].full_name;
				}				
				//TODO check for dublicate names
				return new FixedArrayNavigableMap<String, DTSA_StreamBuffer>(streamBufferNames, streamBuffers);
			}
		}.adapt(groupedStreamBuffers.values().toArray((ArrayList<DTSA_StreamBuffer>[]) new ArrayList[0])));
	}
	
	@SuppressWarnings({ "rawtypes", "unchecked" })
	protected static final FixedEnumMap<DTSA_QueryType, FixedArray<DTSA_Query>> queries(final FixedArray<DTSA_Query> queries){
		ArrayList<DTSA_Query>[] lists = new ArrayList[DTSA_QueryType.values().length];
		for(int type = 0; type < lists.length; ++type){
			lists[type] = new ArrayList<DTSA_Query>();
		}
		
		FixedEnumMap<DTSA_QueryType, ArrayList<DTSA_Query>> groupedQueries = new FixedEnumMap<DTSA_QueryType, ArrayList<DTSA_Query>>(DTSA_QueryType.class, lists);		
		
		for(DTSA_Query query: queries){
			groupedQueries.get(query.type).add(query);
		} 
		
		return new FixedEnumMap<DTSA_QueryType, FixedArray<DTSA_Query>>(DTSA_QueryType.class, new ArrayAdaptor<ArrayList<DTSA_Query>, FixedArray<DTSA_Query>>((Class)FixedArray.class){
			@Override
			public FixedArray<DTSA_Query> map(final ArrayList<DTSA_Query> source) {
				return new FixedArray<DTSA_Query>(false, source.toArray(new DTSA_Query[source.size()]));
			}
		}.adapt((ArrayList<DTSA_Query>[])groupedQueries.values().toArray()));
	}
	
	public void writeTo(final StringBuilder sb, final int indentDepth){
		indentTab(sb, indentDepth);
		if(this.namespace != ""){
			sb.append("NAMESPACE ");
			sb.append(this.namespace);
			sb.append("\n");
			sb.append("\n");
			indentTab(sb, indentDepth);
		}
		
		//FIXME description
		
		sb.append("PROGRAM ");
		sb.append(this.simple_name);
		sb.append("\n\n");
		//FIXME
		for(DTSA_BufferType type: DTSA_BufferType.values()){
			for(DTSA_StreamBuffer temporalStream: this.streamBuffers.get(type).values()){
				temporalStream.writeTo(sb, indentDepth + 1);
				sb.append("\n\n");
			}
		}
		
		for(DTSA_QueryType type: DTSA_QueryType.values()){
			for(DTSA_Query query: this.queries.get(type)){
				query.writeTo(sb, indentDepth + 1);
				sb.append("\n\n");
			}
		}
	}
	
	public void writeTo(final StringBuilder sb){
		this.writeTo(sb, 0);
	}
	
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		this.writeTo(sb);
		return sb.toString();
	}

}
