package de.lmu.ifi.pms.cep.tsa;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;
import static de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula.and;

import org.simpleframework.xml.Attribute;
import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementList;
import org.simpleframework.xml.ElementUnion;
import org.simpleframework.xml.Path;
import org.simpleframework.xml.Root;

import de.lmu.ifi.pms.cep.tsa.conditions.AtomicConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.conditions.TypedAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.analysis.GenericFormulaAnalyser;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_Exception;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_ImbeddingOperator;
import de.lmu.ifi.pms.cep.tsa.schema.ActionStreamSchema;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.cep.tsa.schema.EventStreamSchema;
import de.lmu.ifi.pms.cep.tsa.schema.StateStreamSchema;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchema;
import de.lmu.ifi.pms.cep.tsa.schema.Type;
import de.lmu.ifi.pms.cep.tsa.time.analysis.ArithmeticRelationFormulaExtractor;
import de.lmu.ifi.pms.cep.tsa.time.analysis.TemporalSimpleFocusedFormulaAnalyser;

import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.collection.fixed.FixedCollection;
import de.lmu.ifi.pms.java.util.collection.javaWrapper.FromJavaCollectionWrapper;
import de.lmu.ifi.pms.java.util.filter.AbstractFilter;
import de.lmu.ifi.pms.java.util.filter.Filter;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;


@Root(name="query")
public class TSA_Query
{
	
	public final FixedCollection<String> origin;
	
	@ElementList(name="origin", inline=true)
	java.util.ArrayList<String> getOriginList(){
		return new java.util.ArrayList<String>(this.origin.javaCollection());
	}
	
	@Attribute(name="package", required=false, empty="")
	public final String prog_package;
	
	public final String output_prog_package;
	public final String output_simple_name;
	@Path("./output/")
	@Attribute(name="name")
	public final String output_full_name;
	
	@Path("./output/")
	@ElementUnion({
		@Element(name="event-schema", type=EventStreamSchema.class),
		@Element(name="state-schema", type=StateStreamSchema.class),
		@Element(name="action-schema", type=ActionStreamSchema.class),
		@Element(name="schema", type=TemporalStreamSchema.class)
	})
	public final TemporalStreamSchema output_schema;
	public final boolean output_isStatic;
	
	@Path("./output/")
	@Attribute(name="type")
	public final DTSA_BufferType output_type;
	
	@ElementUnion({
		@Element(name="stream", type=TSA_TemporalStreamExpression.class), 
		@Element(name="apply", type=TSA_CompositeExpression.class)
	})
	public final TSA_Expression expression;
	
	public final boolean isStatic;
	
	@Element(name="description", data=true)
	public final String description;
	
	/*public TSA_Query(final String prog_package, final String output_full_name,
			final TemporalStreamSchema output_schema, final DTSA_BufferType output_type,
			final TSA_Expression expression, final String description, final String... origin) {
		this(prog_package, output_full_name, output_schema, output_type, expression, description, new FixedArray<String>(origin));
	}*/
	
	public TSA_Query(final String prog_package, final String output_full_name,
			final TemporalStreamSchema output_schema, final DTSA_BufferType output_type,
			final TSA_Expression expression, final String description, final String origin) throws TSA_Exception {
		this(prog_package, output_full_name, output_schema, output_type, expression, description, new FixedArray<String>(false, origin));
	}
	
	public TSA_Query(final String prog_package, final String output_full_name,
			final TemporalStreamSchema output_schema, final DTSA_BufferType output_type,
			final TSA_Expression expression, final String description, final java.util.Collection<String> origin) throws TSA_Exception {
		this(prog_package, output_full_name, output_schema, output_type, expression, description, new FromJavaCollectionWrapper<String>(origin));
	}
	
	public TSA_Query(final String prog_package, final String output_full_name,
			final TemporalStreamSchema output_schema, final DTSA_BufferType output_type,
			final TSA_Expression expression, final String description, final FixedCollection<String> origin) throws TSA_Exception {
		this(prog_package, TSA_SyntaxPatterns.bufferFullName(output_full_name), output_schema, output_type, expression, description, origin);
	}
	
/*	public TSA_Query(final String prog_package, final String output_prog_package,
			final String output_simple_name,
			final TemporalStreamSchema output_schema, final DTSA_BufferType output_type,
			final TSA_Expression expression, final String description, final String... origin) {
		this(prog_package, output_prog_package, output_simple_name, output_schema, output_type, expression, description, new FixedArray<String>(origin));
	}*/
	
	public TSA_Query(final String prog_package, final String output_prog_package,
			final String output_simple_name,
			final TemporalStreamSchema output_schema, final DTSA_BufferType output_type,
			final TSA_Expression expression, final String description, final String origin) throws TSA_Exception {
		this(prog_package, output_prog_package, output_simple_name, output_schema, output_type, expression, description, new FixedArray<String>(false, origin));
	}
	
	public TSA_Query(final String prog_package, final String output_prog_package,
			final String output_simple_name,
			final TemporalStreamSchema output_schema, final DTSA_BufferType output_type,
			final TSA_Expression expression, final String description, final java.util.Collection<String> origin) throws TSA_Exception {
		this(prog_package, output_prog_package, output_simple_name, output_schema, output_type, expression, description, new FromJavaCollectionWrapper<String>(origin));
	}
	
	public TSA_Query(final String prog_package, final String output_prog_package,
			final String output_simple_name,
			final TemporalStreamSchema output_schema, final DTSA_BufferType output_type,
			final TSA_Expression expression, final String description, final FixedCollection<String> origin) throws TSA_Exception {
		this(prog_package, new String[]{output_prog_package, output_simple_name}, output_schema, output_type, expression, description, origin);
		
	}
	
	
	public TSA_Query(final String prog_package, final String[] output_full_name,
			final TemporalStreamSchema output_schema, final DTSA_BufferType output_type,
			final TSA_Expression expression, final String description, final FixedCollection<String> origin) throws TSA_Exception {
		super();
		this.prog_package = prog_package;
		this.output_prog_package = output_full_name[0];
		this.output_simple_name = output_full_name[1];
		this.output_full_name = TSA_SyntaxPatterns.bufferFullName(this.output_prog_package, this.output_simple_name);
		this.output_schema = output_schema;
		this.output_isStatic = this.output_schema.isStatic;
		assert output_type == DTSA_BufferType.WORKING | output_type == DTSA_BufferType.OUTPUT | output_type == DTSA_BufferType.LOG;
		this.output_type = output_type;


		this.description = description;
		
		this.origin = origin;
		
		this.expression = this.extendExpression(expression);
		this.isStatic = this.expression.isStatic;
	}
	
	public TSA_Expression extendExpression(final TSA_Expression expression) throws TSA_Exception{
		FixedArrayNavigableSet<String> missingAttributes = FixedArrayNavigableSet.difference(this.output_schema.basicAttributes.keys(), expression.schema().basicAttributes.keys());
		
		if(missingAttributes.size() > 0){
			FixedArrayNavigableMap<String, Term<? extends Type>> implicitAttributeSpecifications = this.output_schema.implicitAttributeSpecifications;
			
			//FIXME what happens if both, default for attribute and subattributes?
			
			//extend to non basic attributes
			
			ArrayList<TypedAttribute<?>> implicitAttributes = new ArrayList<TypedAttribute<?>>(implicitAttributeSpecifications.size());
			ArrayList<Term<?>> definitions = new ArrayList<Term<?>>(implicitAttributeSpecifications.size());
			
			for (int specification = 0; specification < implicitAttributeSpecifications.size(); specification++) {
				String name = implicitAttributeSpecifications.keys().get(specification);
				Term<? extends Type> definition = implicitAttributeSpecifications.values().get(specification);
				assert definition.isTyped;
	
				if(!expression.schema().hasAttribute(name)){
					TypedAttribute<?> implicitAttribute = Term.attr(name, definition.type);
					implicitAttributes.add(implicitAttribute);								
					definitions.add(definition);
				}
			}
			
			TSA_ImbeddingOperator imbeddingOperator = new TSA_ImbeddingOperator(this.origin, new FixedArrayNavigableMap<TypedAttribute<?>, Term<?>>(implicitAttributes.toArray(new TypedAttribute<?>[0]), definitions.toArray(new Term[0])));
			return new TSA_CompositeExpression(imbeddingOperator, expression);
		}
		else{
			return expression;
		}
	}
	
	public boolean isMonotonic(){
		return this.expression.isMonotonic;
	}
	
	public boolean isValid(){
		try{
			this.check();
			//TODO maybe more efficient implementation as isValid does not need to collect exceptions
			return true;
		}
		catch(TSA_Exception e){
			return false;
		}
	}
	
	public void check() throws TSA_Exception{
		this.expression.check();
		this.checkValidOutput();
	}
	
	public void checkValidOutput()  throws TSA_Exception{
		//FIXME check attribute types
		
		FixedArrayNavigableSet<String> commonAttributes = FixedArrayNavigableSet.intersect(this.output_schema.basicAttributes.keys(), this.expression.schema().basicAttributes.keys());
		
		if(commonAttributes.size() < this.output_schema.basicAttributes.size()){
			throw new TSA_Exception(this.origin, "The output schema of the query requires attributes that are not produced by the query expression: \n%s", FixedArrayNavigableSet.difference(this.output_schema.basicAttributes.keys(), commonAttributes));
		}
		
		if(commonAttributes.size() < this.expression.schema().basicAttributes.size()){
			throw new TSA_Exception(this.origin, "The output schema of the query does not contain attributes that are produced by the query expression: \n%s", FixedArrayNavigableSet.difference(this.expression.schema().basicAttributes.keys(), commonAttributes));			
		}
		
		
		if(this.output_schema.isStatic & !this.expression.isStatic){
			throw new TSA_Exception(this.origin, "Queries deriving static data may depend on static data only.");
		}		
		
		TemporalSimpleFocusedFormulaAnalyser identifierAnalyser = new TemporalSimpleFocusedFormulaAnalyser(BasicTypeID.IDENTIFIER.type);
		TemporalSimpleFocusedFormulaAnalyser timestampAnalyser = new TemporalSimpleFocusedFormulaAnalyser(BasicTypeID.TIMESTAMP.type);
		
		GenericFormulaAnalyser gfa = new GenericFormulaAnalyser(identifierAnalyser, timestampAnalyser);
		
		@SuppressWarnings("unchecked")
		Filter<AtomicConditionFormula> focus = AbstractFilter.orFilter(identifierAnalyser.preAdaptfocus, timestampAnalyser.preAdaptfocus);
		
		ConditionFormula expressionTupleConstaints = ArithmeticRelationFormulaExtractor.enhaceClockAnalysis(and(this.expression.schema().tupleConstraints()).flaten().replaceAtomicTerm(ArithmeticRelationFormulaExtractor.enhaceClockAnalysisVariablePrefixer));
		
		if(!gfa.implies(expressionTupleConstaints.focus(focus).normalize(), and(this.output_schema.tupleConstraints()).focus(focus).normalize())){
			throw new TSA_Exception(this.origin, "The query expression does not enforce the tuple constraints of the output type.");
		}
		
		//FIXME implement
		/*if(!gfa.implies(and(this.expression.schema().relationConstraints()).normalize(), and(this.output_schema.relationConstraints()).normalize())){
			throw new TSA_Exception(this.origin, "The query expression does not enforce the relation constraints of the output type.");
		}*/
	}

	public void writeTo(final StringBuilder sb, final int indentDepth){
		indentTab(sb, indentDepth);
		sb.append("DERIVE ");
		sb.append(this.output_type.toString());
		sb.append(" ");
		sb.append(this.output_full_name);
		sb.append(" BY\n");
		this.expression.writeTo(sb, indentDepth + 1);
		sb.append("\n");
		indentTab(sb, indentDepth);
		sb.append("END");
	}
	
	public void writeTo(final StringBuilder sb){
		this.writeTo(sb, 0);
	}
	
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		this.writeTo(sb);
		return sb.toString();
	}

}
