package de.lmu.ifi.pms.cep.tsa;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;

import org.simpleframework.xml.Attribute;
import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementList;
import org.simpleframework.xml.ElementUnion;
import org.simpleframework.xml.Root;

import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_Exception;
import de.lmu.ifi.pms.cep.tsa.schema.ActionStreamSchema;
import de.lmu.ifi.pms.cep.tsa.schema.EventStreamSchema;
import de.lmu.ifi.pms.cep.tsa.schema.StateStreamSchema;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchema;

import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.fixed.FixedCollection;


@Root(name="temporal-stream")
public class TSA_TemporalStream{
	
	public final FixedCollection<String> origin;
	
	@ElementList(name="origin", inline=true)
	java.util.ArrayList<String> getOriginList(){
		return new java.util.ArrayList<String>(this.origin.javaCollection());
	}
	
	public final String prog_package;
	public final String simple_name;
	@Attribute(name="name")
	public final String full_name;
	
	@Element(name="description", data=true)
	public final String description;
	
	@ElementUnion({
		@Element(name="event-schema", type=EventStreamSchema.class),
		@Element(name="state-schema", type=StateStreamSchema.class),
		@Element(name="action-schema", type=ActionStreamSchema.class),
		@Element(name="schema", type=TemporalStreamSchema.class)
	})
	public final TemporalStreamSchema schema;
	
	public final boolean isStatic;
	
	public final FixedArrayNavigableSet<DTSA_BufferType> types;
	
	@ElementList(name="types", entry="type")
	java.util.ArrayList<DTSA_BufferType> getTypeList(){
		return new java.util.ArrayList<DTSA_BufferType>(this.types.javaCollection());
	}
	
	public TSA_TemporalStream(final String prog_package, final String simple_name,
			final String description, final TemporalStreamSchema schema, final FixedArrayNavigableSet<DTSA_BufferType> types, final FixedCollection<String> origin) throws TSA_Exception {
		this(new String[]{prog_package, simple_name}, description, schema, types, origin);		
	}
	
	public TSA_TemporalStream(final String full_name, final String description, final TemporalStreamSchema schema, final FixedArrayNavigableSet<DTSA_BufferType> types, final String origin) throws TSA_Exception {
		this(full_name, description, schema, types, new FixedArray<String>(false, origin));
	}
	
	public TSA_TemporalStream(final String full_name, final String description, final TemporalStreamSchema schema, final FixedArrayNavigableSet<DTSA_BufferType> types, final FixedCollection<String> origin) throws TSA_Exception {
		this(TSA_SyntaxPatterns.bufferFullName(full_name), description, schema, types, origin);
	}

	protected TSA_TemporalStream(final String[] full_name, final String description, 
		final TemporalStreamSchema schema, final FixedArrayNavigableSet<DTSA_BufferType> types, final FixedCollection<String> origin) throws TSA_Exception{
		super();
		this.prog_package = full_name[0];
		this.simple_name = full_name[1];
		this.full_name = TSA_SyntaxPatterns.bufferFullName(this.prog_package, this.simple_name);
		this.description = description;
		//FIXME schema constraints may not contain variables
		this.schema = schema;
		this.isStatic = this.schema.isStatic;
		if(!this.isStatic & !(this.schema.attributes.size() > 0) ){
			throw new TSA_Exception(origin, "A non-static temporal stream schema must have at least one attribute.");
		}
		this.types = types;
		
		this.origin = origin;
	}
	
	public void writeTo(final StringBuilder sb, final int indentDepth, final boolean indentFirstLine, final boolean finish){
		if(indentFirstLine){
			indentTab(sb, indentDepth);
		}
		if(this.prog_package != ""){
			sb.append("PACKAGE ");
			sb.append(this.prog_package);
			sb.append("\n\n");
			indentTab(sb, indentDepth);
		}
		if(this.types.size() > 0){
			for(DTSA_BufferType type: this.types){
				sb.append(type.toString());
				sb.append(" ");
			}
			sb.setLength(sb.length() - " ".length());
			sb.append("\n");
			indentTab(sb, indentDepth);
		}
		this.schema.writeTo(sb, indentDepth, false, this.simple_name, finish);
	}
	
	public void writeTo(final StringBuilder sb, final int indentDepth){
		this.writeTo(sb, indentDepth, false, true);
	}
	
	public void writeTo(final StringBuilder sb){
		this.writeTo(sb, 0);
	}
	
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		this.writeTo(sb);
		return sb.toString();
	}

}
