package de.lmu.ifi.pms.cep.tsa.conditions;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;

import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementUnion;

import de.lmu.ifi.pms.cep.tsa.schema.AType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;

import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.filter.Filter;


public class AtomicStreamBoundFormula extends AtomicConditionFormula {
	
	@ElementUnion({
		@Element(name="bound-identifier", type=StreamBound.class),
		@Element(name="clock", type=ClockFunction.class),
		@Element(name="bound", type=Term.class)
	}
	)
	public final Term<BasicType> bound;
	
	@ElementUnion({
		@Element(name="attribute",type=Attribute.class),
		@Element(name="var",type=Variable.class),
		@Element(name="term", type=Term.class)
	})
	public final AtomicTerm<?> atomicTerm; 
	
	//new according to this stream bound means new in general
	public final boolean isMonotonic;
	
	@org.simpleframework.xml.Attribute(name="monotonic", required=false)
	public Boolean getIsMonotonic(){
		if(this.isMonotonic){
			return true;
		}
		else{
			return null;
		}
	}
	
	public AtomicStreamBoundFormula(final ClockFunction clock, final BasicTypeAttribute attribute, final boolean isMonotonic){
		this(clock, (AtomicTerm<?>) attribute, isMonotonic);
	}
	
	public AtomicStreamBoundFormula(final StreamBound streamBound, final BasicTypeAttribute attribute, final boolean isMonotonic){
		this(streamBound, (AtomicTerm<?>) attribute, isMonotonic);
	}
	
	public AtomicStreamBoundFormula(final String buffer_name, final BasicTypeAttribute attribute, final boolean isMonotonic){
		this(new StreamBound(buffer_name, attribute.name, attribute.type), attribute, isMonotonic);
	}
	
	public AtomicStreamBoundFormula(final String buffer_name, final BasicTypeAttribute attribute){
		this(buffer_name, attribute, false);
	}
	
	public AtomicStreamBoundFormula(final StreamBound streamBound, final Attribute<AType> attribute){
		this(streamBound, attribute, false);
	}
	
	protected AtomicStreamBoundFormula(
		  @Element(name="bound", type=AtomicTerm.class) final Term<BasicType> bound
		, @Element(name="term")final AtomicTerm<?> term
		, @org.simpleframework.xml.Attribute(name="monotonic", required=false) final Boolean isMonotonic) {
		super(true, true);
		if(bound instanceof StreamBound | bound instanceof ClockFunction){
			if((term instanceof Attribute<?> | term instanceof Variable<?>)){//FIXME only untyped or basic
				this.bound = bound;
				this.atomicTerm = term;
				this.isMonotonic = (isMonotonic == null ? false : isMonotonic);
			}
			else{
				throw new IllegalArgumentException();
			}
		}
		else{
			throw new IllegalArgumentException();
		}
	}

	public AtomicStreamBoundFormula(final StreamBound streamBound, final Attribute<AType> attribute, final boolean isMonotonic) {
		this(streamBound, (AtomicTerm<?>) attribute, isMonotonic);
	}
	
	public AtomicStreamBoundFormula(final StreamBound streamBound, final Variable<AType> variable, final boolean isMonotonic) {
		this(streamBound, (AtomicTerm<?>) variable, isMonotonic);
	}
	
	public boolean isRelativeToBoundIdentifier(){
		return this.bound instanceof StreamBound;
	}
	
	public boolean isRelativetoClock(){
		return this.bound instanceof ClockFunction;
	}

	@Override
	public AtomicStreamBoundFormula replaceAtomicTerm(final Adapter<? super AtomicTerm<?>, ? extends Term<?>> replacement) {
		Term<?> newAtomicTerm = this.atomicTerm.replaceAtomicTerm(replacement);
		Term<?> newStreamBound = this.bound.replaceAtomicTerm(replacement);
		if(newStreamBound instanceof StreamBound | newStreamBound instanceof ClockFunction){
			if(newAtomicTerm instanceof Attribute<?>){//FIXME only untyped or basic
				return new AtomicStreamBoundFormula((StreamBound)newStreamBound, (Attribute<?>)newAtomicTerm, this.isMonotonic);
			}
			else if(newAtomicTerm instanceof Variable<?>){//FIXME only untyped or basic
				return new AtomicStreamBoundFormula((StreamBound)newStreamBound, (Variable<?>)newAtomicTerm, this.isMonotonic);
			}
			else{
				throw new IllegalArgumentException("Replacement must be an attribute or variable.");			
			}
		}
		else{
			throw new IllegalArgumentException("Replacement must be an stream bound.");
		}
	}
	
	@Override
	public AtomicStreamBoundFormula replaceTerm(final Adapter<Term<?>, ? extends Term<?>> replacement) {
		return this.replaceAtomicTerm(replacement);
	}
	
	@Override
	protected void collect(final Filter<Term<?>> filter, final ArrayList<Term<?>> terms) {
		this.atomicTerm.collect(filter, terms);
	}

	@Override
	protected int compareToImpl(final AtomicConditionFormula o) {
		int cmp = super.compareToImpl(o);
		if(cmp == 0){
			AtomicStreamBoundFormula c = (AtomicStreamBoundFormula)o;
			cmp = this.atomicTerm.compareTo(c.atomicTerm);
		}
		return cmp;
	}
	
	@Override
	public int hashCode() {
		return super.hashCode() ^ this.atomicTerm.hashCode();
	}
	
	@Override
	public void writeTo(final StringBuilder sb, final int indentDepth) {
		indentTab(sb, indentDepth);
		sb.append("bounded( ");
		this.bound.writeTo(sb, 0);
		sb.append(" , ");
		this.atomicTerm.writeTo(sb, 0);
		sb.append(" , ");
		sb.append("monotonic = ");
		sb.append(this.isMonotonic);
		sb.append(" )");
	}
	
}
