package de.lmu.ifi.pms.cep.tsa.conditions;


import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;

import org.simpleframework.xml.ElementList;
import org.simpleframework.xml.ElementListUnion;

import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.collection.AbstractCollection;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.filter.Filter;


abstract class CompositeConditionFormula<SubType extends CompositeConditionFormula<SubType>> extends ConditionFormula {
	
	protected final ConditionFormula[] subformulas;
	
	@ElementListUnion({//FIXME inline
		@ElementList(entry="and",type=ConjunctiveConditionFormula.class,inline=true,required=false,empty=false),
		@ElementList(entry="or",type=DisjunctiveConditionFormula.class,inline=true,required=false,empty=false),
		@ElementList(entry="not",type=NegatedConditionFormula.class,inline=true,required=false,empty=false),
		@ElementList(entry="top",type=ConditionFormula.TopC.class,inline=true,required=false,empty=false),
		@ElementList(entry="bottom",type=ConditionFormula.BottomC.class,inline=true,required=false,empty=false),
		@ElementList(entry="foreign-key",type=ForeignKeyConstraint.class,inline=true,required=false,empty=false),
		@ElementList(entry="primary-key",type=PrimaryKeyConstraint.class,inline=true,required=false,empty=false),
		@ElementList(entry="stream-bound",type=StreamBoundConstraint.class,inline=true,required=false,empty=false),
		@ElementList(entry="compare",type=BinaryComparissonFormula.class,inline=true,required=false,empty=false),
		@ElementList(entry="user-relation",type=UserRelationFormula.class,inline=true,required=false,empty=false),
		@ElementList(entry="relation",type=RelationFormula.class,inline=true,required=false,empty=false),
		@ElementList(entry="boolean-term",type=BooleanTerm.class,inline=true,required=false,empty=false),
		@ElementList(entry="condition",type=ConditionFormula.class,inline=true,required=false,empty=false)
	})
	java.util.ArrayList<ConditionFormula> getSubFormulaList(){
		return new java.util.ArrayList<ConditionFormula>(java.util.Arrays.asList(this.subformulas));
	}

	CompositeConditionFormula(final ConditionFormulaID ID, final boolean isNormalized, final boolean isFlatened, final ConditionFormula... subformulas) {
		super(ID, isNormalized, isFlatened);
		this.subformulas = subformulas;
		//TODO copy array?
	}
	
	public FixedArray<ConditionFormula> subformulas(){
		return new FixedArray<ConditionFormula>(false, this.subformulas);
	}

	@Override
	protected abstract AbstractCollection<ConditionFormula> DNFiteration(boolean positivePolarity);

	@Override
	protected abstract AbstractCollection<ConditionFormula> CNFiteration(boolean positivePolarity);
	
	@Override
	public SubType focus(final Filter<? super AtomicConditionFormula> topic){
		return this.focus(topic, true);
	}
	
	@SuppressWarnings("unchecked")
	@Override
	protected SubType focus(final Filter<? super AtomicConditionFormula> topic, final boolean positivePolarity) {
		ConditionFormula[] subformulas = new ConditionFormula[this.subformulas.length];
		boolean changed = false;
		for(int index = 0; index < subformulas.length; ++index){
			subformulas[index] = this.subformulas[index].focus(topic, positivePolarity);
			if(subformulas[index] != this.subformulas[index]){
				changed = true;
			}
		}
		if(changed){
			return this.newFormula(subformulas);
		}
		else{
			return (SubType) this;
		}
	}
	
	@SuppressWarnings("unchecked")
	@Override
	public SubType replaceAtomicFormula(final Adapter<AtomicConditionFormula, ? extends ConditionFormula> replacement) {
		ConditionFormula[] subformulas = new ConditionFormula[this.subformulas.length];
		boolean changed = false;
		for(int index = 0; index < subformulas.length; ++index){
			subformulas[index] = this.subformulas[index].replaceAtomicFormula(replacement);
			if(subformulas[index] != this.subformulas[index]){
				changed = true;
			}
		}
		if(changed){
			return this.newFormula(subformulas);
		}
		else{
			return (SubType) this;
		}
	}
	
	@SuppressWarnings("unchecked")
	@Override
	public SubType replaceAtomicTerm(final Adapter<? super AtomicTerm<?>, ? extends Term<?>> replacement) {
		ConditionFormula[] subformulas = new ConditionFormula[this.subformulas.length];
		boolean changed = false;
		for(int index = 0; index < subformulas.length; ++index){
			subformulas[index] = this.subformulas[index].replaceAtomicTerm(replacement);
			if(subformulas[index] != this.subformulas[index]){
				changed = true;
			}
		}
		if(changed){
			return this.newFormula(subformulas);
		}
		else{
			return (SubType) this;
		}
	}
	
	@SuppressWarnings("unchecked")
	@Override
	public SubType replaceTerm(final Adapter<Term<?>, ? extends Term<?>> replacement) {
		ConditionFormula[] subformulas = new ConditionFormula[this.subformulas.length];
		boolean changed = false;
		for(int index = 0; index < subformulas.length; ++index){
			subformulas[index] = this.subformulas[index].replaceTerm(replacement);
			if(subformulas[index] != this.subformulas[index]){
				changed = true;
			}
		}
		if(changed){
			return this.newFormula(subformulas);
		}
		else{
			return (SubType) this;
		}
	}

	protected abstract SubType newFormula(ConditionFormula[] subformulas);
	
	@Override
	protected void collectFormulas(final Filter<? super ConditionFormula> filter, final ArrayList<ConditionFormula> formulas) {
		super.collectFormulas(filter, formulas);
		for(ConditionFormula subFormula: this.subformulas){
			subFormula.collectFormulas(filter, formulas);
		}
	}
	
	@Override
	protected void collect(final Filter<Term<?>> filter, final ArrayList<Term<?>> terms) {
		for(ConditionFormula subFormula: this.subformulas){
			subFormula.collect(filter, terms);
		}
	}
	
	@Override
	public void writeTo(final StringBuilder sb, final int indentDepth) {
		indentTab(sb, indentDepth);
		sb.append(this.ID.toString());
		if(this.subformulas.length > 0){
			sb.append("{\n");
			for(ConditionFormula subFormula: this.subformulas){
				subFormula.writeTo(sb, indentDepth + 1);
				sb.append(",\n");
			}
			sb.setLength(sb.length() - ",\n".length());
			sb.append("\n");
			indentTab(sb, indentDepth);
			sb.append("}");
		}
		else{
			sb.append("{}");
		}
	}

}
