package de.lmu.ifi.pms.cep.tsa.conditions;

import java.util.ArrayList;
import java.util.Arrays;

import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementList;
import org.simpleframework.xml.ElementListUnion;
import org.simpleframework.xml.Root;
import org.simpleframework.xml.Serializer;
import org.simpleframework.xml.core.Persister;
import org.simpleframework.xml.core.Validate;

import de.lmu.ifi.pms.cep.tsa.schema.AType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.cep.tsa.schema.NoType;
import de.lmu.ifi.pms.cep.tsa.schema.Type;

import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.filter.Filter;


@Root(name="tuple")
public abstract class CompositeTerm<TypeClass extends AType> extends Term<TypeClass> {
	
	@Element(name="type", required=false)
	protected TypeClass getType(){
		if(this.type == NoType.instance){
			return null;
		}
		else{
			assert this.type instanceof Type;
			return this.type;
		}
	}
	
	protected final Term<?>[] argumentsArray;
	
	@SuppressWarnings("unchecked")
	@ElementListUnion({
		@ElementList(entry="plus", type=Plus.class, inline=true, required=false, empty=false),
		@ElementList(entry="minus", type=Minus.class, inline=true, required=false, empty=false),
		@ElementList(entry="mult", type=Times.class, inline=true, required=false, empty=false),
		@ElementList(entry="div", type=Division.class, inline=true, required=false, empty=false),
		@ElementList(entry="power", type=Power.class, inline=true, required=false, empty=false),
		@ElementList(entry="greatest", type=Greatest.class, inline=true, required=false, empty=false),
		@ElementList(entry="least", type=Least.class, inline=true, required=false, empty=false),
		@ElementList(entry="function", type=UserTupleFunction.class, inline=true, required=false, empty=false),
		@ElementList(entry="min", type=Min.class, inline=true, required=false, empty=false),
		@ElementList(entry="max", type=Max.class, inline=true, required=false, empty=false),
		@ElementList(entry="avg", type=Average.class, inline=true, required=false, empty=false),
		@ElementList(entry="sum", type=Sum.class, inline=true, required=false, empty=false),
		@ElementList(entry="count", type=Count.class, inline=true, required=false, empty=false),
		@ElementList(entry="aggregation", type=UserAggregationFunction.class, inline=true, required=false, empty=false),
		@ElementList(entry="cast", type=CastFunction.class, inline=true, required=false, empty=false),
		@ElementList(entry="clock", type=ClockFunction.class, inline=true, required=false, empty=false),
		@ElementList(entry="min-time", type=StreamBound.class, inline=true, required=false, empty=false),
		@ElementList(entry="tuple", type=NamedTypeRecord.class, inline=true, required=false, empty=false),
		@ElementList(entry="var", type=Variable.class, inline=true, required=false, empty=false),
		@ElementList(entry="attribute", type=Attribute.class, inline=true, required=false, empty=false),
		@ElementList(entry="timestamp", type=TimestampValue.class, inline=true, required=false, empty=false),
		@ElementList(entry="duration", type=DurationValue.class, inline=true, required=false, empty=false),
		@ElementList(entry="boolean", type=BooleanValue.class, inline=true, required=false, empty=false),
		@ElementList(entry="int", type=IntValue.class, inline=true, required=false, empty=false),
		@ElementList(entry="long", type=LongValue.class, inline=true, required=false, empty=false),
		@ElementList(entry="float", type=FloatValue.class, inline=true, required=false, empty=false),
		@ElementList(entry="double", type=DoubleValue.class, inline=true, required=false, empty=false),
		@ElementList(entry="string", type=StringValue.class, inline=true, required=false, empty=false),
		@ElementList(entry="id", type=IdentifierValue.class, inline=true, required=false, empty=false),
		@ElementList(entry="ref", type=ReferenceValue.class, inline=true, required=false, empty=false),
		@ElementList(entry="literal", type=ConstantLiteral.class, inline=true, required=false, empty=false),
		@ElementList(entry="const", type=Constant.class, inline=true, required=false, empty=false),
		@ElementList(entry="param", type=Parameter.class, inline=true, required=false, empty=false),
		@ElementList(entry="term", type=Term.class, inline=true, required=false, empty=false)
	})
	protected java.util.ArrayList<Term<?>> getArguments(){
		if(this.argumentsArray.length > 0){
			return new ArrayList<Term<?>>(Arrays.asList(this.argumentsArray));
		}
		else{
			return null;
		}
	}
	
	protected CompositeTerm(final TermID ID, final TypeClass type, final boolean copy, final Term<?>... arguments) {
		super(ID, type);
		if(copy){
			this.argumentsArray = Arrays.copyOf(arguments, arguments.length);
		}
		else{
			this.argumentsArray = arguments;
		}
	}
	
	@Validate
	protected void validate(){
		
	}
	
	public FixedArray<Term<?>> arguments(){
		return new FixedArray<Term<?>>(false, this.argumentsArray);
	}
	
	public final int arity(){
		return this.argumentsArray.length;
	}
	
	public Term<?> arg(final int number){
		return this.argumentsArray[number];
		//TODO better exception
	}
	
	@Override
	public CompositeTerm<TypeClass> replaceAtomicTerm(final Adapter<? super AtomicTerm<?>, ? extends Term<?>> replacement) {
		Term<?>[] arguments = new Term<?>[this.argumentsArray.length];
		boolean changed = false;
		for(int index = 0; index < arguments.length; ++index){
			arguments[index] = this.argumentsArray[index].replaceAtomicTerm(replacement);
			if(arguments[index] != this.argumentsArray[index]){
				changed = true;
			}
		}
		if(changed){
			return this.newTerm(arguments);
		}
		else{
			return this;
		}
	}
	
	public abstract CompositeTerm<TypeClass> newTerm(final Term<?>... arguments);
	
	@Override
	protected void collect(final Filter<Term<?>> filter, final de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList<Term<?>> terms) {
		super.collect(filter, terms);
		for(Term<?> argument: this.argumentsArray){
			argument.collect(filter, terms);
		}
	}
	
	@Override
	protected final int compareToImpl(final Term<?> o) {
		int cmp = this.compareHead(o);
		if(cmp == 0){//o is at least of type CompositeTerm
			CompositeTerm<?> f = (CompositeTerm<?>) o;
			this.compareArgs(f);
		}
		return cmp;
	}
	
	protected int compareHead(final Term<?> o) {
		return super.compareToImpl(o);
	}
	
	protected int compareArgs(final CompositeTerm<?> f){
		int cmp = this.arity() - f.arity();
		if(cmp == 0){
			assert this.arity() == f.arity();
			int arg = 0;
			int arity = this.arity();
			while(cmp == 0 & arg < arity){
				cmp = this.arg(arg).compareTo(f.arg(arg));
				++arg;
			}
		}
		return cmp;
	}
	
	@Override
	public int hashCode() {
		int hashcode =  super.hashCode() ^ this.arity();
		for(Term<?> arg: this.argumentsArray){
			hashcode ^= arg.hashCode();
		}
		return hashcode;
	}

	@Override
	public String toString() {
		return String.format("%s", Arrays.toString(this.argumentsArray));
	}
	
	public static void main(final String[] args) throws Exception {
		Serializer s = new Persister();
		s.write(apply(TupleFunctionID.PLUS, BasicTypeID.DOUBLE.type, apply(AggregationFunctionID.AVG, BasicTypeID.DOUBLE.type, attr("a", BasicTypeID.LONG)), attr("b", BasicTypeID.DOUBLE.type)), System.out);
	}
}
