package de.lmu.ifi.pms.cep.tsa.conditions;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;

import java.util.Arrays;

import org.simpleframework.xml.Attribute;
import org.simpleframework.xml.Element;

import de.lmu.ifi.pms.cep.tsa.schema.NoType;

import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.filter.Filter;


public abstract class RelationFormula extends AtomicConditionFormula {
	
	@Attribute(name="name")
	public final String name;
	
	final Term<?>[] terms;
	
	@Element(name="terms")
	Tuple<NoType> getTermTuple(){
		return new Tuple<NoType>(NoType.instance, this.terms);
	}
	
	RelationFormula(final String name, final Term<?>... terms) {
		this(true, true, name, terms);
	}

	RelationFormula(final boolean isNormalized, final boolean isFlatened, final String name, final Term<?>... terms) {
		super(isNormalized, isFlatened);
		this.name = name;
		this.terms = terms;
	}

	@Override
	public RelationFormula replaceAtomicTerm(final Adapter<? super AtomicTerm<?>, ? extends Term<?>> replacement) {
		Term<?>[] arguments = new Term[this.terms.length];
		boolean changed = false;
		for(int index = 0; index < arguments.length; ++index){
			arguments[index] = this.terms[index].replaceAtomicTerm(replacement);
			if(arguments[index] != this.terms[index]){
				changed = true;
			}
		}
		if(changed){
			return this.newFormula(arguments);
		}
		else{
			return this;
		}
	}
	
	@Override
	public RelationFormula replaceTerm(final Adapter<Term<?>, ? extends Term<?>> replacement) {
		Term<?>[] arguments = new Term[this.terms.length];
		boolean changed = false;
		for(int index = 0; index < arguments.length; ++index){
			arguments[index] = replacement.adapt(this.terms[index]);
			//FIXME
			if(arguments[index] != this.terms[index]){
				changed = true;
			}
		}
		if(changed){
			return this.newFormula(arguments);
		}
		else{
			return this;
		}
	}
	
	protected abstract RelationFormula newFormula(final Term<?>... arguments);
	
	public final int arity(){
		return this.terms.length;
	}
	
	public Term<?> term(final int number){
		return this.terms[number];
		//TODO better exception
	}

	@Override
	protected void collect(final Filter<Term<?>> filter, final ArrayList<Term<?>> terms) {
		for(Term<?> term: this.terms){
			term.collect(filter, terms);
		}
	}
	
	public FixedArray<Term<?>> terms(){
		return new FixedArray<Term<?>>(false, this.terms);
	}
	
	@Override
	protected int compareToImpl(final AtomicConditionFormula o) {
		int cmp = super.compareToImpl(o);
		if(cmp == 0){
			RelationFormula rf = (RelationFormula) o;
			cmp = this.name.compareTo(rf.name);
			if(cmp == 0){
				cmp = this.arity() - rf.arity();
				if(cmp == 0){
					assert this.arity() == rf.arity();
					int term = 0;
					int arity = this.arity();
					while(cmp == 0 & term < arity){
						cmp = this.term(term).compareTo(rf.term(term));
						++term;
					}
				}
			}
		}
		return cmp;
	}
	
	@Override
	public int hashCode() {
		int hashcode = super.hashCode() ^ this.name.hashCode();
		for(Term<?> term: this.terms){
			hashcode ^= term.hashCode();
		}
		return hashcode;
	}

	@Override
	public void writeTo(final StringBuilder sb, final int indentDepth) {
		indentTab(sb, indentDepth);
		sb.append(this.name);		
		if(this.terms.length > 0){
			sb.append("( ");
			for(Term<?> term: this.terms){
				term.writeTo(sb, indentDepth);
				sb.append(", ");
			}
			sb.setLength(sb.length() - ", ".length());
			sb.append(" )");
		}
	}
	
	@Override
	public String toString() {
		return String.format("%s%s", this.name, Arrays.toString(this.terms));
	}
}
