package de.lmu.ifi.pms.cep.tsa.conditions;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;

import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementUnion;

import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.filter.Filter;


public class StreamBoundConstraint extends AtomicConditionFormula {
	
	@ElementUnion({
		@Element(name="attribute",type=Attribute.class),
		@Element(name="var",type=Variable.class),
		@Element(name="term", type=Term.class)
	})
	public final AtomicTerm<?> atomicTerm; 
	
	public final boolean isMonotonic;
	
	@org.simpleframework.xml.Attribute(name="monotonic", required=false)
	public Boolean getIsMonotonic(){
		if(this.isMonotonic){
			return true;
		}
		else{
			return null;
		}
	}
	
	public StreamBoundConstraint(final Attribute<?> attribute){
		this(attribute, false);
	}
	
	protected StreamBoundConstraint(	@ElementUnion({
		@Element(name="attribute",type=Attribute.class),
		@Element(name="var",type=Variable.class),
		@Element(name="term")
	}) final AtomicTerm<?> term, @org.simpleframework.xml.Attribute(name="monotonic", required=false) final Boolean isMonotonic) {
		super(true, true);
		if(term instanceof Attribute<?> | term instanceof Variable<?>){
			this.atomicTerm = term;
			this.isMonotonic = (isMonotonic == null ? false : isMonotonic);
		}
		else{
			throw new IllegalArgumentException();
		}
	}

	public StreamBoundConstraint(final Attribute<?> attribute, final boolean isMonotonic) {
		super(true, true);
		this.atomicTerm = attribute;
		this.isMonotonic = isMonotonic;
	}
	
	public StreamBoundConstraint(final Variable<?> variable, final boolean isMonotonic) {
		super(true, true);
		this.atomicTerm = variable;
		this.isMonotonic = isMonotonic;
	}

	@Override
	public StreamBoundConstraint replaceAtomicTerm(final Adapter<? super AtomicTerm<?>, ? extends Term<?>> replacement) {
		Term<?> newAtomicTerm = replacement.adapt(this.atomicTerm);
		if(newAtomicTerm instanceof Attribute){//FIXME untyped or basic type
			return new StreamBoundConstraint((Attribute<?>)newAtomicTerm, this.isMonotonic);
		}
		else if(newAtomicTerm instanceof Variable){
			return new StreamBoundConstraint((Variable<?>)newAtomicTerm, this.isMonotonic);
		}
		else{
			throw new IllegalArgumentException("Replacement must be an attribute");			
		}
	}
	
	@Override
	public StreamBoundConstraint replaceTerm(final Adapter<Term<?>, ? extends Term<?>> replacement) {
		return this.replaceAtomicTerm(replacement);
	}
	
	@Override
	protected void collect(final Filter<Term<?>> filter, final ArrayList<Term<?>> terms) {
		this.atomicTerm.collect(filter, terms);
	}

	@Override
	protected int compareToImpl(final AtomicConditionFormula o) {
		int cmp = super.compareToImpl(o);
		if(cmp == 0){
			StreamBoundConstraint c = (StreamBoundConstraint)o;
			cmp = this.atomicTerm.compareTo(c.atomicTerm);
		}
		return cmp;
	}
	
	@Override
	public int hashCode() {
		return super.hashCode() ^ this.atomicTerm.hashCode();
	}
	
	@Override
	public void writeTo(final StringBuilder sb, final int indentDepth) {
		indentTab(sb, indentDepth);
		sb.append("isStreamBound( ");
		this.atomicTerm.writeTo(sb, 0);
		sb.append(" , ");
		sb.append("monotonic = ");
		sb.append(this.isMonotonic);
		sb.append(" )");
	}
	
}
