package de.lmu.ifi.pms.cep.tsa.conditions.analysis;

import static de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula.*;
import static de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormulaID.*;

import de.lmu.ifi.pms.cep.tsa.conditions.AtomicConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.ConjunctiveConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.NegatedConditionFormula;

import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.collection.composition.FixedComposedCollection;

public class GenericFormulaAnalyser {
	
	public final FixedArray<SimpleFocusedFormulaAnalyser> simpleAnalysers;
	
	public final Adapter<AtomicConditionFormula, ConditionFormula> combinedAdaptor;
	
	public GenericFormulaAnalyser(final SimpleFocusedFormulaAnalyser... simpleAnalysers) {
		this(new FixedArray<SimpleFocusedFormulaAnalyser>(simpleAnalysers));
	}
	
	public GenericFormulaAnalyser(final FixedArray<SimpleFocusedFormulaAnalyser> simpleAnalysers) {
		super();
		this.simpleAnalysers = simpleAnalysers;
		this.combinedAdaptor = new Adapter<AtomicConditionFormula, ConditionFormula>() {
			
			@Override
			public ConditionFormula adapt(final AtomicConditionFormula atom) {
				//TODO maybe class AdoptedAtom with assignment to analyser
				ArrayList<AtomicConditionFormula> adaptedAtoms = new ArrayList<AtomicConditionFormula>();
				boolean addedOriginal = false;
				for(SimpleFocusedFormulaAnalyser simpleAnalyser: simpleAnalysers){
					AtomicConditionFormula adoptedAtom = simpleAnalyser.localType().adapt(atom);
					//IDEA buffer adapters;
					if(adoptedAtom == null){
						
					}
					else if(adoptedAtom == atom){
						if(!addedOriginal){
							adaptedAtoms.add(atom);
							addedOriginal = true;
						}
					}
					else{
						adaptedAtoms.add(adoptedAtom);
					}
				}
				if(adaptedAtoms.size() == 0){
					throw new IllegalArgumentException("Atom is not covered by the given Analysers: " + atom);
				}
				else if(adaptedAtoms.size() == 1){
					return adaptedAtoms.first();
				}
				else{
					return new ConjunctiveConditionFormula(adaptedAtoms);
				}
			}
			
		};
	}

	public boolean isConsitent(final ConditionFormula formula){
		ConditionFormula adoptedConditionFormula = formula.replaceAtomicFormula(this.combinedAdaptor);
		return !this.implies(adoptedConditionFormula, ConditionFormula.Bottom);
	}
	
	public boolean isConsitentFlat(final ConjunctiveConditionFormula conjunction){
		FixedArray<ConditionFormula> literals = conjunction.subformulas();
		//boolean[] isCovered = new boolean[literals.size()];
		
		@SuppressWarnings("unchecked")
		ArrayList<AtomicConditionFormula>[] positiveLiterals = (ArrayList<AtomicConditionFormula>[]) new ArrayList<?>[this.simpleAnalysers.size()];
		@SuppressWarnings("unchecked")
		ArrayList<AtomicConditionFormula>[] negativeLiterals = (ArrayList<AtomicConditionFormula>[]) new ArrayList<?>[this.simpleAnalysers.size()];
		for(int simpleAnalyser = 0; simpleAnalyser < this.simpleAnalysers.size(); ++simpleAnalyser){
			positiveLiterals[simpleAnalyser] = new ArrayList<AtomicConditionFormula>(literals.size());
			negativeLiterals[simpleAnalyser] = new ArrayList<AtomicConditionFormula>(literals.size());
		}
		
		for(int literalIndex = 0; literalIndex < literals.size(); ++literalIndex){
			final ConditionFormula literal = literals.get(literalIndex).flaten();
			assert literal.ID != CONJUNCTIVE;
			assert literal.ID != DISJUNCTIVE;
			if(literal.ID == TOP){
				
			}
			else if(literal.ID == BOTTOM){
				return false; //is already inconsistent
			}
			else if(literal.ID == ATOMIC){
				AtomicConditionFormula atom = (AtomicConditionFormula) literal;
				for(int simpleAnalyser = 0; simpleAnalyser < this.simpleAnalysers.size(); ++simpleAnalyser){
					if(this.simpleAnalysers.get(simpleAnalyser).postAdaptFocus().accepts(atom)){
						positiveLiterals[simpleAnalyser].add(atom);
					}						
				}
			}
			else{
				assert literal.ID == NEGATED;
				assert ((NegatedConditionFormula)literal).subformula().ID == ATOMIC;
				AtomicConditionFormula atom = (AtomicConditionFormula) ((NegatedConditionFormula)literal).subformula();
				
				for(int simpleAnalyser = 0; simpleAnalyser < this.simpleAnalysers.size(); ++simpleAnalyser){
					if(this.simpleAnalysers.get(simpleAnalyser).postAdaptFocus().accepts(atom)){
						negativeLiterals[simpleAnalyser].add(atom);
					}						
				}
			}
		}
		
		for(int simpleAnalyser = 0; simpleAnalyser < this.simpleAnalysers.size(); ++simpleAnalyser){
			if(!this.simpleAnalysers.get(simpleAnalyser).isConsistent(positiveLiterals[simpleAnalyser], negativeLiterals[simpleAnalyser])){
				return false;
			}
		}
		return true;
	}
	
	@SuppressWarnings("unchecked")
	public boolean implies(final ConditionFormula formula1, final ConditionFormula formula2){
		ConditionFormula adoptedConditionFormula1 = formula1.replaceAtomicFormula(this.combinedAdaptor);
		final boolean isBottom2 = (formula2 == Bottom);
		ConditionFormula negatedAdoptedConditionFormula2 = (isBottom2 ? Top : new NegatedConditionFormula(formula2.replaceAtomicFormula(this.combinedAdaptor)));		
		for(ConjunctiveConditionFormula conjunction1: adoptedConditionFormula1.DNFiteration()){
			for(ConjunctiveConditionFormula conjunction2: negatedAdoptedConditionFormula2.DNFiteration()){
				ConjunctiveConditionFormula conjunction;
				if(isBottom2){
					conjunction = conjunction1;
				}
				else{
					conjunction = new ConjunctiveConditionFormula(new FixedComposedCollection<ConditionFormula>(conjunction1.subformulas(), conjunction2.subformulas()));
					//TODO better implementation
				}
				
				//TODO better implementation
				if(this.isConsitentFlat(conjunction)){
					return false;
				}
			}
		}
		return true;
	}

}
