package de.lmu.ifi.pms.cep.tsa.optimization;

import static de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonOperatorID.LESS_EQUAL;
import static de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula.*;
import static de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormulaID.*;
import static de.lmu.ifi.pms.cep.tsa.conditions.Term.attr;
import static de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID.IDENTIFIER;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.event_mill.configuration.DEFAULTS;
import de.lmu.ifi.pms.cep.tsa.DTSA_BufferType;
import de.lmu.ifi.pms.cep.tsa.DTSA_CompositeExpression;
import de.lmu.ifi.pms.cep.tsa.DTSA_Expression;
import de.lmu.ifi.pms.cep.tsa.DTSA_Program;
import de.lmu.ifi.pms.cep.tsa.DTSA_Query;
import de.lmu.ifi.pms.cep.tsa.DTSA_QueryType;
import de.lmu.ifi.pms.cep.tsa.DTSA_StreamBuffer;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.compiler.TSA_DefaultDifferencer;
import de.lmu.ifi.pms.cep.tsa.conditions.AtomicConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.AtomicStreamBoundFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.BasicTypeAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.Parameter;
import de.lmu.ifi.pms.cep.tsa.conditions.StreamBound;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_Exception;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_RuntimeException;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_SchemaException;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_SelectionOperator;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchema;
import de.lmu.ifi.pms.cep.tsa.schema.Type;

import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.collection.array.Arrays;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;
import de.lmu.ifi.pms.java.util.map.fixed.FixedEnumMap;


public class DTSA_GarbageCollector extends DTSA_Optimizer {
	
	@Override
	public DTSA_Program optimize(final DTSA_Program program) {
		try {
			return this.optimizeImpl(program);
		} catch (TSA_Exception e) {
			throw new TSA_RuntimeException(e);
		}
	}

	public DTSA_Program optimizeImpl(final DTSA_Program program) throws TSA_Exception {
		
		final FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>> streamBuffers = program.streamBuffers;	
		final FixedEnumMap<DTSA_QueryType, FixedArray<DTSA_Query>> queries = program.queries;
		
		@SuppressWarnings("unchecked")
		final ArrayList<String>[] newStreamBufferNamesListArray = (ArrayList<String>[]) new ArrayList<?>[DTSA_BufferType.values().length];
		@SuppressWarnings("unchecked")
		final ArrayList<DTSA_StreamBuffer>[] newStreamBuffersListArray = (ArrayList<DTSA_StreamBuffer>[]) new ArrayList<?>[newStreamBufferNamesListArray.length];
		for(int bufferType = 0; bufferType < newStreamBufferNamesListArray.length; ++bufferType){
			newStreamBufferNamesListArray[bufferType] = new ArrayList<String>();
			newStreamBuffersListArray[bufferType] = new ArrayList<DTSA_StreamBuffer>();
		}
		@SuppressWarnings("unchecked")
		final ArrayList<DTSA_Query>[] newQueriesListArray = (ArrayList<DTSA_Query>[]) new ArrayList<?>[DTSA_QueryType.values().length];
		for(int queryType = 0; queryType < newQueriesListArray.length; ++queryType){
			newQueriesListArray[queryType] = new ArrayList<DTSA_Query>();
		}
		
		//generate new buffers and queries, copy old buffers
		@SuppressWarnings("unchecked")
		FixedArrayNavigableMap<String, DTSA_StreamBuffer>[] newStreamBuffersArray = (FixedArrayNavigableMap<String, DTSA_StreamBuffer>[]) new FixedArrayNavigableMap<?,?>[newStreamBufferNamesListArray.length];
		@SuppressWarnings("unchecked")
		FixedArray<DTSA_Query>[] newQueriesArray = (FixedArray<DTSA_Query>[]) new FixedArray<?>[newQueriesListArray.length];
		
		@SuppressWarnings("unchecked")
		FixedEnumMap<DTSA_QueryType, ArrayList<DTSA_Query>> garbageQueries = new FixedEnumMap<DTSA_QueryType, ArrayList<DTSA_Query>>(DTSA_QueryType.GARBAGE_STATIC, DTSA_QueryType.GARBAGE_EVAL, new ArrayList<DTSA_Query>(), new ArrayList<DTSA_Query>(), new ArrayList<DTSA_Query>());
		
		for(DTSA_BufferType bufferType: DTSA_BufferType.values()){
			if(bufferType == DTSA_BufferType.BLACK_HOLE_INPUT){//generate black hole buffer for each input buffer
				FixedArrayNavigableMap<String, DTSA_StreamBuffer> buffers = streamBuffers.get(DTSA_BufferType.INPUT);
				final String[] newBlackHoleBufferNamesArray = new String[buffers.size()];
				final DTSA_StreamBuffer[] newBlackHoleBuffersArray = new DTSA_StreamBuffer[newBlackHoleBufferNamesArray.length];
				@SuppressWarnings("unchecked")
				final ArrayList<DTSA_Query>[] inputQueries = (ArrayList<DTSA_Query>[]) new ArrayList<?>[buffers.size()];
				
				for(int bufferIndex = 0; bufferIndex < inputQueries.length; ++bufferIndex){	
					inputQueries[bufferIndex] = new ArrayList<DTSA_Query>();
				}
				
				for(DTSA_QueryType queryType: new DTSA_QueryType[]{DTSA_QueryType.STATIC, DTSA_QueryType.INIT, DTSA_QueryType.INPUT}){
					for(DTSA_Query query: queries.get(queryType)){
						for(DTSA_StreamBuffer buffer: query.inputs.values()){
							if(buffer.type == DTSA_BufferType.INPUT){
								assert buffers.keys().contains(buffer.full_name);
								int bufferIndex = buffers.keys().indexOf(buffer.full_name);
								inputQueries[bufferIndex].add(query);
							}
						}
					}
				}
				
				for(int bufferIndex = 0; bufferIndex < newBlackHoleBufferNamesArray.length; ++bufferIndex){		
					DTSA_StreamBuffer oldBuffer = buffers.values().get(bufferIndex);
					DTSA_StreamBuffer newBlackHoleBuffer = this.createBlackHoleBuffer(oldBuffer);
					newBlackHoleBufferNamesArray[bufferIndex] = newBlackHoleBuffer.full_name;
					newBlackHoleBuffersArray[bufferIndex] = newBlackHoleBuffer;
					DTSA_Query newGarbageQuery = this.createGarbageQuery(newBlackHoleBuffer, oldBuffer, inputQueries[bufferIndex]);
					if(newGarbageQuery != null){
						garbageQueries.get(newGarbageQuery.type).add(newGarbageQuery);
					}
				}
				
				assert streamBuffers.get(DTSA_BufferType.BLACK_HOLE_INPUT).isEmpty();
				newStreamBuffersArray[bufferType.ordinal()] = new FixedArrayNavigableMap<String, DTSA_StreamBuffer>(newBlackHoleBufferNamesArray, newBlackHoleBuffersArray);			
			}
			else if(bufferType == DTSA_BufferType.BLACK_HOLE_WORKING){//generate black hole buffer for each working buffer				
				FixedArrayNavigableMap<String, DTSA_StreamBuffer> buffers = streamBuffers.get(DTSA_BufferType.WORKING);
				final String[] newBlackHoleBufferNamesArray = new String[buffers.size()];
				final DTSA_StreamBuffer[] newBlackHoleBuffersArray = new DTSA_StreamBuffer[newBlackHoleBufferNamesArray.length];
				@SuppressWarnings("unchecked")
				final ArrayList<DTSA_Query>[] inputQueries = (ArrayList<DTSA_Query>[]) new ArrayList<?>[buffers.size()];
				
				for(int bufferIndex = 0; bufferIndex < inputQueries.length; ++bufferIndex){	
					inputQueries[bufferIndex] = new ArrayList<DTSA_Query>();
				}
				
				for(DTSA_QueryType queryType: new DTSA_QueryType[]{DTSA_QueryType.STATIC, DTSA_QueryType.INIT, DTSA_QueryType.WORKING, DTSA_QueryType.OUTPUT, DTSA_QueryType.LOG}){
					for(DTSA_Query query: queries.get(queryType)){
						for(DTSA_StreamBuffer buffer: query.inputs.values()){
							if(buffer.type == DTSA_BufferType.WORKING){
								assert buffers.keys().contains(buffer.full_name);
								int bufferIndex = buffers.keys().indexOf(buffer.full_name);
								inputQueries[bufferIndex].add(query);
							}
						}
					}
				}
				
				for(int bufferIndex = 0; bufferIndex < newBlackHoleBufferNamesArray.length; ++bufferIndex){		
					DTSA_StreamBuffer oldBuffer = buffers.values().get(bufferIndex);
					DTSA_StreamBuffer newBlackHoleBuffer = this.createBlackHoleBuffer(oldBuffer);
					newBlackHoleBufferNamesArray[bufferIndex] = newBlackHoleBuffer.full_name;
					newBlackHoleBuffersArray[bufferIndex] = newBlackHoleBuffer;
					DTSA_Query newGarbageQuery = this.createGarbageQuery(newBlackHoleBuffer, oldBuffer, inputQueries[bufferIndex]);
					if(newGarbageQuery != null){
						garbageQueries.get(newGarbageQuery.type).add(newGarbageQuery);
					}
				}
				
				assert streamBuffers.get(DTSA_BufferType.BLACK_HOLE_WORKING).isEmpty();
				newStreamBuffersArray[bufferType.ordinal()] = new FixedArrayNavigableMap<String, DTSA_StreamBuffer>(newBlackHoleBufferNamesArray, newBlackHoleBuffersArray);
			}
			else{//copy for other buffer's types
				newStreamBuffersArray[bufferType.ordinal()] = streamBuffers.get(bufferType);				
			}			
		}
		
		@SuppressWarnings("unchecked")
		final FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>> newStreamBuffers = new FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>>(DTSA_BufferType.class, newStreamBuffersArray);
		
		//copy old queries
		for(DTSA_QueryType queryType: DTSA_QueryType.values()){
			if(queryType.isGarbageQuery){
				newQueriesArray[queryType.ordinal()] = new FixedArray<DTSA_Query>(garbageQueries.get(queryType));
			}
			else{
				newQueriesArray[queryType.ordinal()] = queries.get(queryType);
			}
		}
		
		@SuppressWarnings("unchecked")
		final FixedEnumMap<DTSA_QueryType, FixedArray<DTSA_Query>> newQueries = new FixedEnumMap<DTSA_QueryType, FixedArray<DTSA_Query>>(DTSA_QueryType.class, newQueriesArray);
		
		try {
			return new DTSA_Program(program.namespace, program.simple_name, program.description, newStreamBuffers, newQueries, program.origin);
		} catch (TSA_Exception e) {
			throw new TSA_RuntimeException(e);
		}
	}

	public DTSA_StreamBuffer createBlackHoleBuffer(final DTSA_StreamBuffer streamBuffer) throws TSA_Exception {
		if(streamBuffer.type == DTSA_BufferType.INPUT){
			return new DTSA_StreamBuffer(streamBuffer.full_name, "BLACK_HOLE_INPUT for" + streamBuffer.full_name, DTSA_BufferType.BLACK_HOLE_INPUT, streamBuffer.schema, new FixedArrayNavigableMap<String, Term<? extends Type>>(), streamBuffer.origin);
		}
		else if (streamBuffer.type == DTSA_BufferType.WORKING) {
			return new DTSA_StreamBuffer(streamBuffer.full_name, "BLACK_HOLE_WORKING for" + streamBuffer.full_name, DTSA_BufferType.BLACK_HOLE_WORKING, streamBuffer.schema, new FixedArrayNavigableMap<String, Term<? extends Type>>(), streamBuffer.origin);			
		}
		else{
			throw new IllegalArgumentException("BLACK_HOLE buffers can be created only for INPUT or WORKING buffers");
		}
	}
	
	public ConditionFormula garbageFormula(final DTSA_StreamBuffer oldBuffer, final FixedArrayNavigableMap<String, Term<BasicType>> relevanceBoundFunctions, final boolean readByEvalQuery) throws TSA_Exception{
		ConditionFormula outputSBF;
		try{
			outputSBF = oldBuffer.schema.streamBoundFormula();
		} catch (TSA_SchemaException e) {
			throw new TSA_Exception(oldBuffer.origin, e);
		}
		assert TemporalStreamSchema.isValidBufferStreamBoundFormula(outputSBF);
		
		final ConditionFormula upperBoundFormula;
		
		if(oldBuffer.isStatic){
			if(readByEvalQuery){
				upperBoundFormula = Bottom;
			}
			else{
				upperBoundFormula = Top;
			}
		}		
		else if(oldBuffer.type == DTSA_BufferType.INPUT){
			Parameter upperBound = new Parameter("input.now:" + TSA_SyntaxPatterns.streamBoundQualifiedName(oldBuffer.full_name, DEFAULTS._o_block), IDENTIFIER.type);			
			upperBoundFormula = new BinaryComparissonFormula(LESS_EQUAL, attr(DEFAULTS._o_block, IDENTIFIER), upperBound);
		}
		else{
			assert oldBuffer.type == DTSA_BufferType.WORKING;
			upperBoundFormula = outputSBF.replaceAtomicFormula(new Adapter<AtomicConditionFormula, ConditionFormula>() {			
				@Override
				public ConditionFormula adapt(final AtomicConditionFormula source) {
					if(source instanceof AtomicStreamBoundFormula){
						AtomicStreamBoundFormula sbf = (AtomicStreamBoundFormula) source;
						if(sbf.atomicTerm instanceof BasicTypeAttribute){
							assert sbf.bound instanceof StreamBound;
							StreamBound streamBound = (StreamBound) sbf.bound;
	
							if(relevanceBoundFunctions.has(streamBound.qualified_name)){
								Term<BasicType> relevanceBoundFunction = relevanceBoundFunctions.get(streamBound.qualified_name);
								if(relevanceBoundFunction instanceof ConditionFormula){
									//FIXME XML replace by relevanceBoundFunction == Top
									if(((ConditionFormula)relevanceBoundFunction).ID == TOP){
										return Top;
									}
									else{
										assert ((ConditionFormula)relevanceBoundFunction).ID == BOTTOM;
										return Bottom;
									}
								}
								else{
									Parameter upperBound = new Parameter("input.now:" + streamBound.qualified_name, sbf.bound.type);
									return new BinaryComparissonFormula(LESS_EQUAL, sbf.atomicTerm, upperBound);
								}
							}
							return Bottom;
						}
					}				
					throw new Error();
				}			
			}).flaten();
		}

		
		if(upperBoundFormula.flaten().ID == BOTTOM & !(oldBuffer.isStatic & readByEvalQuery)){
			//FIXME throw new TSA_Exception(oldBuffer.origin, "The temporal relations are too weak to derive an garbage collection expression for the non-static buffer \"" + oldBuffer.type + " " + oldBuffer.full_name + "\".");
			System.err.println(new TSA_Exception(oldBuffer.origin, "The temporal relations are too weak to derive an garbage collection expression for the non-static buffer \"" + oldBuffer.type + " " + oldBuffer.full_name + "\".").getMessage());
		}
		
		return upperBoundFormula;
	}
	
	public DTSA_Query createGarbageQuery(final DTSA_StreamBuffer newBuffer, final DTSA_StreamBuffer oldBuffer, final ArrayList<DTSA_Query> queries) throws TSA_Exception {
		
		if(oldBuffer.type == DTSA_BufferType.WORKING | oldBuffer.type == DTSA_BufferType.INPUT){
			if(queries.size() == 0){
				//FIXME
				System.out.println("There is no query using buffer \"" + oldBuffer.type + " " + oldBuffer.full_name + "\" as input. Apparently the buffer is not required.");
				System.out.println(oldBuffer.origin);
			}
		}
			
		@SuppressWarnings("unchecked")
		FixedArray<Term<BasicType>>[] relevanceBoundFunctionsArray = (FixedArray<Term<BasicType>>[]) new FixedArray<?>[queries.size()];
		
		@SuppressWarnings("unchecked")
		FixedArray<Term<BasicType>>[] relevanceBoundClockIncreaseFunctionsArray = (FixedArray<Term<BasicType>>[]) new FixedArray<?>[queries.size()];
		
		boolean readsStatic = false;
		boolean readsInit = false;
		boolean readsEval = false;		

		for(int i = 0; i < queries.size(); ++i){
			DTSA_Query query = queries.get(i);
			switch (query.type) {
				case STATIC:
					readsStatic = true;
					break;
				case INIT:
					readsInit = true;
					break;
				case INPUT:
					readsEval = true;
					break;
				case WORKING:
					readsEval = true;
					break;
				case OUTPUT:
					readsEval = true;
					break;
				case LOG:
					readsEval = true;
					break;	
				default:
					throw new Error();
			}
			
			int firstIndexInclusive = query.relevanceBoundFunctions.keys().ceilingIndex(oldBuffer.full_name + ":");
			int lastIndexExclusive = query.relevanceBoundFunctions.keys().ceilingIndex(oldBuffer.full_name + ":|");
			assert lastIndexExclusive - firstIndexInclusive == oldBuffer.streamBounds.size();
			relevanceBoundFunctionsArray[i] = query.relevanceBoundFunctions.values().subList(firstIndexInclusive, lastIndexExclusive);
			relevanceBoundClockIncreaseFunctionsArray[i] = query.relevanceBoundClockIncreaseFunctions.values().subList(firstIndexInclusive, lastIndexExclusive);
			assert relevanceBoundFunctionsArray[i].size() == oldBuffer.streamBounds.size();
		}
		
		FixedArrayNavigableMap<String, StreamBound> streamBounds = oldBuffer.streamBounds;
		
		@SuppressWarnings("unchecked")
		Term<BasicType>[] streamBoundClockIncreaseFunctionFormulas = (Term<BasicType>[]) new Term<?>[streamBounds.size()];
		Arrays.fill(streamBoundClockIncreaseFunctionFormulas, de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula.Bottom);
		FixedArrayNavigableMap<String, Term<BasicType>> streamBoundClockIncreaseFunctions = new FixedArrayNavigableMap<String, Term<BasicType>>(streamBounds.keys(), new FixedArray<Term<BasicType>>(false, streamBoundClockIncreaseFunctionFormulas));
		
		@SuppressWarnings("unchecked")
		Term<BasicType>[] relevanceBoundFunctionFormulas = (Term<BasicType>[]) new Term<?>[streamBounds.size()];
		Arrays.fill(relevanceBoundFunctionFormulas, de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula.Bottom);
		FixedArrayNavigableMap<String, Term<BasicType>> relevanceBoundFunctions = new FixedArrayNavigableMap<String, Term<BasicType>>(streamBounds.keys(), new FixedArray<Term<BasicType>>(false, relevanceBoundFunctionFormulas));
		
		@SuppressWarnings("unchecked")
		Term<BasicType>[] relevanceBoundClockIncreaseFunctionFormulas = (Term<BasicType>[]) new Term<?>[streamBounds.size()];
		Arrays.fill(relevanceBoundClockIncreaseFunctionFormulas, de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula.Bottom);
		FixedArrayNavigableMap<String, Term<BasicType>> relevanceBoundClockIncreaseFunctions = new FixedArrayNavigableMap<String, Term<BasicType>>(streamBounds.keys(), new FixedArray<Term<BasicType>>(false, relevanceBoundClockIncreaseFunctionFormulas));
		
		@SuppressWarnings("unchecked")
		Term<BasicType>[] minRelevanceBoundFunctionsArray = (Term<BasicType>[]) new Term<?>[streamBounds.size()];
		
		for(int i = 0; i < minRelevanceBoundFunctionsArray.length; ++i){
			@SuppressWarnings("unchecked")
			Term<BasicType>[] minPosition = (Term<BasicType>[]) new Term<?>[relevanceBoundFunctionsArray.length];
			for(int k = 0; k < minPosition.length; ++k){
				minPosition[k] = TSA_DefaultDifferencer.maxBoundFunctions(relevanceBoundFunctionsArray[k].get(i), relevanceBoundClockIncreaseFunctionsArray[k].get(i));
			}
			minRelevanceBoundFunctionsArray[i] = TSA_DefaultDifferencer.minBoundFunctions(minPosition);
			assert minRelevanceBoundFunctionsArray[i] != null;
		}
		
		//garbageCondition = incrementFormula(ConditionFormula outputSBF, FixedArrayNavigableMap<String, Term<BasicType>> outputStreamBoundFunctions, FixedCollection<String> origin),
		//wobei outputStreamBoundFunctions
		ConditionFormula garbageCondition = this.garbageFormula(oldBuffer, new FixedArrayNavigableMap<String, Term<BasicType>>(oldBuffer.streamBounds.keys(), new FixedArray<Term<BasicType>>(false, minRelevanceBoundFunctionsArray)), readsEval);
		
		TSA_SelectionOperator chooseIrrelevantTuples = new TSA_SelectionOperator(newBuffer.origin, garbageCondition);

		DTSA_Expression expression = new DTSA_CompositeExpression(chooseIrrelevantTuples, oldBuffer);
		
		String description;
		if(oldBuffer.type == DTSA_BufferType.INPUT){
			description = "Garbage query for the input buffer" + oldBuffer.full_name;
		}
		else if (oldBuffer.type == DTSA_BufferType.WORKING) {
			description = "Garbage query for the working buffer" + oldBuffer.full_name;			
		}
		else{
			throw new IllegalArgumentException("Garbage queries can be created only for input and working buffers");
		}
		
		DTSA_QueryType garbageQueryType;
		if(readsEval){
			garbageQueryType = DTSA_QueryType.GARBAGE_EVAL;
		}
		else if(readsInit){
			garbageQueryType = DTSA_QueryType.GARBAGE_INIT;
		}
		else if(readsStatic){
			garbageQueryType = DTSA_QueryType.GARBAGE_STATIC;
		}
		else{
			//unused buffer
			garbageQueryType = DTSA_QueryType.GARBAGE_STATIC;
		}
		
		if(garbageCondition != Bottom){
			return new DTSA_Query(newBuffer.prog_package, expression, newBuffer, garbageQueryType, streamBounds, streamBoundClockIncreaseFunctions, relevanceBoundFunctions, relevanceBoundClockIncreaseFunctions, description, newBuffer.origin);
		}
		else{
			return null;
		}
	}
	
	protected static final Logger logger = LoggerFactory.getLogger(DTSA_GarbageCollector.class);
	
	@Override
	protected Logger logger(){
		return logger;
	}

}
