package de.lmu.ifi.pms.cep.tsa.schema;

import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementMap;
import org.simpleframework.xml.core.Resolve;

import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.conditions.AtomicTerm;
import de.lmu.ifi.pms.cep.tsa.conditions.Attribute;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.ConjunctiveConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.conditions.TypedAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.adaptor.PrefixAttributes;
import de.lmu.ifi.pms.cep.tsa.conditions.filter.AttributeFilter;
import de.lmu.ifi.pms.cep.tsa.conditions.filter.VariableFilter;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_TypeException;

import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.filter.Filter;
import de.lmu.ifi.pms.java.util.map.Map;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;
import de.lmu.ifi.pms.java.util.tupel.Pair;

public class CompositeType extends NamelessType {
	
	
	public final FixedArrayNavigableMap<String, Type> attributes;
	
	@ElementMap(name="attributes", entry="attribute", key="name", value="type")
	java.util.HashMap<String, Type> getAttributes(){
		return new java.util.HashMap<String, Type>(this.attributes.javaMap());
	}
	
	final FixedArray<ConditionFormula> tupleConstraints;
	
	@Element(name="tuple-constraints")
	ConjunctiveConditionFormula getTupleConstraintFormula(){
		return ConditionFormula.and(this.tupleConstraints);
	}
	
	CompositeType() {
		super(TypeID.COMPOSITE);
		this.attributes = new FixedArrayNavigableMap<String, Type>();
		this.tupleConstraints = new FixedArray<ConditionFormula>();
	}
	
	CompositeType(@ElementMap(name="attributes", entry="attribute", key="name", value="type") final java.util.HashMap<String, Type> attributes
			, @Element(name="tuple-constraints") final ConjunctiveConditionFormula tupleConstraints) throws TSA_TypeException {
		this(new FixedArrayNavigableMap<String, Type>(attributes), tupleConstraints.subformulas());
	}

	CompositeType(final FixedArrayNavigableMap<String, Type> attributes, final FixedArray<ConditionFormula> tupleConstraints) throws TSA_TypeException {
		super(TypeID.COMPOSITE);
	
		this.attributes = attributes;
		this.checkAttributeNames(attributes.keys());
		
		this.tupleConstraints = tupleConstraints;
		this.checkTupleConstraints(tupleConstraints);
		
	}
	
	@Resolve
	CompositeType resolve(){
		try{
			return get(this.attributes, this.tupleConstraints);
		}
		catch (Exception e) {
			throw new Error("Should not be possible.");
		}
	}
	
	public static CompositeType get(final java.util.Map<String, ? extends Type> attributes) throws TSA_TypeException {
		return get(attributes, new FixedArray<ConditionFormula>());
	}
	
	public static CompositeType get(final java.util.Map<String, ? extends Type> attributes, final FixedArray<ConditionFormula> tupleConstraints)  throws TSA_TypeException {
		return get(new FixedArrayNavigableMap<String, Type>(attributes), tupleConstraints);
	}
	
	public static CompositeType get(final Map<String, ? extends Type, ?> attributes, final FixedArray<ConditionFormula> tupleConstraints) throws TSA_TypeException {
		return get(new FixedArrayNavigableMap<String, Type>(attributes), tupleConstraints);
	}
	
	public static CompositeType get(final FixedArrayNavigableMap<String, Type> attributes, final FixedArray<ConditionFormula> tupleConstraints) throws TSA_TypeException {
		return new CompositeType(attributes, tupleConstraints);
		//TODO cache and return unique object;
	}
	
	public static CompositeType get(final String attribute_name, final Type attribute_type) throws TSA_TypeException {
		return get(new FixedArrayNavigableMap<String, Type>(new FixedArrayNavigableSet<String>(false, false, attribute_name), new FixedArray<Type>(false, attribute_type)));
	}
	
	public static CompositeType get(final FixedArrayNavigableMap<String, ? extends Type> attributes) throws TSA_TypeException {		
		return new CompositeType().extend(attributes);
	}
	
	void checkAttributeNames(final FixedArrayNavigableSet<String> names) throws TSA_TypeException{
		assert names.size() > 0;
		//check valid names
		for(String name: names){
			if(!TSA_SyntaxPatterns.isValidAttributeSimpleName(name)){
				throw new TSA_TypeException("Invalid simple attribute name: " + name);
				//TODO better exception
			}
		}
	}
	
	void checkTupleConstraints(final FixedArray<ConditionFormula> tupleConstraints) throws TSA_TypeException{
		assert this.attributes != null;
		for(ConditionFormula constraint: tupleConstraints){
			for(Term<?> term: constraint.collect(AttributeFilter.instance)){
				Attribute<?> attribute = (Attribute<?>) term;
				//check that only available attributes are used
				if(!this.hasAttribute(attribute.name)){
					throw new TSA_TypeException("Constraint " + constraint + " uses unavailable attribute " + attribute.name); 
				}				
				//FIXME check that attributes have same/compatible type
			}
			//constraints must not contain variables
			if(!constraint.collect(VariableFilter.instance).isEmpty()){
				throw new TSA_TypeException("Constraints must not use variables.");
			}
		}
	}
	
	public boolean hasAttribute(final String name){
		try{
			this.typeOf(name);
			return true;
		}
		catch(TSA_TypeException e){
			return false;
		}
	}
	
	public Pair<Object, String, Type> getAttribute(final String name)  throws TSA_TypeException{
		return new Pair<Object, String, Type>(name, this.typeOf(name));
	}
	
	protected CompositeType extend(final FixedArrayNavigableMap<String, ? extends Type> newAttributes) throws TSA_TypeException {
		if(newAttributes.size() > 0){
			final String[][] newAttributeNames = new String[newAttributes.size()][];
			final Type[] newAttributeTypes = new Type[newAttributeNames.length];
			for(int attribute = 0; attribute < newAttributeNames.length; ++attribute){
				String newName = newAttributes.keys().get(attribute);
				if(newName.equals("")){
					newAttributeNames[attribute] = new String[0];
				}
				else{
					newAttributeNames[attribute] = TSA_SyntaxPatterns.splitAttributeName(newAttributes.keys().get(attribute));
				}
				newAttributeTypes[attribute] = newAttributes.values().get(attribute);
			}
			return this.extend(newAttributeNames, newAttributeTypes, 0, newAttributeNames.length, 0);
			}
		else{
			return this;
		}
	}
	
	@Override
	protected CompositeType extend(final String[][] newAttributeNames, final Type[] newAttributeTypes, final int startIndex, final int endIndex, final int level) throws TSA_TypeException {
		assert startIndex <= endIndex;
		assert endIndex <= newAttributeNames.length;
		assert newAttributeNames.length == newAttributeTypes.length;
		//all new attributes between startIndex and EndIndex should be inserted into this type or as attributes or as attributes of a subtype of this type
		//this means their name up to level (exclusive) should be the same as name of this type seen as attribute of the root type 
		if(startIndex < endIndex){
			assert newAttributeNames[startIndex].length >= level;
			if(newAttributeNames[startIndex].length == level){//at least one of the new attributes is inserted into this type
				int newAttributesIndex = startIndex;
				ArrayList<String> newLocalDirectAttributeNamesList = new ArrayList<String>();
				ArrayList<Type> newLocalDirectAttributeTypesList = new ArrayList<Type>();
				while(newAttributesIndex < endIndex && newAttributeNames[newAttributesIndex].length == level){
					//the name of the new attribute maches the name of this type with respect to the root type
					//i.e. all attributes defined in the type of the new attribute (type must be composite, named composite) should be added to this type
					Type newAttributeType = newAttributeTypes[newAttributesIndex];
					if(newAttributeType.ID == TypeID.NAMED){
						newAttributeType = ((NamedType<?>)newAttributeType).type;//the attributes of the named type become part of the newly created composite type, thus the name does not apply anymore
					}
					assert newAttributeType.ID != TypeID.NAMED;
					CompositeType newAttributeCompositeType;
					
					switch (newAttributeType.ID) {
					case BASIC: {
						throw new TSA_TypeException("There can not be a basic type attribute with the same name as a composite type attribute: " + TSA_SyntaxPatterns.attributeName(newAttributeNames[newAttributesIndex]));
					}
					case COMPOSITE: {
						newAttributeCompositeType = ((CompositeType)newAttributeType);
						break;
					}
					default:
						throw new Error("Unexpected case: " + newAttributeType.ID);
					}					
					
					FixedArrayNavigableMap<String, Type> attributesInCompositeType = newAttributeCompositeType.attributes;
					int numberOfAttributesInCompositeType = attributesInCompositeType.size();
					for(int attribute = 0; attribute < numberOfAttributesInCompositeType; ++attribute){//i.e. all attributes defined in the type of the new attribute should be added to this type
						newLocalDirectAttributeNamesList.add(attributesInCompositeType.keys().get(attribute));
						newLocalDirectAttributeTypesList.add(attributesInCompositeType.values().get(attribute));
					}
					
					++newAttributesIndex;
				}
				
				String[] newLocalDirectAttributeNamesArray = newLocalDirectAttributeNamesList.toArray(new String[newLocalDirectAttributeNamesList.size()]);
				int[] newLocalDirectAttributeNameSortedPositions = new int[newLocalDirectAttributeNamesArray.length];
				de.lmu.ifi.pms.java.util.collection.array.Arrays.sort(newLocalDirectAttributeNamesArray, newLocalDirectAttributeNameSortedPositions, true);
				
				final String[][] newLocalDirectAttributeNames = new String[newLocalDirectAttributeNamesArray.length][]; 
				final Type[] newLocalDirectAttributeTypes = new Type[newLocalDirectAttributeNames.length];
				
				for(int newLocalDirectAttributeOriginalPosition = 0; newLocalDirectAttributeOriginalPosition < newLocalDirectAttributeNames.length; ++newLocalDirectAttributeOriginalPosition){
					int sortedPosition = newLocalDirectAttributeNameSortedPositions[newLocalDirectAttributeOriginalPosition];
					newLocalDirectAttributeNames[sortedPosition] = new String[]{newLocalDirectAttributeNamesArray[sortedPosition]};
					newLocalDirectAttributeTypes[sortedPosition] = newLocalDirectAttributeTypesList.get(newLocalDirectAttributeOriginalPosition);
				}

				CompositeType newLocalType = this.extend(newLocalDirectAttributeNames, newLocalDirectAttributeTypes, 0, newLocalDirectAttributeNames.length, 0);//recursive call with this type as new root type
				
				return newLocalType.extend(newAttributeNames, newAttributeTypes, newAttributesIndex, endIndex, level);
			}
			else{//all new attributes between startIndex and EndIndex should become attributes or as attributes of a subtype of the current type
				assert newAttributeNames[startIndex].length > level;
				int localIndex = 0;//index for already existing attributes of this type
							
				int newAttributesIndex = startIndex;//index for new attributes that should be inserted 
				ArrayList<String> extendedAttributeNames = new ArrayList<String>();
				ArrayList<Type> extendedAttributeTypes = new ArrayList<Type>();
				String newAttribute = newAttributeNames[newAttributesIndex][level];
				if(localIndex < this.attributes.size()){//tests if this type has at least attibutes, if not the merge can be omited
					{
						String localAttribute = this.attributes.keys().get(localIndex);
						int cmp;
						while(true){//doing a modified sorted merge
							assert newAttributesIndex < endIndex;
							cmp = localAttribute.compareTo(newAttribute);
							if(cmp < 0){//existing attribute is not extended
								extendedAttributeNames.add(localAttribute);
								extendedAttributeTypes.add(this.attributes.values().get(localIndex));
								++localIndex;
								if(localIndex < this.attributes.size()){
									localAttribute = this.attributes.keys().get(localIndex);
								}
								else{
									break;
								}
							}
							else if(cmp == 0){//existing attribute is extended by new attributes
								final int subStartIndex = newAttributesIndex;
								assert subStartIndex < endIndex;
								while(localAttribute.equals(newAttribute)){//true at least once, as cmp == 0;
								//collect all new attributes that extend the current existing attribute
									++newAttributesIndex;
									if(newAttributesIndex < endIndex){
										newAttribute = newAttributeNames[newAttributesIndex][level];
									}
									else{
										break;
									}
								}						
								int subEndIndex = newAttributesIndex;
															
								assert startIndex < subEndIndex;
								Type subType = this.attributes.values().get(localIndex);
								Type newSubType = subType.extend(newAttributeNames, newAttributeTypes, subStartIndex, subEndIndex, level + 1);
	
								extendedAttributeNames.add(localAttribute);
								extendedAttributeTypes.add(newSubType);
	
								++localIndex;
								if(newAttributesIndex < endIndex & localIndex < this.attributes.size()){
									localAttribute = this.attributes.keys().get(localIndex);
								}
								else{
									break;
								}
							}
							else{assert cmp > 0;//new attributes do not extend an existing attriute
								final int subStartIndex = newAttributesIndex;
								assert subStartIndex < endIndex;
								while(newAttributesIndex < endIndex && newAttributeNames[newAttributesIndex][level].equals(newAttribute)){//true at least once;
									++newAttributesIndex;
								}						
								int subEndIndex = newAttributesIndex;
								
								assert startIndex < subEndIndex;
								
								Type newAttributeType;
								if(subEndIndex - subStartIndex == 1 & level + 1 == newAttributeNames[subStartIndex].length){
									//this case is particularly important for new basic types, but also a more efficient implementation in the other cases
									newAttributeType = newAttributeTypes[subStartIndex];
								}
								else{
									newAttributeType = new CompositeType().extend(newAttributeNames, newAttributeTypes, subStartIndex, subEndIndex, level + 1);
								}
								extendedAttributeNames.add(newAttribute);
								extendedAttributeTypes.add(newAttributeType);	
								
								if(newAttributesIndex < endIndex){
									newAttribute = newAttributeNames[newAttributesIndex][level];
								}
								else{
									break;
								}
							}
						}
					}
				}
				while(newAttributesIndex < endIndex){
					final int subStartIndex = newAttributesIndex;
					assert subStartIndex < endIndex;
					while(newAttributesIndex < endIndex && newAttributeNames[newAttributesIndex][level].equals(newAttribute)){//true at least once;
						++newAttributesIndex;
					}						
					int subEndIndex = newAttributesIndex;
					
					assert startIndex < subEndIndex;
					
					Type newAttributeType;
					if(subEndIndex - subStartIndex == 1 & level + 1 == newAttributeNames[subStartIndex].length){
						//this case is particularly important for new basic types, but also a more efficient implementation in the other cases
						newAttributeType = newAttributeTypes[subStartIndex];
					}
					else{
						newAttributeType = new CompositeType().extend(newAttributeNames, newAttributeTypes, subStartIndex, subEndIndex, level + 1);
					}
					extendedAttributeNames.add(newAttribute);
					extendedAttributeTypes.add(newAttributeType);
					
					if(newAttributesIndex < endIndex){
						newAttribute = newAttributeNames[newAttributesIndex][level];
					}
					else{
						break;
					}
				}
				for(;localIndex < this.attributes.size(); ++localIndex){
					String localAttribute = this.attributes.keys().get(localIndex);
					extendedAttributeNames.add(localAttribute);
					extendedAttributeTypes.add(this.attributes.values().get(localIndex));
				}
				

				return new CompositeType(new FixedArrayNavigableMap<String, Type>(new FixedArrayNavigableSet<String>(false, false, extendedAttributeNames.toArray(new String[0])), new FixedArray<Type>(false, extendedAttributeTypes.toArray(new Type[0]))), new FixedArray<ConditionFormula>());
			}
		}
		else{
			return this;
		}
	}
	
	protected CompositeType retain(final FixedArrayNavigableSet<String> attributes) throws TSA_TypeException {
		final String[][] retainedAttributes = new String[attributes.size()][];
		for(int attribute = 0; attribute < retainedAttributes.length; ++attribute){
			retainedAttributes[attribute] = TSA_SyntaxPatterns.splitAttributeName(attributes.get(attribute));
		}
		CompositeType newType = this.retain(retainedAttributes, 0, retainedAttributes.length, 0);
		assert newType != null;
			
		return newType;
	}
	
	@Override
	protected CompositeType retain(final String[][] retainedAttributes, final int startIndex, final int endIndex, final int level) throws TSA_TypeException {
		if(startIndex < endIndex){
			assert retainedAttributes[startIndex].length >= level;
			if(retainedAttributes[startIndex].length == level){
				return this;
			}
			else{
				assert this.attributes.size() > 0;
				boolean changed = false;
				int localIndex = 0;
				int retainedAttributesIndex = startIndex;
				ArrayList<String> remainingAttributeNames = new ArrayList<String>();
				ArrayList<Type> remainingAttributeTypes = new ArrayList<Type>();
				String localAttribute = this.attributes.keys().get(localIndex);
				String retainedAttribute = retainedAttributes[retainedAttributesIndex][level];
				int cmp;
				while(true){
					cmp = localAttribute.compareTo(retainedAttribute);
					if(cmp < 0){
						++localIndex;
						if(localIndex < this.attributes.size()){
							localAttribute = this.attributes.keys().get(localIndex);
						}
						else{
							break;
						}
					}
					else if(cmp == 0){
						int subStartIndex = retainedAttributesIndex;
						while(localAttribute.equals(retainedAttribute)){//true at least once, as cmp == 0;
							++retainedAttributesIndex;
							if(retainedAttributesIndex < endIndex){
								retainedAttribute = retainedAttributes[retainedAttributesIndex][level];
							}
							else{
								break;
							}
						}
						int subEndIndex = retainedAttributesIndex;
						assert startIndex < subEndIndex;
						Type subType = this.attributes.values().get(localIndex);
						Type newSubType = subType.retain(retainedAttributes, subStartIndex, subEndIndex, level + 1);
						if(newSubType != null){
							remainingAttributeNames.add(localAttribute);
							remainingAttributeTypes.add(newSubType);
							if(!newSubType.equals(subType)){
								changed = true;
							}
						}
						++localIndex;
						if(localIndex < this.attributes.size()){
							localAttribute = this.attributes.keys().get(localIndex);
						}
						else{
							break;
						}
					}
					else{assert cmp > 0;
						++retainedAttributesIndex;
						if(retainedAttributesIndex < endIndex){
							retainedAttribute = retainedAttributes[retainedAttributesIndex][level];
						}
						else{
							break;
						}
					}
				}
				if(remainingAttributeNames.size() < this.attributes.size()){
					changed = true;
				}
				if(changed){
					if(remainingAttributeNames.size() > 0){
						return new CompositeType(new FixedArrayNavigableMap<String, Type>(new FixedArrayNavigableSet<String>(false, false, remainingAttributeNames.toArray(new String[0])), new FixedArray<Type>(false, remainingAttributeTypes.toArray(new Type[0]))), new FixedArray<ConditionFormula>());
					}
					else{
						return new CompositeType();
					}
				}
				else{
					return this;
				}
			}
		}
		else{
			return new CompositeType();
		}
	}
	
	@Override
	protected Type typeOf(final String[] name, final int index)  throws TSA_TypeException{
		assert TSA_SyntaxPatterns.isValidAttributeName(name);
		assert index >= 0;
		if(index < name.length){
			String localName = name[index];
			int keyIndex = this.attributes.keys().indexOf(localName);
			if(keyIndex >= 0){
				return this.attributes.values().get(keyIndex).typeOf(name, index + 1);
			}
			else{
				throw new TSA_TypeException("No Attribute with this name: " + TSA_SyntaxPatterns.attributeName(name));
			}
		}
		else{
			assert index == name.length;
			return this;
		}
	}
	
	@Override
	public boolean isEmpty() {
		return this.attributes.isEmpty();
	}
	
	@Override
	protected void basicAttributes(final StringBuilder namePrefix, final ArrayList<String> nameAcc, final ArrayList<BasicType> basicTypeAcc){
		final int prefixLength = namePrefix.length();
		for(int index = 0; index < this.attributes.size(); ++index){
			TSA_SyntaxPatterns.prefixAttribute(namePrefix, this.attributes.keys().get(index));
			this.attributes.values().get(index).basicAttributes(namePrefix, nameAcc, basicTypeAcc);
			namePrefix.setLength(prefixLength);
		}
	}
	
	public FixedArray<ConditionFormula> tupleConstraints(){
		return this.tupleConstraints("");		
	}
	
	public FixedArray<ConditionFormula> tupleConstraints(final Filter<String> attributeFilter){
		return this.tupleConstraints("", attributeFilter);
	}
	
	@Override
	protected void tupleConstraints(final String ownName, final ArrayList<ConditionFormula> constraintAcc, final Filter<String> attributeFilter){
		if(attributeFilter.accepts(ownName)){
			//add own constraints
			Adapter<AtomicTerm<?>, AtomicTerm<?>> prefixer = new PrefixAttributes(ownName);
			
			for(ConditionFormula constraint: this.tupleConstraints){
				//prefix attribute names
				constraintAcc.add(constraint.replaceAtomicTerm(prefixer));			
			}
			
			//add constraints of Attributes		
			for(int index = 0; index < this.attributes.size(); ++index){
				this.attributes.values().get(index).tupleConstraints(TSA_SyntaxPatterns.prefixAttribute(ownName, this.attributes.keys().get(index)), constraintAcc, attributeFilter);
			}
		}
	}
	
	protected TypedAttribute<?>[] subAttributesAsAttributeTerms(final String ownName){
		TypedAttribute<?>[] attributeTerms = new TypedAttribute<?>[this.attributes.size()];
		for(int attribute = 0; attribute < attributeTerms.length; ++attribute){
			attributeTerms[attribute] = TypedAttribute.create(TSA_SyntaxPatterns.prefixAttribute(ownName, this.attributes.keys().get(attribute)), this.attributes.values().get(attribute));
		}
		return attributeTerms;
	}
	
	@Override
	protected int compareToImpl(final Type o) {
		int cmp = super.compareToImpl(o);
		if(cmp == 0){
			CompositeType ctype = (CompositeType) o;
			FixedArray<String> ownAttributeNames = this.attributes.keys();
			FixedArray<String> otherAttributeNames = ctype.attributes.keys();
			cmp = ownAttributeNames.size() - otherAttributeNames.size();
			if(cmp == 0){//both have the same number of attributes
				assert ownAttributeNames.size() == otherAttributeNames.size();
				int numberOfAttributes = ownAttributeNames.size();
				int attribute = 0;
				while(cmp == 0 & attribute < numberOfAttributes){
					cmp = ownAttributeNames.get(attribute).compareTo(otherAttributeNames.get(attribute));
					++attribute;
				}
				FixedArray<Type> ownAttributeTypes = this.attributes.values();
				FixedArray<Type> otherAttributeTypes = ctype.attributes.values();
				attribute = 0;
				while(cmp == 0 & attribute < numberOfAttributes){
					cmp = ownAttributeTypes.get(attribute).compareTo(otherAttributeTypes.get(attribute));
					++attribute;
				}
				if(cmp == 0){
					//TODO compare constraints?
				}
			}
		}
		return cmp;
	}
	
	@Override
	public int hashCode() {
		int hashcode = super.hashCode();
		FixedArray<String> attributeNames = this.attributes.keys();
		FixedArray<Type> attributeTypes = this.attributes.values();
		int numberOfAttributes = attributeNames.size();
		for(int attribute = 0; attribute < numberOfAttributes; ++attribute){
			hashcode ^= attributeNames.get(attribute).hashCode();
		}
		for(int attribute = 0; attribute < numberOfAttributes; ++attribute){
			hashcode ^= attributeTypes.get(attribute).hashCode();
		}
		return hashcode;
	}
	
	@Override
	protected boolean structuralEqualsImpl(final Type type) {
		if(type instanceof CompositeType){
			CompositeType ctype = (CompositeType) type;
			if(this.ID == ctype.ID && ctype.attributes.size() == this.attributes.size()){
				int numberOfAttributes = this.attributes.size();
				for(int attribute = 0; attribute < numberOfAttributes; ++attribute){
					if(ctype.attributes.keys().get(attribute).equals(this.attributes.keys().get(attribute))){
						if(ctype.attributes.values().get(attribute).structuralEquals(this.attributes.values().get(attribute))){
							continue;
						}
						else{
							return false; 
						}
					}
					else{
						return false;
					}
				}
				return true;
			}
			else{
				return false;
			}
		}
		else{
			return false;
		}
	}
	
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append("{");
		if(this.attributes.size() > 0){
			for(int argument = 0; argument < this.attributes.size(); ++argument){
				sb.append(this.attributes.keys().get(argument));
				sb.append("{ ");
				sb.append(this.attributes.values().get(argument));
				sb.append(" }");
				sb.append(", ");
			}
			sb.setLength(sb.length() - ", ".length());
		}
		sb.append("}");
		return sb.toString();
	}

}
