package de.lmu.ifi.pms.cep.tsa.time.analysis;

import java.io.IOException;
import java.io.Writer;

import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormulaID;
import de.lmu.ifi.pms.cep.tsa.time.constraints.ConjunctiveTemporalRelationFormula;
import de.lmu.ifi.pms.cep.tsa.time.constraints.SimpleBoundTemporalFormula;
import de.lmu.ifi.pms.cep.tsa.time.constraints.SimpleTemporalFormula;
import de.lmu.ifi.pms.cep.tsa.time.constraints.TemporalRelationFormula;
import de.lmu.ifi.pms.cep.tsa.time.constraints.enums.BinaryComparisonOperator;

import de.lmu.ifi.pms.java.util.collection.AbstractCollection;
import de.lmu.ifi.pms.java.util.collection.array.Arrays;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;


public class SimpleArithmeticAnalyzer extends ArithmeticAnalyzer{
	
	public static void initWorkMatrix(final long[][] matrix){
		for (int i = 0; i < matrix.length; i++) {
			Arrays.fill(matrix[i], PLUS_INFTY);
			matrix[i][i] = ZERO;
		}
	}
	
	public static void initMatrix(final long[][] matrix){
		for (int i = 0; i < matrix.length; i++) {
			Arrays.fill(matrix[i], MINUS_INFTY);
		}
	}
	
	public static String whitespace(final int count)
	{
		return repeatString(" ", count);
	}
	
	public static String repeatString(final String s, final int count)
	{
		int finalLength = s.length() * count;
		String result = s;
		while(result.length() < finalLength){
			result+=result;
		}
		return result.substring(0, finalLength);
	}
	
	public static void writeMatrix(final Writer writer, final FixedArray<String> names, final long[][] matrix){
		try{
			assert names.size() == matrix.length;
			if(names.size() > 0){
				int maxNameLength = 0;
				int[] columnLength = new int[names.size() + 1];
				int maxColSize = 0;
				
				{
				String colName;
				for (int j = 0; j < matrix.length; j++) {
					colName = names.get(j);
					if(colName.length() > maxNameLength){
						maxNameLength = colName.length();					
					}
					if(colName.length() > 22){
						columnLength[j] = colName.length();
					}
					else{
						columnLength[j] = 22;
					}
					if(columnLength[j] > maxColSize){
						maxColSize = columnLength[j];
					}
				}
				if(maxNameLength > maxColSize){
					maxColSize = maxNameLength;
				}
				}
				
				String whitespace = whitespace(maxColSize);
				
				writer.append(whitespace.substring(0, maxNameLength));
				writer.append(" | ");
				String colName;
				for (int j = 0; j < matrix.length; ++j) {
					if(j > 0){
						writer.append(", ");
					}
					colName = names.get(j);
					writer.append(whitespace.substring(0, columnLength[j] - colName.length()));
					writer.append(colName);
					
					writer.append(", ");
				}
				writer.append("\n");
				
				writer.append(repeatString("-", maxNameLength));
				writer.append("-+-");
				for (int j = 0; j < matrix.length; j++) {
					writer.append(repeatString("-", columnLength[j] + ", ".length()));
				}
				writer.append("\n");
				
				
				for (int i = 0; i < matrix.length;++i) {
					
					String rowName = names.get(i);
					writer.append(whitespace.substring(0, maxNameLength - rowName.length()));
					writer.append(rowName);
					writer.append(" | ");
					
					for (int j = 0; j < matrix.length; j++) {
						if(j > 0){
							writer.append(", ");
						}
						
						long value = matrix[i][j];
						String valueString;
						if(value == PLUS_INFTY){
							valueString = "+infty";
						}
						else if (value == MINUS_INFTY){
							valueString = "-infty";
						}
						else{
							valueString = Long.toString(constant(value)) + (ArithmeticAnalyzer.minusEpsilon(value) ? "-e": "  ");
						}
						writer.append(whitespace.substring(0, columnLength[j] - valueString.length()));
						writer.append(valueString);					
					}
					writer.append("\n");
				}
			}
			writer.append("\n");
			writer.flush();
		} catch (IOException e){
			throw new RuntimeException(e);
		}
	}

	public static void writeMatrix(final StringBuilder sb, final FixedArray<String> names, final long[][] matrix){
		assert names.size() == matrix.length;
		if(names.size() > 0){
			int maxNameLength = 0;
			int[] columnLength = new int[names.size() + 1];
			int maxColSize = 0;
			
			{
			String colName;
			for (int j = 0; j < matrix.length; j++) {
				colName = names.get(j);
				if(colName.length() > maxNameLength){
					maxNameLength = colName.length();					
				}
				if(colName.length() > 22){
					columnLength[j] = colName.length();
				}
				else{
					columnLength[j] = 22;
				}
				if(columnLength[j] > maxColSize){
					maxColSize = columnLength[j];
				}
			}
			if(maxNameLength > maxColSize){
				maxColSize = maxNameLength;
			}
			}
			
			String whitespace = whitespace(maxColSize);
			
			sb.append(whitespace.substring(0, maxNameLength));
			sb.append(" | ");
			String colName;
			for (int j = 0; j < matrix.length; j++) {
				colName = names.get(j);
				sb.append(whitespace.substring(0, columnLength[j] - colName.length()));
				sb.append(colName);
				
				sb.append(", ");
			}
			sb.setLength(sb.length() - ", ".length());
			sb.append("\n");
			
			sb.append(repeatString("-", maxNameLength));
			sb.append("-+-");
			for (int j = 0; j < matrix.length; j++) {
				sb.append(repeatString("-", columnLength[j] + ", ".length()));
			}
			sb.append("\n");
			
			
			for (int i = 0; i < matrix.length; i++) {
				String rowName = names.get(i);
				sb.append(whitespace.substring(0, maxNameLength - rowName.length()));
				sb.append(rowName);
				sb.append(" | ");
				
				for (int j = 0; j < matrix.length; j++) {
					long value = matrix[i][j];
					String valueString;
					if(value == PLUS_INFTY){
						valueString = "+infty";
					}
					else if (value == MINUS_INFTY){
						valueString = "-infty";
					}
					else{
						valueString = Long.toString(constant(value)) + (ArithmeticAnalyzer.minusEpsilon(value) ? "-e": "  ");
					}
					sb.append(whitespace.substring(0, columnLength[j] - valueString.length()));
					sb.append(valueString);
					
					sb.append(", ");
				}
				sb.setLength(sb.length() - ", ".length());
				sb.append("\n");
			}
		}
		sb.append("\n");		
	}
	
	boolean isComputed;
	
	final TemporalRelationFormula formula;
	final FixedArrayNavigableSet<String> relevantAttributeNames;
	final FixedArrayNavigableSet<String> irrelevantAttributeNames;
	final FixedArrayNavigableSet<String> relevantVariableNames;
	final FixedArrayNavigableSet<String> irrelevantVariableNames;
	final int numberOfRelevantAttributes;
	final int numberOfRelevantAttributesAndVariables;
	final int numberOfRelevantAttributesAndVariablesAndIrrelevantAttributes;
	final int size;
	final long[][] matrix;
	long[][] workMatrix;
	boolean isConsitent = false;

	public SimpleArithmeticAnalyzer(final TemporalRelationFormula formula, final FixedArrayNavigableSet<String> relevantAttributes, final FixedArrayNavigableSet<String> relevantVariables){
		this.isComputed = false;
		
		this.formula = formula.simpleForm();
		this.relevantAttributeNames = relevantAttributes;
		this.irrelevantAttributeNames = FixedArrayNavigableSet.difference(this.formula.AttributeNames(),this.relevantAttributeNames);
		this.numberOfRelevantAttributes = this.relevantAttributeNames.size();
		this.relevantVariableNames = relevantVariables;
		this.numberOfRelevantAttributesAndVariables = this.numberOfRelevantAttributes + this.relevantVariableNames.size();
		this.numberOfRelevantAttributesAndVariablesAndIrrelevantAttributes = this.numberOfRelevantAttributesAndVariables + this.irrelevantAttributeNames.size();
		this.irrelevantVariableNames = FixedArrayNavigableSet.difference(this.formula.VariableNames(),this.relevantVariableNames);
		this.size = this.numberOfRelevantAttributesAndVariablesAndIrrelevantAttributes + this.irrelevantVariableNames.size();
		this.matrix = new long[this.numberOfRelevantAttributesAndVariables][this.numberOfRelevantAttributesAndVariables];
		//TODO check constants for small enough size
	}
	
	public SimpleArithmeticAnalyzer(final TemporalRelationFormula formula){
		this(formula, new FixedArrayNavigableSet<String>(), new FixedArrayNavigableSet<String>());
	}
	

	
	private void addWorkMatrix(){
		for(int row=0;row<this.numberOfRelevantAttributesAndVariables;row++){
			for(int col=0;col<this.numberOfRelevantAttributesAndVariables;col++){
				long currentValue = this.matrix[row][col];
				long newValue = this.workMatrix[row][col];
				this.matrix[row][col] = (newValue > currentValue ? newValue : currentValue);
			}	
		}
	}
	
	public static void setDistanceMatrix(final long[][] matrix, final AbstractCollection<SimpleTemporalFormula> flatConjunction, final int numberOfRelevantAttributes, final int numberOfRelevantAttributesAndVariables, final int numberOfRelevantAttributesAndVariablesAndIrrelevantAttributes, final FixedArrayNavigableSet<String> relevantAttributeNames, final FixedArrayNavigableSet<String> irrelevantAttributeNames, final FixedArrayNavigableSet<String> relevantVariableNames, final FixedArrayNavigableSet<String> irrelevantVariableNames){
		assert matrix.length == relevantAttributeNames.size() + relevantVariableNames.size() + irrelevantAttributeNames.size() + irrelevantVariableNames.size();
		assert numberOfRelevantAttributes == relevantAttributeNames.size();
		
		SimpleArithmeticAnalyzer.initWorkMatrix(matrix);
		
		for(SimpleTemporalFormula simpleFormula: flatConjunction){
			assert simpleFormula.op == BinaryComparisonOperator.LESS_EQUAL;
			
			int col;
			int row;
			if(simpleFormula.isAttribute1){
				int index = relevantAttributeNames.indexOf(simpleFormula.v1);
				if(index >= 0){
					col = relevantAttributeNames.indexOf(simpleFormula.v1);
				}
				else{assert irrelevantAttributeNames.contains(simpleFormula.v1);
					col = numberOfRelevantAttributesAndVariables + irrelevantAttributeNames.indexOf(simpleFormula.v1);
				}
			}
			else{
				int index = relevantVariableNames.indexOf(simpleFormula.v1);
				if(index >= 0){
					col = numberOfRelevantAttributes + relevantVariableNames.indexOf(simpleFormula.v1);
				}
				else{assert irrelevantVariableNames.contains(simpleFormula.v1);
					col = numberOfRelevantAttributesAndVariablesAndIrrelevantAttributes + irrelevantVariableNames.indexOf(simpleFormula.v1);
				}
			}
			if(simpleFormula.isAttribute2){
				int index = relevantAttributeNames.indexOf(simpleFormula.v2);
				if(index >= 0){
					row = relevantAttributeNames.indexOf(simpleFormula.v2);
				}
				else{assert irrelevantAttributeNames.contains(simpleFormula.v2);
					row = numberOfRelevantAttributesAndVariables + irrelevantAttributeNames.indexOf(simpleFormula.v2);
				}
			}
			else{
				int index = relevantVariableNames.indexOf(simpleFormula.v2);
				if(index >= 0){
					row = numberOfRelevantAttributes +  relevantVariableNames.indexOf(simpleFormula.v2);
				}
				else{assert irrelevantVariableNames.contains(simpleFormula.v2);
					row = numberOfRelevantAttributesAndVariablesAndIrrelevantAttributes + irrelevantVariableNames.indexOf(simpleFormula.v2);
				}
			}
			
			long encodedConstant;
			if(simpleFormula.constant < CONSTANT_MIN_SIZE){
				encodedConstant = MINUS_INFTY;
			}
			else if(simpleFormula.constant > CONSTANT_MAX_SIZE){
				encodedConstant = PLUS_INFTY;
			}
			else{
				assert CONSTANT_MIN_SIZE <= simpleFormula.constant & simpleFormula.constant <= CONSTANT_MAX_SIZE;
				encodedConstant = encode(simpleFormula.constant, simpleFormula.minusEpsilon);//TODO more efficient if done only once per Formula, i.e. change formulas
			}
			long currentValue = matrix[row][col];
			
			matrix[row][col] = (encodedConstant < currentValue? encodedConstant : currentValue);			
		}
	}
	
	public static void applyDontCareToDistanceMatrix(final long[][] matrix, final AbstractCollection<SimpleBoundTemporalFormula> flatConjunction, final int numberOfRelevantAttributes, final int numberOfRelevantAttributesAndVariables, final int numberOfRelevantAttributesAndVariablesAndIrrelevantAttributes, final FixedArrayNavigableSet<String> relevantAttributeNames, final FixedArrayNavigableSet<String> irrelevantAttributeNames, final FixedArrayNavigableSet<String> relevantVariableNames, final FixedArrayNavigableSet<String> irrelevantVariableNames){
		assert matrix.length == relevantAttributeNames.size() + relevantVariableNames.size() + irrelevantAttributeNames.size() + irrelevantVariableNames.size();
		assert numberOfRelevantAttributes == relevantAttributeNames.size();
		
		for(SimpleBoundTemporalFormula simpleFormula: flatConjunction){
			
			int pos;
			if(simpleFormula.isAttribute){
				int index = relevantAttributeNames.indexOf(simpleFormula.v);
				if(index >= 0){
					pos = relevantAttributeNames.indexOf(simpleFormula.v);
				}
				else{assert irrelevantAttributeNames.contains(simpleFormula.v);
					pos = numberOfRelevantAttributesAndVariables + irrelevantAttributeNames.indexOf(simpleFormula.v);
				}
			}
			else{
				int index = relevantVariableNames.indexOf(simpleFormula.v);
				if(index >= 0){
					pos = numberOfRelevantAttributes + relevantVariableNames.indexOf(simpleFormula.v);
				}
				else{assert irrelevantVariableNames.contains(simpleFormula.v);
					pos = numberOfRelevantAttributesAndVariablesAndIrrelevantAttributes + irrelevantVariableNames.indexOf(simpleFormula.v);
				}
			}
			
			for(int row = 0; row < numberOfRelevantAttributesAndVariables; ++row){
				matrix[row][pos] = MINUS_INFTY;
			}
			
			for(int col = 0; col < numberOfRelevantAttributesAndVariables; ++col){
				matrix[pos][col] = MINUS_INFTY;
			}
					
		}
	}
	
	@SuppressWarnings("unchecked")
	public void calculate(){
		assert this.formula.isFlatened;
		
		if(this.isComputed == false){
			if(this.formula.ID == ConditionFormulaID.BOTTOM){
				initMatrix(this.matrix);
				this.isConsitent = false;
			}
			else if(this.formula.ID == ConditionFormulaID.TOP){
				SimpleArithmeticAnalyzer.initWorkMatrix(this.matrix);
				this.isConsitent = true;
			}
			else{
				initMatrix(this.matrix);
				this.workMatrix = new long[this.size][this.size];
				
				for(ConjunctiveTemporalRelationFormula trf : this.formula.tempDNFiteration()){
					ArrayList<SimpleTemporalFormula> simpleFormulas = new ArrayList<SimpleTemporalFormula>();
					ArrayList<SimpleBoundTemporalFormula> boundedFormulas = new ArrayList<SimpleBoundTemporalFormula>();
					for(TemporalRelationFormula atrf: trf.subFormulas()){
						if(atrf instanceof SimpleTemporalFormula){
							simpleFormulas.add((SimpleTemporalFormula)atrf);
						}
						else if(atrf instanceof SimpleBoundTemporalFormula){
							boundedFormulas.add((SimpleBoundTemporalFormula)atrf);
						}
						else{
							throw new Error();
						}
					}
					
					setDistanceMatrix(this.workMatrix, simpleFormulas, this.numberOfRelevantAttributes, this.numberOfRelevantAttributesAndVariables, this.numberOfRelevantAttributesAndVariablesAndIrrelevantAttributes, this.relevantAttributeNames, this.irrelevantAttributeNames, this.relevantVariableNames, this.irrelevantVariableNames);
					SimpleFloydWarshall sfw = new SimpleFloydWarshall(this.workMatrix, this.numberOfRelevantAttributesAndVariables);
					sfw.calculate();
					
					if(sfw.isConsistent()){
						applyDontCareToDistanceMatrix(this.workMatrix, boundedFormulas, this.numberOfRelevantAttributes, this.numberOfRelevantAttributesAndVariables, this.numberOfRelevantAttributesAndVariablesAndIrrelevantAttributes, this.relevantAttributeNames, this.irrelevantAttributeNames, this.relevantVariableNames, this.irrelevantVariableNames);
						
						this.addWorkMatrix();
						this.isConsitent = true;
					}
				}
				
				this.workMatrix = null;
			}
			
			this.isComputed = true;
		}
	}
	
	@Override
	public long distance(final String name1, final boolean isAttribute1, final String name2, final boolean isAttribute2){
		if(!this.isComputed){
			this.calculate();
		}
		int index1 = (isAttribute1 ? this.relevantAttributeNames.indexOf(name1) : this.relevantVariableNames.indexOf(name1));
		int index2 = (isAttribute2 ? this.relevantAttributeNames.indexOf(name2) : this.relevantVariableNames.indexOf(name2));
		if(index1 < 0 | index2 < 0){
			return PLUS_INFTY;
		}
		else{
			return this.matrix[(isAttribute1 ? index1 : this.numberOfRelevantAttributes + index1)][(isAttribute2 ? index2 : this.numberOfRelevantAttributes + index2)];
		}
	}
	
	public long[][] distances(){
		if(!this.isComputed){
			this.calculate();
		}
		return this.matrix;
	}
	
	@Override
	public boolean isConsistent(){
		if(!this.isComputed){
			this.calculate();
		}
		return this.isConsitent;
	}
	
	@SuppressWarnings("unchecked")
	@Override
	public void writeTo(final StringBuilder sb) {
		SimpleArithmeticAnalyzer.writeMatrix(sb, FixedArray.concat(this.relevantAttributeNames, this.relevantVariableNames), this.matrix);
		sb.append("\n");
		this.formula.writeTo(sb);
		if((!this.isComputed) & this.workMatrix != null){
			sb.append("\n");
			SimpleArithmeticAnalyzer.writeMatrix(sb, FixedArray.concat(this.relevantAttributeNames, this.relevantVariableNames, this.irrelevantAttributeNames, this.irrelevantVariableNames), this.workMatrix);
		}
	}

}
