package de.lmu.ifi.pms.cep.tsa.time.analysis;

import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.cep.tsa.time.constraints.TemporalRelationFormula;

import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;

public class SimpleArithmeticAnalyzerFactory extends ArithmeticAnalyzerFactory{
	
	final ArithmeticRelationFormulaExtractor formulaExtractor;		

	public SimpleArithmeticAnalyzerFactory(final ArithmeticRelationFormulaExtractor formulaExtractor) {
		super();
		this.formulaExtractor = formulaExtractor;
	}

	public SimpleArithmeticAnalyzerFactory(final BasicType type) {
		this(new ArithmeticRelationFormulaExtractor(type));
	}

	public SimpleArithmeticAnalyzerFactory(final BasicTypeID type) {
		this(type.type);
	}
	
	@Override
	public SimpleArithmeticAnalyzer make(final ConditionFormula formula){
		return this.make(this.formulaExtractor.extract(formula));
	}
	
	@Override
	public SimpleArithmeticAnalyzer make(final ConditionFormula formula, final FixedArrayNavigableSet<String> relevantAttributes, final FixedArrayNavigableSet<String> relevantVariables){
		return this.make(this.formulaExtractor.extract(formula), relevantAttributes, relevantVariables);
	}

	public SimpleArithmeticAnalyzer make(final TemporalRelationFormula f){
		return new SimpleArithmeticAnalyzer(f);
	}
	
	public SimpleArithmeticAnalyzer make(final TemporalRelationFormula temporalRelationFormula, final FixedArrayNavigableSet<String> relevantAttributes, final FixedArrayNavigableSet<String> relevantVariables) {
		return new SimpleArithmeticAnalyzer(temporalRelationFormula, relevantAttributes, relevantVariables);
	}
	
}
