package de.lmu.ifi.pms.cep.tsa.time.constraints;

import org.simpleframework.xml.Element;

import de.lmu.ifi.pms.cep.tsa.conditions.AtomicTerm;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormulaID;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;

import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.collection.AbstractCollection;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.filter.Filter;
import de.lmu.ifi.pms.java.util.tupel.Single;

public abstract class AtomicTemporalRelationFormula extends TemporalRelationFormula {
	
	@SuppressWarnings("hiding")
	@Element(name="type")
	public final BasicTypeID type;
	
	public AtomicTemporalRelationFormula(final BasicTypeID type) {
		this(type, true, true);
	}
	
	AtomicTemporalRelationFormula(final BasicTypeID type, final boolean isNormalized, final boolean isFlatened) {
		this(ConditionFormulaID.ATOMIC, type, isNormalized, isFlatened);
	}
	
	AtomicTemporalRelationFormula(final ConditionFormulaID ID, final BasicTypeID type, final boolean isNormalized, final boolean isFlatened) {
		super(ID, isNormalized, isFlatened);
		this.type = type;
	}
	
	AtomicTemporalRelationFormula(final ConditionFormulaID superID, final BasicTypeID type, final ConditionFormulaID ID, final boolean isNormalized, final boolean isFlatened) {
		super(superID, ID, isNormalized, isFlatened);
		this.type = type;
	}

	@Override
	protected abstract void VariableNames(ArrayList<String> variables);
	
	@Override
	protected abstract void AttributeNames(final ArrayList<String> attributes);
	
	@Override
	protected TemporalRelationFormula normalizeImpl() {
		return this;
	}
	
	@Override
	protected TemporalRelationFormula flatenImpl() {
		return this;
	}

	@Override
	public int tempDNFsize(){
		return 1;
	}
	
	@Override
	protected AbstractCollection<TemporalRelationFormula> tempDNFiteration(final boolean positivePolarity) {
		if(positivePolarity){
			return new Single<TemporalRelationFormula, TemporalRelationFormula>(this);
		}
		else{
			return new Single<TemporalRelationFormula, TemporalRelationFormula>(new NegatedTemporalRelationFormula(this));
		}
	}

	
	@Override
	public AbstractCollection<ConjunctiveTemporalRelationFormula> tempDNFiteration() {
		return new Single<ConjunctiveTemporalRelationFormula,ConjunctiveTemporalRelationFormula>(new ConjunctiveTemporalRelationFormula(this));
	}
	
	@Override
	public int tempCNFsize() {
		return 1;
	}
	
	@Override
	protected AbstractCollection<TemporalRelationFormula> tempCNFiteration(final boolean positivePolarity) {
		if(positivePolarity){
			return new Single<TemporalRelationFormula, TemporalRelationFormula>(this);
		}
		else{
			return new Single<TemporalRelationFormula, TemporalRelationFormula>(new NegatedTemporalRelationFormula(this));
		}
	}
	
	@Override
	public AbstractCollection<DisjunctiveTemporalRelationFormula> tempCNFiteration() {
		return new Single<DisjunctiveTemporalRelationFormula,DisjunctiveTemporalRelationFormula>(new DisjunctiveTemporalRelationFormula(this));
	}
	
	@Override
	protected abstract void collect(final Filter<Term<?>> filter, final ArrayList<Term<?>> terms);
	
	@Override
	public abstract TemporalRelationFormula replaceAtomicTerm(final Adapter<? super AtomicTerm<?>, ? extends Term<?>> replacement);

}
