package de.lmu.ifi.pms.cep.tsa.time.constraints;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;
import de.lmu.ifi.pms.cep.tsa.conditions.AtomicTerm;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormulaID;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.filter.Filter;
  
public abstract class CompositeTemporalRelationFormula extends TemporalRelationFormula{
	
	protected final TemporalRelationFormula[] subformulas;
	
	CompositeTemporalRelationFormula(final ConditionFormulaID ID, final boolean isNormalized, final boolean isFlatened, final TemporalRelationFormula... subformulas) {
		super(ID, isNormalized, isFlatened);
		this.subformulas = subformulas;
		//TODO copy array?
	}
	
	public FixedArray<TemporalRelationFormula> subFormulas() {
		return new FixedArray<TemporalRelationFormula>(false, this.subformulas);
	}
	
	@Override
	protected void VariableNames(final ArrayList<String> variableNames) {
		for(TemporalRelationFormula subFormula : this.subformulas){
			subFormula.VariableNames(variableNames);
		}
	}
	
	@Override
	protected void AttributeNames(final ArrayList<String> attributeNames) {
		for(TemporalRelationFormula subFormula : this.subformulas){
			subFormula.AttributeNames(attributeNames);
		}
	}
	
	@Override
	protected void collect(final Filter<Term<?>> filter, final ArrayList<Term<?>> terms) {
		for(TemporalRelationFormula subFormula: this.subformulas){
			subFormula.collect(filter, terms);
		}
	}
	
	@Override
	public TemporalRelationFormula replaceAtomicTerm(final Adapter<? super AtomicTerm<?>, ? extends Term<?>> replacement) {
		TemporalRelationFormula[] subformulas = new TemporalRelationFormula[this.subformulas.length];
		boolean changed = false;
		for(int index = 0; index < subformulas.length; ++index){
			subformulas[index] = this.subformulas[index].replaceAtomicTerm(replacement);
			if(subformulas[index] != this.subformulas[index]){
				changed = true;
			}
		}
		if(changed){
			return this.newFormula(subformulas);
		}
		else{
			return this;
		}
	}
	
	@Override
	public TemporalRelationFormula replaceTerm(final Adapter<Term<?>, ? extends Term<?>> replacement) {
		TemporalRelationFormula[] subformulas = new TemporalRelationFormula[this.subformulas.length];
		boolean changed = false;
		for(int index = 0; index < subformulas.length; ++index){
			subformulas[index] = this.subformulas[index].replaceTerm(replacement);
			if(subformulas[index] != this.subformulas[index]){
				changed = true;
			}
		}
		if(changed){
			return this.newFormula(subformulas);
		}
		else{
			return this;
		}
	}
	
	protected abstract CompositeTemporalRelationFormula newFormula(TemporalRelationFormula[] subformulas);
	
	@Override
	public void writeTo(final StringBuilder sb, final int indentDepth) {
		indentTab(sb, indentDepth);
		sb.append(this.ID.toString());
		if(this.subformulas.length > 0){
			sb.append("{\n");
			for(ConditionFormula subFormula: this.subformulas){
				subFormula.writeTo(sb, indentDepth + 1);
				sb.append(",\n");
			}
			sb.setLength(sb.length() - ",\n".length());
			sb.append("\n");
			indentTab(sb, indentDepth);
			sb.append("}");
		}
		else{
			sb.append("{}");
		}
	}
	
}
