package de.lmu.ifi.pms.cep.tsa.time.constraints;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;
import static de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormulaID.NEGATED;

import de.lmu.ifi.pms.cep.tsa.conditions.AtomicTerm;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;

import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.collection.AbstractCollection;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.filter.Filter;

public class NegatedTemporalRelationFormula extends TemporalRelationFormula {
	
	protected final TemporalRelationFormula subformula;
	
	public NegatedTemporalRelationFormula(final TemporalRelationFormula subformula) {
		this(initialIsNormalized(subformula), subformula);
	}
	
	NegatedTemporalRelationFormula(final boolean isNormalized, final TemporalRelationFormula subformula) {
		this(isNormalized, initialIsFlatened(isNormalized, subformula), subformula);		
	}
	
	NegatedTemporalRelationFormula(final boolean isNormalized, final boolean isFlatened, final TemporalRelationFormula subformula) {
		super(NEGATED, isNormalized, isFlatened);
		assert isNormalized == initialIsNormalized(subformula);
		assert isFlatened == initialIsFlatened(isNormalized, subformula);
		assert !(isNormalized & !isFlatened & subformula.isFlatened);
		this.subformula = subformula;
	}
	
	protected static boolean initialIsNormalized(final TemporalRelationFormula subformula){
		return subformula.isNormalized & (subformula.ID != NEGATED);
	}
	
	protected static boolean initialIsFlatened(final boolean isNormalized, final TemporalRelationFormula subformula){
		return isNormalized & subformula.isFlatened;
	}

	@Override
	protected void VariableNames(final ArrayList<String> variables) {
		this.subformula.VariableNames(variables);
	}
	
	@Override
	protected void AttributeNames(final ArrayList<String> attributes) {
		this.subformula.AttributeNames(attributes);
	}
	
	@Override
	protected TemporalRelationFormula normalizeImpl() {
		assert !this.isNormalized;
		if(this.subformula.ID == NEGATED){
			return ((NegatedTemporalRelationFormula)this.subformula).subformula.normalize();
		}
		else{
			assert !this.subformula.isNormalized;
			TemporalRelationFormula normalizedSubFormula = this.subformula.normalizeImpl();
			switch (normalizedSubFormula.ID) {
			case ATOMIC:
				return new NegatedTemporalRelationFormula(true, normalizedSubFormula);
			case BOTTOM:
				return Top;
			case CONJUNCTIVE:
				return new NegatedTemporalRelationFormula(true, normalizedSubFormula);
			case DISJUNCTIVE:
				return new NegatedTemporalRelationFormula(true, normalizedSubFormula);
			case NEGATED:
				return ((NegatedTemporalRelationFormula)normalizedSubFormula).subformula;
			case TOP:
				return Bottom;
			}
			//	default:
			//		break;
		}
		throw new Error("Forgotten Case");
	}

	@Override
	protected TemporalRelationFormula flatenImpl() {
		assert this.isNormalized;
		assert !this.isFlatened;
		assert this.subformula.ID != NEGATED;
		assert this.subformula.isNormalized;
		assert !this.subformula.isFlatened;
		return new NegatedTemporalRelationFormula(true, true, this.subformula.flatenImpl());
	}
	
	@Override
	protected TemporalRelationFormula simpleForm(final boolean positivePolarity) {
		return this.subformula.simpleForm(!positivePolarity);
	}


	@Override
	public int tempDNFsize() {
		return this.subformula.tempCNFsize();
	}

	@Override
	protected AbstractCollection<TemporalRelationFormula> tempDNFiteration(final boolean positivePolarity) {
		return this.subformula.tempCNFiteration(!positivePolarity);
	}

	@Override
	public int tempCNFsize() {
		return this.subformula.tempDNFsize();
	}

	@Override
	protected AbstractCollection<TemporalRelationFormula> tempCNFiteration(final boolean positivePolarity) {
		return this.subformula.tempDNFiteration(!positivePolarity);
	}
	
	@Override
	protected void collect(final Filter<Term<?>> filter, final ArrayList<Term<?>> terms) {
		this.subformula.collect(filter, terms);
	}
	
	@Override
	public TemporalRelationFormula replaceAtomicTerm(final Adapter<? super AtomicTerm<?>, ? extends Term<?>> replacement) {
		TemporalRelationFormula subformula = this.subformula.replaceAtomicTerm(replacement);
		if(subformula == this.subformula){
			return this;
		}
		else{
			return new NegatedTemporalRelationFormula(subformula);
		}
	}
	
	@Override
	public TemporalRelationFormula replaceTerm(final Adapter<Term<?>, ? extends Term<?>> replacement) {
		TemporalRelationFormula subformula = this.subformula.replaceTerm(replacement);
		if(subformula == this.subformula){
			return this;
		}
		else{
			return new NegatedTemporalRelationFormula(subformula);
		}
	}
	
	@Override
	public void writeTo(final StringBuilder sb, final int indentDepth) {
		indentTab(sb, indentDepth);
		sb.append(this.ID);
		sb.append("{\n");
		this.subformula.writeTo(sb, indentDepth + 1);
		sb.append("\n");
		indentTab(sb, indentDepth);
		sb.append("}");
	}

}
