package de.lmu.ifi.pms.cep.tsa.time.constraints;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;

import de.lmu.ifi.pms.cep.tsa.conditions.AtomicTerm;
import de.lmu.ifi.pms.cep.tsa.conditions.BasicTypeAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.BasicTypeVariable;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.cep.tsa.time.constraints.enums.BinaryComparisonOperator;
import de.lmu.ifi.pms.cep.tsa.time.constraints.enums.MinMax;

import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.collection.AbstractCollection;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.filter.Filter;

public class TemporalMinMax extends AtomicTemporalRelationFormula{
	public final String v;
	boolean isAttributeV;
	private final String[] arguments;
	private final boolean[] isAttribute;
	public final MinMax operator;
	
	/**
	 * This class represents  v=min(variables) and v=max(variables)
	 * @param m this decides whether the type is minimum oder maximum
	 * @param v 
	 * @param identifier
	 */
	public TemporalMinMax(final BasicTypeID type, final MinMax m, final String v, final AbstractCollection<String> arguments){
		this(type, v, m, arguments.toArray(new String[arguments.size()]));
	}

	public TemporalMinMax(final BasicTypeID type, final String v1, final boolean isAttribute1, final MinMax m, final String[] arguments, final boolean[] isAttribute){
		super(type);
		this.operator = m;
		this.v = v1;
		this.isAttributeV = isAttribute1;
		this.arguments = arguments;
		this.isAttribute = isAttribute;
		//TODO copy arguments?
		
		if(this.arguments.length != this.isAttribute.length){
			throw new IllegalArgumentException();
			//TODO better exception
		}

	}
	
	public TemporalMinMax(final BasicTypeID type, final String v, final MinMax m, final String... arguments){
		this(type, v, false, m, arguments, new boolean[arguments.length]);
	}

	public FixedArray<String> arguments() {
		return new FixedArray<String>(this.arguments);
	}
	
	@Override
	public void writeTo(final StringBuilder sb, final int indentDepth) {
		indentTab(sb, indentDepth);
		sb.append((this.isAttributeV ? "attr " : "var "));
		sb.append("v = ");
		if(this.operator==MinMax.MIN){
			sb.append("min( ");
		}else{
			sb.append("max( ");
		}
		for(int arg = 0; arg < this.arguments.length; ++arg){
			sb.append((this.isAttribute[arg] ? "attr " : "var "));
			sb.append(this.arguments[arg]);
			sb.append(", ");
		}
		sb.setLength(sb.length()-", ".length());
		sb.append(" )");
	}
	
	@Override
	protected void VariableNames(final ArrayList<String> variables) {
		if(!this.isAttributeV){
			variables.add(this.v);
		}
		for(int arg = 0; arg < this.arguments.length; ++arg){
			if(!this.isAttribute[arg]){
				variables.add(this.arguments[arg]);
			}
		}
	}
	
	@Override
	protected void AttributeNames(final ArrayList<String> attributes) {
		if(this.isAttributeV){
			attributes.add(this.v);
		}
		for(int arg = 0; arg < this.arguments.length; ++arg){
			if(this.isAttribute[arg]){
				attributes.add(this.arguments[arg]);
			}
		}
	}
	
	@Override
	protected TemporalRelationFormula simpleForm(final boolean positivePolarity){
		if(positivePolarity){
			ArrayList<TemporalRelationFormula> leftSide = new ArrayList<TemporalRelationFormula>();
			ArrayList<TemporalRelationFormula> rightSide = new ArrayList<TemporalRelationFormula>();
			for(int arg = 0; arg < this.arguments.length; ++arg){
				String argument = this.arguments[arg];
				boolean isAttribute = this.isAttribute[arg];
				if(this.operator==MinMax.MIN){
					leftSide.add(new SimpleTemporalFormula(this.type, this.v, this.isAttributeV, BinaryComparisonOperator.LESS_EQUAL, argument, isAttribute, false));
					rightSide.add(new SimpleTemporalFormula(this.type, argument, isAttribute, BinaryComparisonOperator.LESS_EQUAL,this.v, this.isAttributeV, false));
				}else{assert this.operator==MinMax.MAX;
					leftSide.add(new SimpleTemporalFormula(this.type, argument, isAttribute, BinaryComparisonOperator.LESS_EQUAL, this.v, this.isAttributeV, false));
					rightSide.add(new SimpleTemporalFormula(this.type, this.v, this.isAttributeV, BinaryComparisonOperator.LESS_EQUAL,argument, isAttribute, false));
				}
			}
			return new ConjunctiveTemporalRelationFormula(new ConjunctiveTemporalRelationFormula(leftSide), new DisjunctiveTemporalRelationFormula(rightSide));
		}
		else{
			return this.simpleForm(true).simpleForm(false);
			//Does not produce loop, as this.simpleForm(true) does not contain a min or max anymore
		}
	}
	
	@Override
	protected void collect(final Filter<Term<?>> filter, final ArrayList<Term<?>> terms) {
		AtomicTerm<?> v = this.arg1();
		if(filter.accepts(v)){
			terms.add(v);
		}
		for(int arg = 0; arg < this.arguments.length; ++arg){
			AtomicTerm<?> argument = this.arg(arg);
			if(filter.accepts(argument)){
				terms.add(argument);
			}
		}
	}
	
	protected AtomicTerm<BasicType> arg1(){
		if(this.isAttributeV){
			return BasicTypeAttribute.create(this.v, this.type);
		}
		else{
			return BasicTypeVariable.create(this.v, this.type);
		}
	}
	
	protected AtomicTerm<BasicType> arg(final int index){
		if(this.isAttribute[index]){
			return BasicTypeAttribute.create(this.arguments[index], this.type);
		}
		else{
			return BasicTypeVariable.create(this.arguments[index], this.type);
		}
	}
	
	@Override
	public TemporalRelationFormula replaceAtomicTerm(final Adapter<? super AtomicTerm<?>, ? extends Term<?>> replacement) {
		Term<?> t1 = replacement.adapt(this.arg1());
		String v1;
		BasicTypeID type1;
		boolean isAttribute1;
		if(t1 instanceof BasicTypeAttribute){
			v1 = ((BasicTypeAttribute) t1).name;
			type1 = ((BasicTypeAttribute) t1).type.ID;
			isAttribute1 = true;
		}
		else if(t1 instanceof BasicTypeVariable){
			v1 = ((BasicTypeVariable) t1).name;
			type1 = ((BasicTypeVariable) t1).type.ID;
			isAttribute1 = false;
		}
		else{
			throw new IllegalArgumentException();
			//TODO better exception
		}
		
		String[] arguments = new String[this.arguments.length];
		boolean[] isAttribute = new boolean[arguments.length];
		boolean changed = false;
		
		for(int argIndex = 0; argIndex < this.arguments.length; ++argIndex){
			Term<?> t2 = replacement.adapt(this.arg(argIndex));
			if(t2 instanceof BasicTypeAttribute){
				arguments[argIndex] = ((BasicTypeAttribute) t2).name;
				isAttribute[argIndex] = true;
				if(((BasicTypeAttribute) t2).type.ID != type1){
					throw new IllegalArgumentException();
					//TODO better exception
				}
			}
			else if(t2 instanceof BasicTypeVariable){
				arguments[argIndex] = ((BasicTypeVariable) t2).name;
				isAttribute[argIndex] = false;
			}
			else{
				throw new IllegalArgumentException();
				//TODO better exception
			}
			if(!this.arguments[argIndex].equals(arguments[argIndex]) | this.isAttribute[argIndex] != isAttribute[argIndex]){
				changed = true;
			}
		}
		
		
		
		if(v1.equals(this.v) & type1 == this.type & isAttribute1 == this.isAttributeV & !changed){
			return this;
		}
		else{
			return new TemporalMinMax(type1, v1, isAttribute1, this.operator, arguments, isAttribute); 
		}
	}
	
	
	@Override
	public TemporalRelationFormula replaceTerm(final Adapter<Term<?>, ? extends Term<?>> replacement) {
		return this.replaceAtomicTerm(replacement);
	}
	
}