package de.lmu.ifi.pms.cep.tsa.time.functions;

import org.simpleframework.xml.Attribute;

public class AddDuration extends SimpleTimeFunction {
	
	@Attribute(name="duration") 
	public final long duration;

	public AddDuration(@Attribute(name="duration") final long duration) {
		super();
		this.duration = duration;
	}
	
	public static final long PLUS_INFTY = Long.MAX_VALUE;
	public static final long MINUS_INFTY = Long.MIN_VALUE;
	
	
	@Override
	long apply(final long[] arguments, final int argStart) {
		final long argument = arguments[argStart];
		if(argument == MINUS_INFTY | this.duration == MINUS_INFTY){
			return MINUS_INFTY;
		}
		else{
			long result = argument + this.duration;
			
			long overflow = ((argument ^ result) & (this.duration ^ result)) >> 63;
			
			//overflow occured if overflow = -1 (only 1 bits) else overflow = 0 (only 0 bits);		
			long plusInfty = (overflow & result) >> 63;
		
			//MINUS_INFTY + MINUS_INFTY = MINUS_INFTY => no overflow detected but correct result
			//MINUS_INFTY + neg value = pos value => overflow detected => minusInfty is set
			long minusInfty = overflow & argument & MINUS_INFTY;//(overflow & m) >> 63 & MINUS_INFTY would be independent from the value of MINUS_INFTY
			//is MINUS_INFTY if an underflow occurs else 0 		
			return (plusInfty & PLUS_INFTY) | minusInfty | (~overflow & result);
		}
	}

	@Override
	public int argumentArity() {
		return 1;
	}

}
