package de.lmu.ifi.pms.cep.tsa.time.functions;

import java.util.ArrayList;
import java.util.Arrays;

import org.simpleframework.xml.ElementList;
import org.simpleframework.xml.ElementListUnion;
import org.simpleframework.xml.core.Complete;
import org.simpleframework.xml.core.Persist;
import org.simpleframework.xml.core.Resolve;



public class ParallelCompositeTimeFunction extends TimeFunction {
	
	final int arity;
	final int resultArity;
	
	protected final TimeFunction[] subFunctions;
	
	@ElementListUnion({
		@ElementList(entry="unity", type=UnityTimeFunction.class, inline=true, required=false),
		@ElementList(entry="identity", type=IdentityTimeFunction.class, inline=true, required=false),
		@ElementList(entry="const", type=ConstantTimeFunction.class, inline=true, required=false),
		@ElementList(entry="add", type=AddDuration.class, inline=true, required=false),
		@ElementList(entry="min", type=Min.class, inline=true, required=false),
		@ElementList(entry="max", type=Max.class, inline=true, required=false),
		@ElementList(entry="map", type=MapTimeFunction.class, inline=true, required=false),
		@ElementList(entry="all", type=ParallelCompositeTimeFunction.class, inline=true, required=false),
		@ElementList(entry="compose", type=SequencialCompositeTimeFunction.class, inline=true, required=false)
	})
	java.util.ArrayList<TimeFunction> getSubFunctions = new ArrayList<TimeFunction>();
	
	@Persist
	protected void persist(){
		if(this.subFunctions.length > 0){
			this.getSubFunctions = new ArrayList<TimeFunction>(Arrays.asList(this.subFunctions));
		}
		else{
			this.getSubFunctions = new ArrayList<TimeFunction>();
			this.getSubFunctions.add(UnityTimeFunction.instance);
			this.getSubFunctions.add(null);
			//FIXME remove when simple-xml framework bug is fixed
		}
	}
	
	@Complete
	protected void complete(){
		this.getSubFunctions = null;
	}
	
	@Resolve
	protected ParallelCompositeTimeFunction resolve(){
		TimeFunction[] subfunctions = this.getSubFunctions.toArray(new TimeFunction[this.getSubFunctions.size()]);
		if(subfunctions.length == 1 && subfunctions[0] == UnityTimeFunction.instance){
			return new ParallelCompositeTimeFunction(new TimeFunction[]{});
			//FIXME remove case when simple-xml framework bug is fixed
		}
		else{
			return new ParallelCompositeTimeFunction(subfunctions);
		}
	}
	
	final int[] argStarts;
	final int[] resultStarts;
	
	@SuppressWarnings("unused")
	private ParallelCompositeTimeFunction() {
		this.arity = 0;
		this.resultArity = 0;
		this.subFunctions = null;
		this.argStarts = null;
		this.resultStarts = null;
	}
	
	public ParallelCompositeTimeFunction(final TimeFunction... subFunctions) {
		super();
		int argumentArity = 0;
		
		int resultArity = 0;
		int[] argumentStarts = new int[subFunctions.length];
		int[] resultStarts = new int[subFunctions.length];
		
		for(int index = 0; index < subFunctions.length; ++index){
			argumentStarts[index] = argumentArity;
			argumentArity += subFunctions[index].argumentArity();
			
			resultStarts[index] = resultArity;
			resultArity += subFunctions[index].resultArity();
		}
		
		this.arity = argumentArity;
		this.resultArity = resultArity;
		this.subFunctions = subFunctions;
		this.argStarts = argumentStarts;
		this.resultStarts = resultStarts;
	}

	@Override
	public void apply(final long[] arguments, final int argStart, final long[] results, final int resultStart) {
		assert argStart >= 0;
		assert resultStart >= 0;
		assert arguments.length >= argStart + this.argumentArity();
		assert results.length >= resultStart + this.resultArity(); 
		
		for(int index = 0; index < this.subFunctions.length; ++index){
			this.subFunctions[index].apply(arguments, resultStart + this.argStarts[index], results, resultStart + this.resultStarts[index]);
		}
	}

	@Override
	public final int argumentArity() {
		return this.arity;
	}

	@Override
	public final int resultArity() {
		return this.resultArity;
	}

}
