package de.lmu.ifi.pms.cep.tsa.time.functions;

import org.simpleframework.xml.Element;


public class SequencialCompositeTimeFunction extends TimeFunction {
	
	public final TimeFunction secondFunction;
	
	@Element(name="second")
	ParallelCompositeTimeFunction getSecondFunction(){
		return new ParallelCompositeTimeFunction(this.secondFunction);
	}
	
	public final TimeFunction firstFunction;
	
	@Element(name="first")
	ParallelCompositeTimeFunction getFirstFunction(){
		return new ParallelCompositeTimeFunction(this.firstFunction);
	}

	public SequencialCompositeTimeFunction(final TimeFunction secondFunction, final TimeFunction firstFunction) {
		super();
		this.firstFunction = firstFunction;
		this.secondFunction = secondFunction;
		
		//check
		if(this.firstFunction.resultArity() != this.secondFunction.argumentArity()){
			throw new IllegalArgumentException("Incompatible Functions");
		}
	}
	
	@SuppressWarnings("unused")
	private SequencialCompositeTimeFunction(@Element(name="second") final ParallelCompositeTimeFunction secondFunction, @Element(name="first") final ParallelCompositeTimeFunction firstFunction) {
		this(secondFunction.subFunctions[0], firstFunction.subFunctions[0]);
	}

	@Override
	public void apply(final long[] arguments, final int argStart, final long[] results, final int resultStart) {
		assert argStart >= 0;
		assert resultStart >= 0;
		assert arguments.length >= argStart + this.argumentArity();
		assert results.length >= resultStart + this.resultArity() : "Result arity is " + this.resultArity() + " , start Index is " + resultStart + " and array length is " + results.length; 
		
		long[] intermediateResult = new long[this.firstFunction.resultArity()];
		this.firstFunction.apply(arguments, argStart, intermediateResult, 0);
		this.secondFunction.apply(intermediateResult, 0, results, resultStart);
	}

	@Override
	public final int argumentArity() {
		return this.firstFunction.argumentArity();
	}

	@Override
	public final int resultArity() {
		return this.secondFunction.resultArity();
	}
	
}
