package de.lmu.ifi.pms.java.lang.iteration;

import java.util.Comparator;
import java.util.Iterator;

import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.adaption.IterationAdapterImpl;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.extendible.ExtendibleCollection;
import de.lmu.ifi.pms.java.util.collection.fixed.wrapper.HeadTailWrapper;
import de.lmu.ifi.pms.java.util.collection.set.SortedSet;
import de.lmu.ifi.pms.java.util.comparator.AbstractComparator;



public abstract class AbstractOrderedIterator<Type, IteratorType extends Iterator<? extends Type>, CollectionType extends ExtendibleCollection<IteratorType>> extends AbstractFixedIterator<Type> {

	protected final CollectionType iterators;
	
	private final Comparator<? super Type> comparator;

	public AbstractOrderedIterator(final Collection<? extends SortedSet<Type>> sortedSets, final Adapter<SortedSet<Type>, IteratorType> adapter, final CollectionType emptyCollection) {
		this.iterators = emptyCollection;
		for(final IteratorType iterator: new IterationAdapterImpl<SortedSet<Type>, IteratorType>(new FilteredIteration<SortedSet<Type>>(sortedSets){
			
			/* (non-Javadoc)
			 * @see extension.java.lang.iteration.FilteredIteration#accepts(java.lang.Object)
			 */
			@Override
			public boolean accepts(final SortedSet<Type> element) {
				return !element.isEmpty(); //TODO implement
			}
			
		}, adapter)){
			emptyCollection.add(iterator);
		}
		
		
		if(sortedSets.isEmpty()){
			this.comparator = null;
		}
		else{
			final HeadTailWrapper<SortedSet<Type>> ht = new HeadTailWrapper<SortedSet<Type>>(sortedSets);
			this.comparator = ht.head().comparator();
			for(final SortedSet<Type> s: ht){
				if((this.comparator == null && !(s.comparator() == null)) || (this.comparator != null && !this.comparator.equals(s.comparator()))){
					throw new IllegalArgumentException(); //TODO better exception
				}
			}
		}
	}
	
	protected final int iteratorCount(){
		return this.iterators.size();
	}
	
	protected final boolean isEmptyIterators(){
		return this.iterators.isEmpty();
	}
	
	public final int compare(final Type t1, final Type t2){
		return AbstractComparator.compare(t1, t2, this.comparator);
	}
	
	@Override
	public abstract boolean hasNext();

	@Override
	public abstract Type next();

}
