package de.lmu.ifi.pms.java.lang.iteration;

import java.util.Comparator;
import java.util.Iterator;
import java.util.NoSuchElementException;

import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.compatibility.TreeSet;
import de.lmu.ifi.pms.java.util.collection.set.SortedSet;
import de.lmu.ifi.pms.java.util.comparator.AbstractComparator;



public class OrderedUnionIterator<Type> extends AbstractOrderedIterator<Type, OrderedUnionIterator.ComparableExaminingIteratorWrapper<Type>, TreeSet<OrderedUnionIterator.ComparableExaminingIteratorWrapper<Type>>>{

	public OrderedUnionIterator(final Collection<? extends SortedSet<Type>> sortedSets) {
		super(sortedSets, new Adapter<SortedSet<Type>, ComparableExaminingIteratorWrapper<Type>>(){
			
			int i = 0;
			
			@Override
			public ComparableExaminingIteratorWrapper<Type> adapt(final SortedSet<Type> source) {
				return new ComparableExaminingIteratorWrapper<Type>(source.iterator(), (sortedSets.isEmpty() ? null : sortedSets.iterator().next().comparator()) ,this.i++);
			}
		}, new TreeSet<ComparableExaminingIteratorWrapper<Type>>());
	}
	
	@Override
	public boolean hasNext() {
		return !this.iterators.isEmpty();
	}

	@Override
	public Type next() {
		if(this.hasNext()){
			ComparableExaminingIteratorWrapper<Type> first = this.iterators.first();
			this.iterators.removeFirst();
			final Type next = first.next();
			if(first.hasNext()){
				this.iterators.add(first);
			}
			while(!this.iterators.isEmpty() && this.compare(this.iterators.first().examineNext(),next) == 0){
				first = this.iterators.first();
				this.iterators.removeFirst();
				first.skipNext();
				if(first.hasNext()){
					this.iterators.add(first);
				}
			}
			return next;
		}
		else{
			throw new NoSuchElementException();//TODO better exception
		}
	}
	
	protected static class ComparableExaminingIteratorWrapper<Type> extends ExaminingIteratorWrapper<Type> implements Comparable<ComparableExaminingIteratorWrapper<Type>>{
		
		private final Comparator<? super Type> comparator;
		
		final int id;
		
		public ComparableExaminingIteratorWrapper(final Iterator<Type> iterator, final Comparator<? super Type> comparator, final int id) {
			super(iterator);
			this.comparator = comparator;
			this.id = id;
		}

		@Override
		public int compareTo(final ComparableExaminingIteratorWrapper<Type> i2) {
			final int cmp =  this.compare(this.examineNext(), i2.examineNext());
			if(cmp == 0){
				return this.id - i2.id;
			}
			else{
				return cmp;
			}
		}
		
		@SuppressWarnings("unchecked")
		@Override
		public boolean equals(final Object obj) {
			if(obj instanceof ComparableExaminingIteratorWrapper){
				return this.compareTo((ComparableExaminingIteratorWrapper<Type>)obj) == 0;
			}
			else{
				return false;
			}
		}
		
		public final int compare(final Type t1, final Type t2){
			return AbstractComparator.compare(t1, t2, this.comparator);
		}
		
	}

}
