package de.lmu.ifi.pms.java.lang.iteration;

import java.util.Iterator;
import java.util.Map.Entry;
import java.util.SortedMap;
import java.util.TreeMap;

public class TreeMapIterator<KeyType, ValueType>  extends AbstractIterator<TreeMap<KeyType, ValueType>>{

	private final TreeMap<KeyType, Iterator<ValueType>> iteratorMap;
	
	public TreeMapIterator(final SortedMap<KeyType, ? extends Iterable<ValueType>> iterableMap) {
		this.iteratorMap = new TreeMap<KeyType, Iterator<ValueType>>(iterableMap.comparator());
		for(final Entry<KeyType, ? extends Iterable<ValueType>> entry: iterableMap.entrySet()){
			this.iteratorMap.put(entry.getKey(), entry.getValue().iterator());
		}
	}
	
	@Override
	public boolean hasNext() {
		for(final Iterator<ValueType> iterator: this.iteratorMap.values()){
			if(!iterator.hasNext()){
				return false;
			}
		}
		return true;
	}
	
	@Override
	public TreeMap<KeyType, ValueType> next() {
		final TreeMap<KeyType, ValueType> treeMap = new TreeMap<KeyType, ValueType>(this.iteratorMap.comparator());
		for(final Entry<KeyType, Iterator<ValueType>> entry: this.iteratorMap.entrySet()){
			treeMap.put(entry.getKey(), entry.getValue().next());
		}
		return treeMap;
	}
	
	@SuppressWarnings("deprecation")
	@Override
	public void remove() {
		for(final Iterator<ValueType> iterator: this.iteratorMap.values()){
			iterator.remove();
		}
	}
	
}
