package de.lmu.ifi.pms.java.math;

import de.lmu.ifi.pms.java.lang.ArithmeticException;

public final class Math {
	
	static{
		//System.loadLibrary("UtilSource");
	}
	
	public final static int MaxIntValue = Integer.MAX_VALUE;
	
	public final static int MinIntValue = -MaxIntValue;
	
	public final static long MaxLongValue = Long.MAX_VALUE;
	
	public final static long MinLongValue = -MaxLongValue;
	
	public final static int TwoExp30 = 0x40000000;

	public final static long TwoExp62 = 0x4000000000000000L;
	
	public final static float IntMaxValueAsFloat = toFloat(MaxIntValue);
	
	public final static float IntMinValueAsFloat = toFloat(MinIntValue);
	
	public final static double IntMaxValueAsDouble = toDouble(MaxIntValue);
	
	public final static double IntMinValueAsDouble = toDouble(MinIntValue);
	
	public final static float LongMaxValueAsFloat = toFloat(MaxLongValue);
	
	public final static float LongMinValueAsFloat = toFloat(MinLongValue);
	
	public final static double LongMaxValueAsDouble = toDouble(MaxLongValue);
	
	public final static double LongMinValueAsDouble = toDouble(MinLongValue);
	
	private Math() {

	}
	
	public static int sign(final int n){
		return java.lang.Integer.signum(n);
	}
	
	public static long sign(final long n){
		return java.lang.Long.signum(n);
	}
	
	public static float sign(final float f){
		return java.lang.Math.signum(f);
	}
	
	public static double sign(final double d){
		return java.lang.Math.signum(d);
	}
	
	public static int abs(final int n){
		return java.lang.Math.abs(n);
	}
	
	public static long abs(final long n){
		return java.lang.Math.abs(n);		
	}
	
	public static float abs(final float f){
		return java.lang.Math.abs(f);
	}
	
	public static double abs(final double d){
		return java.lang.Math.abs(d);
	}
	 
	//public static native int plus(int m, int n);
	
	public static int plus(final int m, final int n){
		int k = m + n;
		//(m ^ n) < 0 iff m and n have different signs
		//~(m ^ k) < 0 iff m and k have same signs
		//(m ^ n) | ~(m ^ k)) < 0 iff m and n have different signs or m, n and k have the same sign
		//m and n have different sign => no overflow
		//m, n and k have the same sign 
		//k != -2147483648 => k > -2147483648
		if( ((m ^ n) | ~(m ^ k)) < 0 & k != -2147483648){//More efficient implementation?
			return k;
		}
		else{
			if(m < 0){
				throw new ArithmeticException("Underflow");
			}
			else{
				throw new ArithmeticException("Overflow");
			}
		}
	}
	
	public static long plus(final long m, final long n){
		long k = m + n;
		if( ((m ^ n) | ~(m ^ k)) < 0 & k != -9223372036854775808l){//More efficient implementation?
			return k;
		}
		else{
			if(m < 0){
				throw new ArithmeticException("Underflow");
			}
			else{
				throw new ArithmeticException("Overflow");
			}
		}
	}
	
	public static float plus(final float f1, final float f2){
		return f1 + f2;
	}
	
	public static double plus(final double d1, final double d2){
		return d1 + d2;
	}
	
	public static int sub(final int m, final int n){
		return plus(m, -n);
	}
	
	public static long sub(final long m, final long n){
		return plus(m, -n);
	}
	
	public static float sub(final float f1, final float f2){
		return f1 - f2;
	}
	
	public static double sub(final double d1, final double d2){
		return d1 - d2;
	}
	
	public static int mult(final int m, final int n){
		if(m != 0 && (-2147483647 / m > n || n > 2147483647 / m)){
			throw new ArithmeticException("Overflow");
		}
		else{
			return m * n;
		}
	}
	
	public static long mult(final long m, final long n){
		if(m != 0 && (-9223372036854775807l / m > n || n > 9223372036854775807l / m)){
			throw new ArithmeticException("Overflow");
		}
		else{
			return m * n;
		}
	}
	
	public static float mult(final float f1, final float f2){
		return f1 * f2;
	}
	
	public static double mult(final double d1, final double d2){
		return d1 * d2;
	}
	
	public static int div(final int m, final int n){//TODO more efficient implementation
		return (m - mod(m, n))/n;
	}
	
	public static long div(final long m, final long n){//TODO more efficient implementation
		return (m - mod(m, n))/n;
	}
	
	public static float div(final float f1, final float f2){
		return f1 / f2;
	}
	
	public static double div(final double d1, final double d2){
		return d1 / d2;
	}
	
	public static int mod(final int m, final int n){
		if ((m ^ n) < 0){//Unterschiedliche Vorzeichen
			return ((m % n) + n) % n;
		}	
		else{
			return m % n;
		}
	}
	
	public static long mod(final long m, final long n){
		if ((m ^ n) < 0){//Unterschiedliche Vorzeichen
			return ((m % n) + n) % n;
		}	
		else{
			return m % n;
		}
	}
	
	public static int potenz(final int m, int n){//n muss >= 0 sein
		int p = 1;//TODO more efficient implemention
		for(;n > 0; --n){
			p = mult(p, m);
		}
		return m;
	}
	
	public static long potenz(final long m, long n){//n muss >= 0 sein
		long p = 1;//TODO more efficient implemention
		for(;n > 0; --n){
			p = mult(p, m);
		}
		return m;
	}
	
	public static boolean equal(final int m, final int n){
		return m == n;
	}
	
	public static boolean greater(final int m, final int n){
		return m > n;
	}
	
	public static boolean greaterOrEqual(final int m, final int n){
		return m >= n;
	}
	
	public static boolean less(final int m, final int n){
		return m < n;
	}
	
	public static boolean lessOrEqual(final int m, final int n){
		return m <= n;
	}
	
	public static int min(final int m, final int n){
		return (greater(m,n) ? n : m);
	}
	
	public static int max(final int m, final int n){
		return (greater(m,n) ? m : n);
	}
	
	public static int max(final int... numbers){
		if(numbers.length > 0){
			int max = numbers[0];
			int n;
			for(int i = 1; i < numbers.length; ++i)
			{
				n = numbers[i];
				if(n > max){
					max = n;
				}
			}
			return max;
		}
		else{
			throw new IllegalArgumentException(); //TODO better Exception
		}
	}
	
	public static int min(final int... numbers){
		if(numbers.length > 0){
			int min = numbers[0];
			int n;
			for(int i = 1; i < numbers.length; ++i)
			{
				n = numbers[i];
				if(n < min){
					min = n;
				}
			}
			return min;
		}
		else{
			throw new IllegalArgumentException(); //TODO better Exception
		}
	}
	
	public static boolean equal(final long m, final long n){
		return m == n;
	}
	
	public static boolean greater(final long m, final long n){
		return m > n;
	}
	
	public static boolean greaterOrEqual(final long m, final long n){
		return m >= n;
	}
	
	public static boolean less(final long m, final long n){
		return m < n;
	}
	
	public static boolean lessOrEqual(final long m, final long n){
		return m <= n;
	}
	
	public static long min(final long m, final long n){
		return (greater(m,n) ? n : m);
	}
	
	public static long max(final long m, final long n){
		return (greater(m,n) ? m : n);
	}
	
	public static long max(final long... numbers){
		if(numbers.length > 0){
			long max = numbers[0];
			long n;
			for(int i = 1; i < numbers.length; ++i)
			{
				n = numbers[i];
				if(n > max){
					max = n;
				}
			}
			return max;
		}
		else{
			throw new IllegalArgumentException(); //TODO better Exception
		}
	}
	
	public static long min(final long... numbers){
		if(numbers.length > 0){
			long min = numbers[0];
			long n;
			for(int i = 1; i < numbers.length; ++i)
			{
				n = numbers[i];
				if(n < min){
					min = n;
				}
			}
			return min;
		}
		else{
			throw new IllegalArgumentException(); //TODO better Exception
		}
	}
	
	public static boolean equal(final float f1, final float f2){
		return f1 == f2;
	}
	
	public static boolean greater(final float f1, final float f2){
		return f1 > f2;
	}
	
	public static boolean greaterOrEqual(final float f1, final float f2){
		return f1 >= f2;
	}
	
	public static boolean less(final float f1, final float f2){
		return f1 < f2;
	}
	
	public static boolean lessOrEqual(final float f1, final float f2){
		return f1 <= f2;
	}
	
	public static boolean equal(final double d1, final double d2){
		return d1 == d2;
	}
	
	public static boolean greater(final double d1, final double d2){
		return d1 > d2;
	}
	
	public static boolean greaterOrEqual(final double d1, final double d2){
		return d1 >= d2;
	}
	
	public static boolean less(final double d1, final double d2){
		return d1 < d2;
	}
	
	public static boolean lessOrEqual(final double d1, final double d2){
		return d1 <= d2;
	}

	
	private static int positivGgt(int m, final int n){
		if(m >= n){
			m = mod(m, n);
			if(m == 0){
				return n;
			}
			else{
				return positivGgt(n, m);
			}
		}
		else{
			return positivGgt(n,m);
		}
	}
	
	public static int ggt(final int m, int n){
		n = positivGgt(abs(m), abs(n));
		return (m > 0 ? n : -n);
	}
	
	private static long positivGgt(long m, final long n){
		if(m >= n){
			m = mod(m, n);
			if(m == 0){
				return n;
			}
			else{
				return positivGgt(n, m);
			}
		}
		else{
			return positivGgt(n,m);
		}
	}
	
	public static long ggt(final long m, long n){
		n = positivGgt(abs(m), abs(n));
		return (m > 0 ? n : -n);
	}
	
	public static int floorlog2(final int n){
		if(n < 0){
			throw new ArithmeticException("log from number smaller than 0");
		}
		else{
			return 31 - Integer.numberOfLeadingZeros(n);
		}
	}
	
	public static int floorlog2(final long n){
		if(n < 0){
			throw new ArithmeticException("log from number smaller than 0");
		}
		else{
			return 63 - Long.numberOfLeadingZeros(n);
		}
	}
	
	public static int ceillog2(final int n){
		if(n < 0){
			throw new ArithmeticException("log from number smaller than 0");
		}
		else{
			return 32 - Integer.numberOfLeadingZeros(n-1);
		}
	}
	
	public static int ceillog2(final long n){
		if(n < 0){
			throw new ArithmeticException("log from number smaller than 0");
		}
		else{
			return 64 - Long.numberOfLeadingZeros(n-1);
		}
	}
	
	public static float log(final float f, final float base){
		return toFloat(log(toDouble(f), toDouble(base)));
	}
	
	public static double log(final double d, final double base){
		return div(java.lang.Math.log(d), java.lang.Math.log(base));
	}
	
	public static int sum(final int[] values){
		int s = 0;
		for(int value: values){
			s = plus(s, value);
		}
		return s;
	}
	
	public static long sum(final long[] values){
		long s = 0;
		for(long value: values){
			s = plus(s, value);
		}
		return s;
	}
	
	public static float sum(final float[] values){
		float s = 0;
		for(float value: values){
			s = plus(s, value);
		}
		return s;
	}
	
	public static double sum(final double[] values){
		double s = 0;
		for(double value: values){
			s = plus(s, value);
		}
		return s;
	}
	
	
	public static int prod(final int[] values){
		int p = 1;
		for(int value: values){
			p = mult(p, value);
		}
		return p;
	}
	
	public static long prod(final long[] values){
		long p = 1;
		for(long value: values){
			p = mult(p, value);
		}
		return p;
	}
	
	public static float prod(final float[] values){
		float p = 1;
		for(float value: values){
			p = mult(p, value);
		}
		return p;
	}
	
	public static double prod(final double[] values){
		double p = 1;
		for(double value: values){
			p = mult(p, value);
		}
		return p;
	}
	
	public static float mean(final float[] values){
		return sum(values) / values.length;
	}
	
	public static double mean(final double[] values){
		return sum(values) / values.length;
	}
	
	public static void scalarMult(final double scalar, final double[] values, final double[] result){
		assert values.length == result.length;
		for(int i = 0; i < result.length; ++i){
			result[i] = scalar * values[i];
		}
	}
	
	public static void abs(final double[] values, final double[] result){
		assert values.length == result.length;
		for(int i = 0; i < result.length; ++i){
			result[i] = abs(values[i]);
		}
	}
	
	public static void add(final double[] values1, final double[] values2, final double[] result){
		assert values1.length == values2.length && values2.length == result.length;
		for(int i = 0; i < result.length; ++i){
			result[i] = values1[i] + values2[i];
		}
	}
	
	public static void add(final double[] values1, final double value2, final double[] result){
		assert values1.length == result.length;
		for(int i = 0; i < result.length; ++i){
			result[i] = values1[i] + value2;
		}
	}
	
	public static void sub(final double[] values1, final double[] values2, final double[] result){
		assert values1.length == values2.length && values2.length == result.length;
		for(int i = 0; i < result.length; ++i){
			result[i] = values1[i] - values2[i];
		}
	}
	
	public static void sub(final double[] values1, final double value2, final double[] result){
		assert values1.length == result.length;
		for(int i = 0; i < result.length; ++i){
			result[i] = values1[i] - value2;
		}
	}
	
	public static void vectorProd(final double[] values1, final double[] values2, final double[] result){
		assert values1.length == values2.length && values2.length == result.length;
		for(int i = 0; i < result.length; ++i){
			result[i] = values1[i] * values2[i];
		}
	}
	
	public static double scalarProd(final double[] values1, final double[] values2){
		double[] vectorProduct = new double[values1.length];
		vectorProd(values1, values2, vectorProduct);
		return sum(vectorProduct);
	}
	
	public static int[] newRat(final int n){
		int[] result = new int[3];
		result[0] = n;
		result[2] = 1;
		return result;
	}
	
	public static long[] newRat(final long n){
		long[] result = new long[3];
		result[0] = n;
		result[2] = 1;
		return result;
	}
	
	public static int[] newRat(final int m, final int n){
		int[] result = new int[3];
		result[1] = m;
		result[2] = n;
		if(n == 0){
			throw new IllegalArgumentException();
			//TODO better Exception
		}
		normalize(result, result);
		return result;
	}
	
	public static long[] newRat(final long m, final long n){
		long[] result = new long[3];
		result[1] = m;
		result[2] = n;
		if(n == 0){
			throw new IllegalArgumentException();
			//TODO better Exception
		}
		normalize(result, result);
		return result;
	}
	
	public static int[] newRat(final int l, final int m, final int n){
		int[] result = new int[3];
		result[0] = l;
		result[1] = m;
		result[2] = n;
		if(n == 0){
			throw new IllegalArgumentException();
			//TODO better Exception
		}
		normalize(result, result);
		return result;
	}
	
	public static long[] newRat(final long l, final long m, final long n){
		long[] result = new long[3];
		result[0] = l;
		result[1] = m;
		result[2] = n;
		if(n == 0){
			throw new IllegalArgumentException();
			//TODO better Exception
		}
		normalize(result, result);
		return result;
	}
	
	public static int[] newRat(final int[] rat){
		if(rat.length == 1){
			return newRat(rat[0]);
		}
		else if(rat.length == 2){
			return newRat(rat[0],rat[1]);
		}
		else if(rat.length == 3){
			return newRat(rat[0], rat[1], rat[2]);
		}
		else{
			throw new IllegalArgumentException();
			//TODO better Exception
		}
	}
	
	public static long[] newRat(final long[] rat){
		if(rat.length == 1){
			return newRat(rat[0]);
		}
		else if(rat.length == 2){
			return newRat(rat[0],rat[1]);
		}
		else if(rat.length == 3){
			return newRat(rat[0], rat[1], rat[2]);
		}
		else{
			throw new IllegalArgumentException();
			//TODO better Exception
		}
	}
	
	public static int[] normalize(final int[] rat, final int[] result){
		result[0] = plus(rat[0], div(rat[1],rat[2]));
		if(rat[2] > 0){
			result[1] = mod(rat[1],rat[2]);
			result[2] = abs(rat[2]);
		}
		else{
			result[1] = -mod(rat[1],rat[2]);
			result[2] = abs(rat[2]);
		}		
		if(result[0] > 0 && result[1] < 0){
			result[0] = result[0] - 1; //No Overflow possible here
			result[1] = result[2] +  result[1]; //No Overflow possible here
		}
		else if(result[0] < 0 && result[1] > 0){
			result[0] = result[0] + 1; //No Overflow possible here
			result[1] = result[1] - result[2];//No Overflow possible here
		}
		if(result[1] == 0){
			result[2] = 1;
		}
		kuerzen(result, result);
		return result;
	}
	
	public static long[] normalize(final long[] rat, final long[] result){
		result[0] = plus(rat[0], div(rat[1],rat[2]));
		if(rat[2] > 0){
			result[1] = mod(rat[1],rat[2]);
			result[2] = abs(rat[2]);
		}
		else{
			result[1] = -mod(rat[1],rat[2]);
			result[2] = abs(rat[2]);
		}		
		if(result[0] > 0 && result[1] < 0){
			result[0] = result[0] - 1; //No Overflow possible here
			result[1] = result[2] +  result[1]; //No Overflow possible here
		}
		else if(result[0] < 0 && result[1] > 0){
			result[0] = result[0] + 1; //No Overflow possible here
			result[1] = result[1] - result[2];//No Overflow possible here
		}
		if(result[1] == 0){
			result[2] = 1;
		}
		kuerzen(result, result);
		return result;
	}
	
	public static boolean isZero(final int[] rat){
		return rat[0] == 0 && rat[1] == 0;
	}
	
	public static boolean isZero(final long[] rat){
		return rat[0] == 0 && rat[1] == 0;
	}
	
	public static void zero(final int[] result){
		result[0] = 0;
		result[1] = 0;
		result[2] = 1;
	}
	
	public static void zero(final long[] result){
		result[0] = 0;
		result[1] = 0;
		result[2] = 1;
	}
	
	public static int[] kuerzen(final int[] rat, final int[] result){
		result[0] = rat[0];
		if(rat[1] == 0){
			result[1] = 0;
			result[2] = 1;
		}
		else{
			int ggt = ggt(rat[2],rat[1]);
			result[1] = div(rat[1],ggt);
			result[2] = div(rat[2],ggt);
		}
		return result;
	}
	
	public static long[] kuerzen(final long[] rat, final long[] result){
		result[0] = rat[0];
		if(rat[1] == 0){
			result[1] = 0;
			result[2] = 1;
		}
		else{
			long ggt = ggt(rat[2],rat[1]);
			result[1] = div(rat[1],ggt);
			result[2] = div(rat[2],ggt);
		}
		return result;
	}
	
	public static int ratSign(final int[] rat){
		if(rat[0] == 0){
			return sign(rat[1]);
		}
		else{
			return sign(rat[0]);
		}
	}
	
	public static long ratSign(final long[] rat){
		if(rat[0] == 0){
			return sign(rat[1]);
		}
		else{
			return sign(rat[0]);
		}
	}
	
	public static int[] ratAbs(final int[] rat, final int[] result){
		result[0] = abs(rat[0]);
		result[1] = abs(rat[1]);
		result[2] = rat[2];
		return result;
	}
	
	public static long[] ratAbs(final long[] rat, final long[] result){
		result[0] = abs(rat[0]);
		result[1] = abs(rat[1]);
		result[2] = rat[2];
		return result;
	}
	
	public static int[] ratNeg(final int[] rat, final int[] result){
		result[0] = -rat[0];
		result[1] = -rat[1];
		result[2] = rat[2];
		return result;
	}
	
	public static long[] ratNeg(final long[] rat, final long[] result){
		result[0] = -rat[0];
		result[1] = -rat[1];
		result[2] = rat[2];
		return result;
	}
	
	public static int precision(final int[] rat){
		return floorlog2(rat[2]) + 1;
	}
	
	public static int precision(final long[] rat){
		return floorlog2(rat[2]) + 1;
	}
	
	public static int[] cutPrecision(final int[] rat, int precision, final int[] result){
		if(precision < 1){
			throw new IllegalArgumentException();
		}
		else{
			precision = precision(rat) - precision;
			if(precision > 0){
				result[0] = rat[0];
				result[1] = rat[1] >> precision;
				result[2] = rat[2] >> precision;
				normalize(result, result);
			}
			else{
				result[0] = rat[0];
				result[1] = rat[1];
				result[2] = rat[2];
			}
			return result;
		}
	}
	
	protected static void cutPrecision(final int[] rat1, final int[] rat2, final int[] result1, final int[] result2){
		int p1 = precision(rat1);
		int p2;
		if(p1 < 15){
			p2 = 30 - p1;
		}
		else{
			p2 = precision(rat2);
			if(p2 < 15){
				p1 = 30 - p2;
			}
			else{
				p1 = 15;
				p2 = 15;
			}
		}
		cutPrecision(rat1, p1, result1);
		cutPrecision(rat2, p1, result2);
	}
	
	public static long[] cutPrecision(final long[] rat, long precision, final long[] result){
		if(precision < 1){
			throw new IllegalArgumentException();
		}
		else{
			precision = precision(rat) - precision;
			if(precision > 0){
				result[0] = rat[0];
				result[1] = rat[1] >> precision;
				result[2] = rat[2] >> precision;
				return normalize(result, result);
			}
			else{
				result[0] = rat[0];
				result[1] = rat[1];
				result[2] = rat[2];
			}
			return result;
		}
	}
	
	protected static void cutPrecision(final long[] rat1, final long[] rat2, final long[] result1, final long[] result2){
		long p1 = precision(rat1);
		long p2;
		if(p1 < 15){
			p2 = 30 - p1;
		}
		else{
			p2 = precision(rat2);
			if(p2 < 15){
				p1 = 30 - p2;
			}
			else{
				p1 = 15;
				p2 = 15;
			}
		}
		cutPrecision(rat1, p1, result1);
		cutPrecision(rat2, p1, result2);
	}
	
	public static int[] ratPlus(final int[] rat1, final int[] rat2, final int[] result){
		int ggt = ggt(rat1[2],rat2[2]);		
		int n = div(rat2[2],ggt);
		try{
			result[1] = plus(mult(rat1[1], n) , mult(rat2[1], div(rat1[2],ggt)));
		}
		catch(ArithmeticException e){
			int[] result2 = new int[3];
			cutPrecision(rat1, rat2, result, result2);
			return ratPlus(result, result2, result);
		}
		try{
			result[2] = mult(rat1[2], n);
		}
		catch(ArithmeticException e){
			int[] result2 = new int[3];
			cutPrecision(rat1, rat2, result, result2);
			return ratPlus(result, result2, result);
		}
		result[0] = plus(rat1[0], rat2[0]);
		return normalize(result,result);
	}
	
	public static long[] ratPlus(final long[] rat1, final long[] rat2, final long[] result){
		long ggt = ggt(rat1[2],rat2[2]);		
		long n = div(rat2[2],ggt);
		try{
			result[1] = plus(mult(rat1[1], n) , mult(rat2[1], div(rat1[2],ggt)));
		}
		catch(ArithmeticException e){
			long[] result2 = new long[3];
			cutPrecision(rat1, rat2, result, result2);
			return ratPlus(result, result2, result);
		}
		try{
			result[2] = mult(rat1[2], n);
		}
		catch(ArithmeticException e){
			long[] result2 = new long[3];
			cutPrecision(rat1, rat2, result, result2);
			return ratPlus(result, result2, result);
		}
		result[0] = plus(rat1[0], rat2[0]);
		return normalize(result,result);
	}
	
	public static int[] ratSub(final int[] rat1, final int[] rat2, final int[] result){
		return ratPlus(rat1,ratNeg(rat2, newRat(0)), result);
	}
	
	public static long[] ratSub(final long[] rat1, final long[] rat2, final long[] result){
		return ratPlus(rat1,ratNeg(rat2, rat2), result);
	}
	
	public static int[] ratMult(final int[] rat1, final int[] rat2, final int[] result){
		try{
			int[] k1 = {0,rat1[1],rat2[2]};
			int[] k2 = {0,rat2[1],rat1[2]};
			kuerzen(k1, k1);
			kuerzen(k2, k2);
			k1[0] = mult(rat1[0],rat2[0]);
			k1[1] = mult(k1[1] , k2[1]);
			k1[2] = mult(k1[2] , k2[2]);
			k2[1] = rat2[0];
			k2[2] = rat1[2];
			kuerzen(k2, k2);
			k2[1] = mult(k2[1], rat1[1]);
			ratPlus(k1, k2, k1);
			k2[1] = rat1[0];
			k2[2] = rat2[2];
			kuerzen(k2, k2);
			k2[1] = mult(k2[1] , rat2[1]);
			ratPlus(k1, k2, result);
			return result;
		}
		catch(ArithmeticException e){
			int[] result2 = new int[3];
			cutPrecision(rat1, rat2, result, result2);
			return ratMult(result, result2, result);
		}
	}		
	
	public static long[] ratMult(final long[] rat1, final long[] rat2, final long[] result){
		try{
			long[] k1 = {0,rat1[1],rat2[2]};
			long[] k2 = {0,rat2[1],rat1[2]};
			kuerzen(k1, k1);
			kuerzen(k2, k2);
			k1[0] = mult(rat1[0],rat2[0]);
			k1[1] = mult(k1[1] , k2[1]);
			k1[2] = mult(k1[2] , k2[2]);
			k2[1] = rat2[0];
			k2[2] = rat1[2];
			kuerzen(k2, k2);
			k2[1] = mult(k2[1], rat1[1]);
			ratPlus(k1, k2, k1);
			k2[1] = rat1[0];
			k2[2] = rat2[2];
			kuerzen(k2, k2);
			k2[1] = mult(k2[1] , rat2[1]);
			ratPlus(k1, k2, result);
			return result;
		}
		catch(ArithmeticException e){
			long[] result2 = new long[3];
			cutPrecision(rat1, rat2, result, result2);
			return ratPlus(result, result2, result);
		}
	}
	
	public static int[] ratInv(final int[] rat, final int[] result){
		if(rat[0] == 0){//Betrag kleiner 1
			result[0] = 0;
			result[1] = rat[2];
			result[2] = rat[1];
			normalize(result, result);
		}
		else{//Betrag gr�sser 1
			try{
				int n = rat[2];
				result[2] = plus(rat[1], mult(rat[0], n));
				result[1] = n;	
				result[0] = 0;
			}
			catch(ArithmeticException e){//TODO �berpr�fen, insbesondere ob es auch f�r 'plus' funktioniert
				int s = ceillog2(rat[2]);
				s = max(min(s - 1, ceillog2(abs(rat[0])) + s - 30),1);
				result[0] = rat[0];
				result[1] = rat[1] >>> s;
				result[2] = rat[2] >>> s;
				ratInv(result, result);
			}
		}
		return result;
	}		
	
	public static long[] ratInv(final long[] rat, final long[] result){
		if(rat[0] == 0){//Betrag kleiner 1
			result[0] = 0;
			result[1] = rat[2];
			result[2] = rat[1];
			normalize(result, result);
		}
		else{//Betrag gr�sser 1
			try{
				long n = rat[2];
				result[2] = plus(rat[1], mult(rat[0], n));
				result[1] = n;	
				result[0] = 0;
			}
			catch(ArithmeticException e){//TODO �berpr�fen, insbesondere ob es auch f�r 'plus' funktioniert
				long s = ceillog2(rat[2]);
				s = max(min(s - 1, ceillog2(abs(rat[0])) + s - 30),1);
				result[0] = rat[0];
				result[1] = rat[1] >>> s;
				result[2] = rat[2] >>> s;
				ratInv(result, result);
			}
		}
		return result;
	}	
	
	public static int[] ratDiv(final int[] rat1, final int[] rat2, final int[] result){
		if(isZero(rat2)){
			throw new ArithmeticException("Division by zero");
		}
		else if(isZero(rat1)){
			zero(result);
		}
		else{
			try{
				int[] k1 = {0,ggt(rat1[0],rat1[1]),ggt(rat2[0],rat2[1])};
				int[] k2 = {0,plus(mult(div(rat1[0],k1[1]), rat1[2]), div(rat1[1],k1[1])),plus(mult(div(rat2[0],k1[2]), rat2[2]), div(rat2[1],k1[2]))};
				result[0] = 0;
				result[1] = rat2[2];
				result[2] = rat1[2];
				normalize(k1, k1);
				normalize(k2, k2);
				normalize(result,result);
				ratMult(k2, result, result);
				ratMult(k1, result, result);
			}
			catch (ArithmeticException e) {
				return ratMult(rat1, ratInv(rat2, new int[3]), result);//TODO better Implementation?
			}
		}
		return result;
	}
	
	public static long[] ratDiv(final long[] rat1, final long[] rat2, final long[] result){		
		if(isZero(rat2)){
			throw new ArithmeticException("Division by zero");
		}
		else if(isZero(rat1)){
			zero(result);
		}
		else{
			try{
				long[] k1 = {0l,ggt(rat1[0],rat1[1]),ggt(rat2[0],rat2[1])};
				long[] k2 = {0l,plus(mult(div(rat1[0],k1[1]), rat1[2]), div(rat1[1],k1[1])),plus(mult(div(rat2[0],k1[2]), rat2[2]), div(rat2[1],k1[2]))};
				result[0] = 0l;
				result[1] = rat2[2];
				result[2] = rat1[2];
				normalize(k1, k1);
				normalize(k2, k2);
				normalize(result,result);
				ratMult(k2, result, result);
				ratMult(k1, result, result);
			}
			catch (ArithmeticException e) {
				return ratMult(rat1, ratInv(rat2, new long[3]), result);//TODO better Implementation?
			}
		}
		return result;
	}
	
	public static boolean ratEqual(final int[] rat1, final int[] rat2){
		return rat1[0] == rat2[0] && rat1[1] == rat2[1] && rat1[2] == rat2[2]; 
	}
	
	public static boolean ratGreater(final int[] rat1, final int[] rat2){
		if(rat1[0] > rat2[0]){
			return true;
		}
		else{
			return ratSign(ratSub(rat1, rat2, newRat(0))) > 0; 
		}		
	}
	
	public static boolean ratGreaterOrEqual(final int[] rat1, final int[] rat2){
		return ratSign(ratSub(rat1, rat2, newRat(0))) >= 0; 
	}
	
	public static boolean ratLess(final int[] rat1, final int[] rat2){
		return ratGreater(rat2, rat1); 
	}
	
	public static boolean ratLessOrEqual(final int[] rat1, final int[] rat2){
		return ratGreaterOrEqual(rat2, rat1); 
	}
	
	public static boolean ratEqual(final long[] rat1, final long[] rat2){
		return rat1[0] == rat2[0] && rat1[1] == rat2[1] && rat1[2] == rat2[2]; 
	}
	
	public static boolean ratGreater(final long[] rat1, final long[] rat2){
		if(rat1[0] > rat2[0]){
			return true;
		}
		else{
			return ratSign(ratSub(rat1, rat2, newRat((long)0))) > 0; 
		}		
	}
	
	public static boolean ratGreaterOrEqual(final long[] rat1, final long[] rat2){
		return ratSign(ratSub(rat1, rat2, newRat((long)0))) >= 0; 
	}
	
	public static boolean ratLess(final long[] rat1, final long[] rat2){
		return ratGreater(rat2, rat1); 
	}
	
	public static boolean ratLessOrEqual(final long[] rat1, final long[] rat2){
		return ratGreaterOrEqual(rat2, rat1); 
	}
	
	public static int drop(final int[] rat){
		return rat[0];
	}
	
	public static long drop(final long[] rat){
		return rat[0];
	}
	
	public static int round(final int[] rat){//Achtung: Wertebereich!!
		if(rat[1] == 0){
			return rat[0];
		}
		else if(rat[1] > 0){
			return (ratLess(rat,newRat(rat[0],1,2)) ? rat[0] : plus(rat[0],1));
		}
		else{
			return (ratLess(rat,newRat(rat[0],-1,2)) ? sub(rat[0],1) : rat[0]);
		}		
	}
	
	public static long round(final long[] rat){//Achtung: Wertebereich!!
		if(rat[1] == 0){
			return rat[0];
		}
		else if(rat[1] > 0){
			return (ratLess(rat,newRat(rat[0],1,2)) ? rat[0] : plus(rat[0],1));
		}
		else{
			return (ratLess(rat,newRat(rat[0],-1,2)) ? sub(rat[0],1) : rat[0]);
		}			
	}	
	
	public static long round(final float f){
		return round(toDouble(f));
	}
	
	public static long round(final double d){
		return java.lang.Math.round(d);
	}
	
	public static long drop(final float f){
		return drop(toDouble(f));
	}
	
	public static long drop(final double d){
		if(d >= 0d){
			return toLong(java.lang.Math.floor(d));
		}
		else{
			return toLong(java.lang.Math.ceil(d));
		}
	}
	
	public static int toInt(final int i){
		return i;
	}
	
	public static int toInt(final long l){
		if(MinIntValue <= l && l <= MaxIntValue){
			return (int) l;
		}
		else{
			throw new ArithmeticException();//TODO better exception
		}
	}
	
	public static int toInt(final float f){
		if(IntMinValueAsFloat <= f && f <= IntMaxValueAsFloat){
			return (int) f;
		}
		else{
			throw new ArithmeticException();//TODO better exception
		}
	}
	
	public static int toInt(final double d){
		if(IntMinValueAsDouble <= d && d <= IntMaxValueAsDouble){
			return (int) d;
		}
		else{
			throw new ArithmeticException();//TODO better exception
		}
	}
	
	public static int toInt(final int[] rat){
		return round(rat);
	}
	
	public static int toInt(final long[] rat){
		return toInt(round(rat));
	}
	
	@SuppressWarnings("cast")
	public static long toLong(final int i){
		return (long) i;
	}
	
	public static long toLong(final long l){
		return l;
	}
	
	public static long toLong(final float f){
		if(LongMinValueAsFloat <= f && f <= LongMaxValueAsFloat){
			return (long) f;
		}
		else{
			throw new ArithmeticException();//TODO better exception
		}
	}
	
	public static long toLong(final double d){
		if(LongMinValueAsDouble <= d && d <= LongMaxValueAsDouble){
			return (long) d;
		}
		else{
			throw new ArithmeticException();//TODO better exception
		}
	}
	
	public static long toLong(final int[] rat){
		return toLong(round(rat));//TODO �berpr�fen Wertebereich
	}
	
	public static long toLong(final long[] rat){
		return round(rat);
	}
	
	@SuppressWarnings("cast")
	public static float toFloat(final int i){
		return (float) i;
	}
	
	@SuppressWarnings("cast")
	public static float toFloat(final long l){
		return (float) l;
	}
	
	public static float toFloat(final float f){
		return f;
	}
	
	public static float toFloat(final double d){
		return (float) d;
	}
	
	public static float toFloat(final int[] rat){
		return toFloat(rat[0]) + toFloat(rat[1]) / toFloat(rat[2]);
	}
	
	public static float toFloat(final long[] rat){
		return toFloat(rat[0]) + toFloat(rat[1]) / toFloat(rat[2]);
	}
	
	@SuppressWarnings("cast")
	public static double toDouble(final int i){
		return (double) i;
	}
	
	@SuppressWarnings("cast")
	public static double toDouble(final long l){
		return (double) l;
	}
	
	@SuppressWarnings("cast")
	public static double toDouble(final float f){
		return (double) f;
	}
	
	public static double toDouble(final double d){
		return d;
	}
	
	public static double toDouble(final int[] rat){
		return toDouble(rat[0]) + toDouble(rat[1]) / toDouble(rat[2]);
	}
	
	public static double toDouble(final long[] rat){
		return toDouble(rat[0]) + toDouble(rat[1]) / toDouble(rat[2]);
	}
	
	public static int[] toIntRat(final int i){
		return newRat(i);
	}
	
	public static int[] toIntRat(final long l){
		return newRat(toInt(l));
	}
	
	public static int[] toIntRat(final float f){//TODO better conversion?
		int i = Math.toInt(drop(f));
		return newRat(i,toInt(drop((f - i) * TwoExp30)), TwoExp30);
	}
	
	public static int[] toIntRat(final double d){//TODO better conversion?
		int i = Math.toInt(drop(d));
		return newRat(i,toInt(drop((d - i) * TwoExp30)), TwoExp30);
	}
	
	public static int[] toIntRat(final int[] rat){
		int[] irat = new int[rat.length];
		for(int i = 0; i < rat.length; ++i){
			irat[i] = rat[i];
		}
		return irat;
	}
	
	public static int[] toIntRat(final long[] rat){
		int[] irat = new int[rat.length];
		for(int i = 0; i < rat.length; ++i){
			irat[i] = toInt(rat[i]);
		}
		return irat;
	}
	
	public static long[] toLongRat(final int i){
		return newRat(toLong(i));
	}
	
	public static long[] toLongRat(final long l){
		return newRat(l);
	}
	
	public static long[] toLongRat(final float f){//TODO better conversion?
		long l = Math.toInt(drop(f));
		return newRat(l,toInt(drop((f - l) * TwoExp30)), TwoExp30);
	}
	
	public static long[] toLongRat(final double d){//TODO better conversion?
		long l = Math.toInt(drop(d));
		return newRat(l,toInt(drop((d - l) * TwoExp30)), TwoExp30);
	}
	
	public static long[] toLongRat(final int[] rat){
		long[] lrat = new long[rat.length];
		for(int i = 0; i < rat.length; ++i){
			lrat[i] = toLong(rat[i]);
		}
		return lrat;
	}
	
	public static long[] toLongRat(final long[] rat){
		long[] lrat = new long[rat.length];
		for(int i = 0; i < rat.length; ++i){
			lrat[i] = rat[i];
		}
		return lrat;
	}
	
	public static double sqrt(final double d){
		return java.lang.Math.sqrt(d);
	}
	
	public static void main(final String[] args){
		/*
		System.out.println(mod(5,3));
		System.out.println(mod(-5,3));
		System.out.println(mod(5,-3));
		System.out.println(mod(-5,-3));
		System.out.println(mod(6,2));
		System.out.println(mod(-6,2));
		System.out.println(mod(6,-2));
		System.out.println(mod(-6,-2));

		System.out.println(div(5,3));
		System.out.println(div(-5,3));
		System.out.println(div(5,-3));
		System.out.println(div(-5,-3));
		System.out.println(div(6,2));
		System.out.println(div(-6,2));
		System.out.println(div(6,-2));
		System.out.println(div(-6,-2));*/
		int i = Integer.MIN_VALUE;
		String s = Integer.toBinaryString(i);
		s = "00000000000000000000000000000000".substring(s.length()) + s;
		System.out.println(s);

	}
	

}
