package de.lmu.ifi.pms.java.util;

import de.lmu.ifi.pms.java.math.Math;
import de.lmu.ifi.pms.java.util.collection.array.Arrays;


public class Buffer<Type>{

	private Type[] values;
	
	private int start = -1;
	
	private int end = 0;
	
	public Buffer(final int capacity) {
		this(capacity, (Class<Type>)Object.class);
	}
	
	public Buffer(final int capacity, final Class<Type> type) {
		super();
		this.values = Arrays.getArray(type, capacity);
	}
	
	protected void setCapacity(final int capacity){
		if(this.values == null){
			this.values = (Type[])new Object[capacity];
		}
		else{
			this.values = Arrays.getArray((Class<Type>)this.values.getClass().getComponentType(), capacity);
		}
	}
	
	protected void setCapacity(final int capacity, final Class<Type> type){
		this.values = Arrays.getArray(type, capacity);
	}

	public int capacity(){
		return this.values.length;
	}
	
	public boolean isEmpty(){
		return this.start == -1;
	}
	
	protected void waitWhileEmpty(){
		if(this.isEmpty()){
			throw new IllegalStateException(); //TODO better Exception
		}
	}
	
	public boolean isFull(){
		return this.end == this.start;
	}
	
	protected void waitWhileFull(){
		if(this.isFull()){
			throw new IllegalStateException(); //TODO better Exception
		}
	}
	
	public int size(){
		if(this.isEmpty()){
			return 0;
		}
		else if(this.isFull()){
			return this.capacity();
		}
		else{
			return Math.mod(this.end - this.start, this.capacity());
		}
	}
	
	public int remainingCapacity(){
		return this.capacity() - this.size();
	}
	
	protected void waitForSize(final int size){
		if(this.size() < size){
			throw new IllegalStateException(); //TODO better Exception
		}
	}
	
	protected void signalAdd(){

	}
	
	public void add(final Type value){
		this.addLast(value);
	}
	
	public void addFirst(final Type value){
		this.waitWhileFull();
		if(this.isEmpty()){
			this.start = this.end;
		}
		this.start = Math.mod(this.start -1, this.capacity());
		this.values[this.start] = value;
		this.signalAdd();
	}
	
	public void addLast(final Type value){
		this.waitWhileFull();
		if(this.isEmpty()){
			this.start = this.end;
		}
		this.values[this.end] = value;
		this.end = Math.mod(this.end + 1, this.capacity());
		this.signalAdd();
	}
	
	public boolean has(final int index){
		return 0 <= index && index < this.size();
	}
	
	public boolean hasFirst(){
		return !this.isEmpty();
	}
	
	public boolean hasLast(){
		return !this.isEmpty();
	}
	
	public Type get(final int index){
		if(index < this.capacity()){
			this.waitForSize(index +1 );
			return this.values[Math.mod(this.start + index, this.capacity())];
		}
		else{
			throw new IndexOutOfBoundsException();//TODO betterexception
		}
	}
	
	public Type first(){
		this.waitWhileEmpty();
		return this.values[this.start];
	}
	
	public Type last(){
		this.waitWhileEmpty();
		return this.values[Math.mod(this.end - 1, this.capacity())];
	}
	
	protected void signalRemove(){
		
	}
	
	public void remove(){
		this.removeFirst();
	}
	
	public void removeFirst(){
		this.waitWhileEmpty();
		this.values[this.start] = null;
		this.start = Math.mod(this.start + 1, this.capacity());
		if(this.start == this.end){
			this.start = -1;
			this.end = 0;
		}
		this.signalRemove();
	}
	
	public void removeLast(){
		this.waitWhileEmpty();
		this.end = Math.mod(this.end - 1, this.capacity());
		this.values[this.end] = null;
		if(this.start == this.end){
			this.start = -1;
			this.end = 0;
		}
		this.signalRemove();
	}
	
}
