package de.lmu.ifi.pms.java.util.collection.array;

import de.lmu.ifi.pms.java.lang.iteration.ExtendibleIterator;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.extendible.ExtendibleCollection;
import de.lmu.ifi.pms.java.util.collection.javaWrapper.FromJavaCollectionWrapper;


public abstract class AbstractExtendibleArray<Type> extends AbstractModifyableArray<Type> implements ExtendibleCollection<Type>{
	
	public AbstractExtendibleArray(){
		
	}
	
	@Override
	public ExtendibleArrayIterator<Type, ? extends AbstractExtendibleArray<Type>> iterator() {
		return new ExtendibleArrayIterator<Type, AbstractExtendibleArray<Type>>(this);
	}
	
	public static class ExtendibleArrayIterator<Type, ArrayType extends AbstractExtendibleArray<Type>> extends ModifyableArrayIterator<Type, ArrayType> implements ExtendibleIterator<Type>{

		public ExtendibleArrayIterator(final ArrayType array, final int firstIndexInclusive, final int lastIndexExclusive, final int position) {
			super(array, firstIndexInclusive, lastIndexExclusive, position);
		}

		public ExtendibleArrayIterator(final ArrayType array) {
			super(array);
		}
		
		/* (non-Javadoc)
		 * @see extension.java.lang.iteration.ExtendibleIterator#insert(java.lang.Object)
		 */
		@Override
		public void insert(final Type value) {
			 this.array.insert(this.position - 1, value);
		}
		
		/* (non-Javadoc)
		 * @see extension.java.lang.iteration.AbstractIterator#remove()
		 */
		@SuppressWarnings("deprecation")
		@Override
		public void remove() {
			 this.array.remove(this.position - 1);
		}

		/* (non-Javadoc)
		 * @see extension.java.util.collection.array.AbstractArray.ArrayIterator#add(java.lang.Object)
		 */
		@SuppressWarnings("deprecation")
		@Override
		public final void add(final Type e) {
			this.insert(e);
		}
	}
	
	@Override
	public abstract AbstractExtendibleArray<Type> subArray(int offset, int length);

	@Override
	public boolean add(final Type value){
		this.addLast(value);
		return true;
	}
	
	public void addFirst(final Type value){
		this.insert(this.minIndexInc(), value);
	}
	
	public abstract void addLast(Type value);
	
	public abstract void insert(int index, Type value);
	
	public abstract void setSize(int size);
	
	public abstract void setExpSize(int exp);
	
	public abstract void remove(int index);
	
	public void removeFirst(){
		this.remove(this.minIndexInc());
	}
	
	public void removeFirst(int n){
		for(;n > 0; --n){
			this.removeFirst();
		}
	}
	
	public abstract void removeLast();
	
	public void removeLast(int n) {
		for(;n > 0;--n){
			this.removeLast();
		}
	}
	
	@Override
	public boolean addAll(final java.util.Collection<? extends Type> c) {
		return this.addAll(new FromJavaCollectionWrapper<Type>(c));
	}
	
	/* (non-Javadoc)
	 * @see extension.java.util.collection.AbstractCollection#addAll(java.util.Collection)
	 */
	@Override
	public boolean addAll(final Collection<? extends Type> c) {
		boolean modified = false;
		for(Type t: c){
			if(this.add(t)){
				modified = true;
			}
		}
		return modified;
	}
	
	@Override
	//@SuppressWarnings("deprecation")
	public void clear() {
		this.removeLast(this.size());
	}

	@Override
	public boolean remove(final Object o) {
		ExtendibleIterator<?> e = this.iterator();
		if (o==null) {
		    while (e.hasNext()) {
			if (e.next()==null) {
			    e.remove();
			    return true;
			}
		    }
		} else {
		    while (e.hasNext()) {
			if (o.equals(e.next())) {
			    e.remove();
			    return true;
			}
		    }
		}
		return false;
	}
	
	
	@SuppressWarnings({ "unchecked", "rawtypes" })
	@Override
	public boolean removeAll(final java.util.Collection<?> c) {
		return this.removeAll(new FromJavaCollectionWrapper(c));
	}

	@Override
	public boolean removeAll(final Collection<?> c) {
		boolean modified = false;
		ExtendibleIterator<?> e = this.iterator();
		while (e.hasNext()) {
		    if (c.contains(e.next())) {
			e.remove();
			modified = true;
		    }
		}
		return modified;
	}
	
	@SuppressWarnings({ "unchecked", "rawtypes" })
	@Override
	public boolean retainAll(final java.util.Collection<?> c) {
		return this.retainAll(new FromJavaCollectionWrapper(c));
	}

	@Override
	public boolean retainAll(final Collection<?> c) {
		boolean modified = false;
		ExtendibleIterator<Type> e = this.iterator();
		while (e.hasNext()) {
		    if (!c.contains(e.next())) {
			e.remove();
			modified = true;
		    }
		}
		return modified;
	}
	
}
