/**
 * 
 */
package de.lmu.ifi.pms.java.util.collection.extendible.list;

import java.util.NoSuchElementException;


public abstract class AbstractDoubleLinkedEntry<Type, ListEntryType extends AbstractDoubleLinkedEntry<Type, ListEntryType>> extends AbstractSingleLinkedEntry<Type, ListEntryType> implements DoubleLinkedListEntry<Type, ListEntryType>{
	
	protected ListEntryType next;

	public AbstractDoubleLinkedEntry() {
		super();
	}

	public AbstractDoubleLinkedEntry(final AbstractSingleLinkedList<Type, ListEntryType> linkingList) {
		super(linkingList);
	}
	
	@SuppressWarnings("unchecked")
	public AbstractDoubleLinkedEntry(final ListEntryType head) {
		super(head.linkingList);
		this.setNext(head);
		head.setPrevious((ListEntryType)this);
	}

	@Override
	public boolean hasPrevious() {
		return !this.previous().isFoot();
	}

	@Override
	public boolean hasNext() {
		return !this.isHead();
	}

	@Override
	public boolean isHead() {
		return this.next == null;
	}

	protected void checkNext(){
		if(this.hasNext()){
			throw new NoSuchElementException();
		}
	}

	@Override
	public ListEntryType next() {
		this.checkNext();
		return this.next;
	}
	
	protected void setNext(final ListEntryType entry){
		this.next = entry;
	}
	
	@SuppressWarnings("unchecked")
	@Override
	public void insertBefore(final ListEntryType entry) {
		super.insertBefore(entry);
		entry.setNext((ListEntryType)this);
		this.previous.setPrevious(entry);
	}

	@Override
	public void insertBehind(final ListEntryType entry) {
		this.next.insertBefore(entry);
	}

	@Override
	public void remove() {
		this.unlink();
		this.previous.setNext(this.next());
		this.next.setPrevious(this.previous());
	}



	@Override
	public void removeFollowing() {
		while(this.hasNext()){
			this.removeNext();
		}
	}

	@Override
	public void removeNext() {
		this.next().remove();
	}
	
}