package de.lmu.ifi.pms.java.util.map.fixed;

import de.lmu.ifi.pms.java.util.collection.array.Arrays;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;



public class FixedEnumMap<KeyType extends Enum<KeyType>, ValueType> extends FixedArrayNavigableMap<KeyType, ValueType>{
	
	protected final int offset;
	
	public FixedEnumMap(final Class<KeyType> enumType, final ValueType... values) {
		this(enumType.getEnumConstants(), values);
	}
	
	protected FixedEnumMap(final KeyType[] all_enum_constants, final ValueType[] values) {
		this(all_enum_constants[0], all_enum_constants[all_enum_constants.length-1], values);
	}

	public FixedEnumMap(final KeyType start, final KeyType end, final ValueType... values) {
		super( new FixedEnumSet<KeyType>(start, end), new FixedArray<ValueType>(false, values));
		this.offset = start.ordinal();
	}
	
	public static class FixedEnumSet<Type extends Enum<Type>> extends FixedArrayNavigableSet<Type>{
		
		protected final int offset;
		
		public FixedEnumSet(final Type start, final Type end) {
			super(true);
			this.offset = start.ordinal();
			Type[] all_values = start.getDeclaringClass().getEnumConstants();
			Type[] sub_values = Arrays.getArray(all_values, Math.max(end.ordinal() - this.offset + 1, 0));
			System.arraycopy(all_values, this.offset, sub_values, 0, sub_values.length);
			this.fillBody(false, false, sub_values);
		}
		
		@Override
		protected int search(final Type e) {
			int ordinal = e.ordinal();
			int index = ordinal - this.offset;
			if(this.has(index)){
				return index;
			}
			else{
				return -1;
			}
		}
		
		@Override
		public FixedArrayNavigableSet<Type> subArray(final int offset, final int length) {
			return super.subArray(offset, length);
			//TODO return FixedEnumSet
		}
		
		
	}

}
