package de.lmu.ifi.pms.java.util.string;

import java.io.IOException;
import java.io.Writer;
import java.nio.CharBuffer;

public class StringBuilder extends Writer implements Readable, CharSequence{

	private boolean closed = false;
	private int length = 0;
	
	private final java.util.LinkedList<String> strings = new java.util.LinkedList<String>();
	
	public StringBuilder() {
		super();	
	}
	
	public void subString(int beginIndex, int endIndex){
		if(beginIndex >= 0 && endIndex <= this.length() && beginIndex <= endIndex){
			if(beginIndex < endIndex){
				this.length = this.length - beginIndex;
				endIndex = endIndex - beginIndex;
				while(beginIndex >= this.strings.getFirst().length()){
					beginIndex = beginIndex - this.strings.getFirst().length();
					this.strings.removeFirst();
				}
				String first = this.strings.getFirst();
				this.strings.removeFirst();
				this.strings.addFirst(first.substring(beginIndex));
				endIndex = this.length() - endIndex;
				while(endIndex >= this.strings.getLast().length()){
					this.strings.removeLast();
				}
				String last = this.strings.getLast();
				this.strings.removeLast();
				this.strings.addLast(last.substring(0, last.length() - endIndex));
			}
			else{// beginIndex == endIndex
				this.strings.clear();
				this.length = 0;
			}
		}
		else{
			throw new IndexOutOfBoundsException();//TODO better Exception;
		}		
	}
	
	public void subString(final int beginIndex){
		this.subString(beginIndex, this.length());
	}
	
	public void clear(){
		this.length = 0;
		this.strings.clear();
	}
	
	public Writer append(final String string){
		this.strings.addLast(string);
		this.length = this.length + string.length();
		return this;
	}
	
	public Writer append(final boolean i){
		this.append(String.valueOf(i));
		return this;
	}
	
	public Writer append(final byte i){
		this.append(String.valueOf(i));
		return this;
	}
	
	public Writer append(final short i){
		this.append(String.valueOf(i));
		return this;
	}
	
	public Writer append(final int i){
		this.append(String.valueOf(i));
		return this;
	}
	
	public Writer append(final long i){
		this.append(String.valueOf(i));
		return this;
	}
	
	public Writer append(final float i){
		this.append(String.valueOf(i));
		return this;
	}
	
	public Writer append(final double i){
		this.append(String.valueOf(i));
		return this;
	}
	
	@Override
	public Writer append(final char i){
		this.append(String.valueOf(i));
		return this;
	}
	
	@Override
	public Writer append(final CharSequence csq) throws IOException {
		this.append(csq.toString());
		return this;
	}
	
	@Override
	public Writer append(final CharSequence csq, final int start, final int end) throws IOException {
		this.append(csq.subSequence(start, end));
		return this;
	}
	
	public void append(final Iterable<String> strings){
		for(String string: strings){
			this.append(string);
		}
	}
	
	@Override
	public int length(){
		return this.length;
	}
	
	public void flatten(){
		java.lang.StringBuilder buffer = new java.lang.StringBuilder(this.length());
		for(String string: this.strings){
			buffer.append(string);
		}
		this.strings.clear();
		this.strings.add(buffer.toString());
	}
	
	@Override
	public String toString() {
		if(this.strings.size() == 0){
			return "";
		}
		else if(this.strings.size() == 1){
			return this.strings.getFirst();
		}
		else{
			this.flatten();
			return this.toString();
		}
	}
	
	@Override
	public void write(final char[] cbuf, final int off, final int len){		
		this.append(new String(cbuf,off,len));
	}
	
	@Override
	public void write(final String str, final int off, final int len){
		this.write(str.substring(off, off + this.length));
	}
	
	@Override
	public void write(final String str){
		if(this.closed){
			throw new IllegalStateException("Already closed!");
		}
		else{
			this.append(str);
		}
	}
	
	@Override
	public void close(){
		this.closed = true;
	}
	
	@Override
	public void flush(){

	}

	@Override
	public int read(final CharBuffer cb) throws IOException {
		if(this.closed){
			return -1;
		}
		else{
			int readLength = 0;
			int remainingCapacity = cb.remaining();
			String string;
			while(remainingCapacity > 0 && !this.strings.isEmpty()){			
				string = this.strings.getFirst();
				int length = string.length();
				if(length <= remainingCapacity){
					cb.append(string);
					readLength += length;
					this.length -= length;
					remainingCapacity -= length;
					this.strings.removeFirst();
				}
				else{
					cb.append(string, 0, remainingCapacity);
					readLength += remainingCapacity;
					this.length -= remainingCapacity;
					remainingCapacity = 0;
					this.strings.removeFirst();
					this.strings.addFirst(string.substring(remainingCapacity, length));
				}
			}
			return readLength;
		}
	}
	
	@Override
	public CharSequence subSequence(final int start, final int end) {
		return this.toString().subSequence(start, end);
		//TODO better implementation?
	}
	
	@Override
	public char charAt(int index) {
		if(0 <= index & index < this.length()){
			for(String string: this.strings){
				if(index < string.length()){
					return string.charAt(index);
				}
				else{
					index-=string.length();
				}
			}
			throw new Error();
		}
		else{
			throw new IndexOutOfBoundsException();
		}
	}
	
}
