package de.lmu.ifi.pms.cep.unitTests.dura;

import java.io.File;
import java.io.FileFilter;
import java.io.IOException;
import java.util.ArrayList;

import nl.cwi.monetdb.event_mill.MonetDriver;

import org.junit.BeforeClass;
import org.junit.Test;

import de.lmu.ifi.pms.cep.compilation.Compiler;
import de.lmu.ifi.pms.cep.compilation.StackCompiler;
import de.lmu.ifi.pms.cep.dura.compiler.dura.DuraCompiler;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraCompilationException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraParseException;
import de.lmu.ifi.pms.cep.event_mill.sql.SQL_Program_Spec;
import de.lmu.ifi.pms.cep.event_mill.sql.compiler.DTSA_SQL_Compiler;
import de.lmu.ifi.pms.cep.event_mill.sql.compiler.DTSA_SQL_WriterImpl;
import de.lmu.ifi.pms.cep.tsa.TSA_Program;
import de.lmu.ifi.pms.cep.tsa.compiler.TSA_DTSA_Compiler;
import de.lmu.ifi.pms.cep.tsa.compiler.TSA_DefaultDifferencer;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_ComprehensiveException;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_RuntimeException;


public class DuraTests {
	
	String testFileDirectory = "src/de/lmu/ifi/pms/cep/unitTests/dura/files/";
	String exampleFileDirectory = "examples/";
	
	static Compiler<TSA_Program, SQL_Program_Spec> tsaSQLCompiler;
	
	@BeforeClass
	public static void init() throws ClassNotFoundException {
		tsaSQLCompiler = new StackCompiler<TSA_Program, SQL_Program_Spec>(TSA_Program.class, SQL_Program_Spec.class, new TSA_DTSA_Compiler(new TSA_DefaultDifferencer()), new DTSA_SQL_Compiler(new DTSA_SQL_WriterImpl(new MonetDriver().translation())));
	}
	
	@Test
	public void playground() throws IOException {
		File file = new File(this.testFileDirectory + "Playground.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}
	
	@Test
	public void compileExamples() {
		File dir = new File(this.exampleFileDirectory);
		
		FileFilter duraFilter = new FileFilter() {
			@Override
		    public boolean accept(final File file) {
		        return file.isFile() && file.getName().endsWith(".dura");
		    }
		};

		ArrayList<Throwable> exceptions = new ArrayList<Throwable>();
		for (File file : dir.listFiles(duraFilter)) {
			try{
				DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
			}
			catch (TSA_RuntimeException e) {
				Throwable t;
				if(e.getMessage() != null){
					t = new TSA_RuntimeException("\nException in file \"" + file.getName() + "\"\n" + e.getMessage(), e.getCause());					
				}
				else{
					t = new TSA_RuntimeException("\nException in file \"" + file.getName() + "\"\n", e.getCause());
				}
				t.setStackTrace(new StackTraceElement[0]);
				exceptions.add(t);
			}
			catch (DuraCompilationException e) {
				Throwable t;
				if(e.getMessage() != null){
					t = new DuraCompilationException("\nException in file \"" + file.getName() + "\"\n" +  e.getMessage(), e.getCause());					
				}
				else{
					t = new DuraCompilationException("\nException in file \"" + file.getName() + "\"\n",  e.getCause());
				}
				t.setStackTrace(new StackTraceElement[0]);
				exceptions.add(t);
			}
			catch (RuntimeException e) {
				exceptions.add(new RuntimeException("\nException in file \"" + file.getName() + "\"" , e));
			}
			catch (Exception e) {
				exceptions.add(new Exception("\nException in file \"" + file.getName() + "\"", e));
			}
			catch (Error e) {
				exceptions.add(new Error("\nError in file \"" + file.getName() , e));
			}
		}
		if(exceptions.size() > 0){
			throw new RuntimeException(TSA_ComprehensiveException.message(exceptions));
		}
	}
	
	@Test
	public void foreignKeys() throws IOException {
		File file = new File(this.testFileDirectory + "ForeignKeys.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}

	
	@Test
	public void reactiveRule() throws IOException {
		File file = new File(this.testFileDirectory + "ReactiveRules.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}
	
	@Test
	public void implicitActionStreams() throws IOException {
		File file = new File(this.testFileDirectory + "ImplicitActionStreams.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}
	
	@Test
	public void implicitStatefulObjectStreams() throws IOException {
		File file = new File(this.testFileDirectory + "ImplicitStatefulObjectStreams.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}
	
	@Test
	public void streamConstraints() throws IOException {
		File file = new File(this.testFileDirectory + "StreamConstraints.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}

	@Test
	public void timeIntervalFunctionsAndRelations() throws IOException {
		File file = new File(this.testFileDirectory + "TimeIntervalFunctionsAndRelations.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}
	
	@Test
	public void grouping() throws IOException {
		File file = new File(this.testFileDirectory + "Grouping.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}
	
	@Test
	public void literals() throws IOException {
		File file = new File(this.testFileDirectory + "Literals.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}
	
	@Test
	public void simpleModule() throws IOException {
		File file = new File(this.testFileDirectory + "SimpleModule.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}
	
	@Test
	public void negation() throws IOException {
		File file = new File(this.testFileDirectory + "Negation.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}
	
	@Test
	public void statefulObject() throws IOException {
		File file = new File(this.testFileDirectory + "StatefulObject.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}
	
	@Test
	public void atomicQuerySupplement() throws IOException {
		File file = new File(this.testFileDirectory + "AtomicQuerySupplement.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}
	
	@Test
	public void pathConditions() throws IOException {
		File file = new File(this.testFileDirectory + "PathConditions.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}
	
	@Test
	public void dynamicStatefulObjects() throws IOException {
		File file = new File(this.testFileDirectory + "DynamicStatefulObjects.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}
	
	@Test
	public void dynamicStatefulObjectsMinimal() throws IOException {
		File file = new File(this.testFileDirectory + "DynamicStatefulObjectsMinimal.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}
	
	@Test(expected=DuraCompilationException.class)
	public void syntaxRestrictionsStreamName() throws IOException {
		File file = new File(this.testFileDirectory + "SyntaxRestrictionsStreamName.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}
	
	@Test(expected=DuraCompilationException.class)
	public void syntaxRestrictionsAttributeName() throws IOException {
		File file = new File(this.testFileDirectory + "SyntaxRestrictionsAttributeName.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}
	
	@Test(expected=DuraParseException.class)
	public void parseException() throws IOException {
		File file = new File(this.testFileDirectory + "ParseException.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}
	
	@Test
	public void staticStatefulObjects() throws IOException {
		File file = new File(this.testFileDirectory + "StaticStatefulObjects.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}
	
	@Test(expected=DuraCompilationException.class)
	public void filteredAttributes2() throws IOException {
		File file = new File(this.testFileDirectory + "FilteredAttributes2.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}
	
	@Test(expected=DuraCompilationException.class)
	public void filteredAttributes3() throws IOException {
		File file = new File(this.testFileDirectory + "FilteredAttributes3.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}
	
	@Test(expected=DuraCompilationException.class)
	public void filteredAttributes4() throws IOException {
		File file = new File(this.testFileDirectory + "FilteredAttributes4.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}
	
	@Test(expected=DuraCompilationException.class)
	public void filteredAttributes() throws IOException {
		File file = new File(this.testFileDirectory + "FilteredAttributes.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}
	
	@Test(expected=DuraCompilationException.class)
	public void rangeRestrictionNegation() throws IOException {
		File file = new File(this.testFileDirectory + "RangeRestrictionNegation.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}
	
	@Test
	public void negatedStaticStatefulObject() throws IOException {
		File file = new File(this.testFileDirectory + "NegatedStaticStatefulObject.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}
	
	@Test
	public void statefulObjectGrouping() throws IOException {
		File file = new File(this.testFileDirectory + "StatefulObjectGrouping.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}
	
	@Test(expected=DuraCompilationException.class)
	public void incompatibleTypes() throws IOException {
		File file = new File(this.testFileDirectory + "IncompatibleTypes.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}
	
	@Test(expected=DuraCompilationException.class)
	public void incompatibleTypes2() throws IOException {
		File file = new File(this.testFileDirectory + "IncompatibleTypes2.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}
	
	@Test(expected=DuraCompilationException.class)
	public void incompatibleTypes3() throws IOException {
		File file = new File(this.testFileDirectory + "IncompatibleTypes3.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}
	
	@Test
	public void termTypes() throws IOException {
		File file = new File(this.testFileDirectory + "TermTypes.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}
	
	@Test
	public void timeIntervalBinding() throws IOException {
		File file = new File(this.testFileDirectory + "TimeIntervalBinding.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}
	
	@Test
	public void exists() throws IOException {
		File file = new File(this.testFileDirectory + "Exists.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}
	
	@Test
	public void eventIdentifier() throws IOException {
		File file = new File(this.testFileDirectory + "EventIdentifier.dura");
		
		DuraTests.tsaSQLCompiler.compile(DuraCompiler.compile("", "test", file));
	}
	public static void main(final String[] args) {
		new DuraTests().compileExamples();
	}
}
